@@ -59,8 +59,6 @@ typedef struct mapcache_context_apache mapcache_context_apache;
 typedef struct mapcache_context_apache_request mapcache_context_apache_request;
 typedef struct mapcache_context_apache_server mapcache_context_apache_server;
 
-apr_pool_t *pchild = NULL;
-
 struct mapcache_context_apache {
   mapcache_context ctx;
 };
@@ -75,6 +73,19 @@ struct mapcache_context_apache_request {
   request_rec *request;
 };
 
+typedef struct mapcache_alias_entry mapcache_alias_entry;
+
+struct mapcache_alias_entry {
+  char *endpoint;
+  char *configfile;
+  mapcache_cfg *cfg;
+  mapcache_connection_pool *cp;
+};
+
+struct mapcache_server_cfg {
+  apr_array_header_t *aliases; /**< list of mapcache configurations aliased to a server uri */
+};
+
 void apache_context_server_log(mapcache_context *c, mapcache_log_level level, char *message, ...)
 {
   mapcache_context_apache_server *ctx = (mapcache_context_apache_server*)c;
@@ -185,21 +196,42 @@ static mapcache_context_apache_request* apache_request_context_create(request_re
 {
   mapcache_context_apache_request *ctx = apr_pcalloc(r->pool, sizeof(mapcache_context_apache_request));
   mapcache_server_cfg *cfg = NULL;
-  mapcache_cfg *config = NULL;
+  const char *mapcache_alias;
+  mapcache_alias_entry *alias_entry;
+  int i;
+  mapcache_context *mctx = (mapcache_context*)ctx;
 
-  ctx->ctx.ctx.pool = r->pool;
-  ctx->ctx.ctx.process_pool = pchild;
+  mctx->pool = r->pool;
 #ifdef APR_HAS_THREADS
-  ctx->ctx.ctx.threadlock = thread_mutex;
+  mctx->threadlock = thread_mutex;
 #endif
 
   /* lookup the configuration object given the configuration file name */
   cfg = ap_get_module_config(r->server->module_config, &mapcache_module);
-  config = apr_hash_get(cfg->aliases,(void*)r->filename,APR_HASH_KEY_STRING);
-  ctx->ctx.ctx.config = config;
-  ctx->request = r;
-  init_apache_request_context(ctx);
-  return ctx;
+  if(!cfg || !cfg->aliases) {
+    return NULL;
+  }
+
+  mapcache_alias = apr_table_get(r->notes,"mapcache_alias_entry");
+  //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "using mapcache config %s", mapcache_config_file);
+  if(!mapcache_alias) {
+    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mapcache module bug? no mapcache_alias_entry found");
+    return NULL;
+  }
+
+  for(i=0; i<cfg->aliases->nelts; i++) {
+    alias_entry = APR_ARRAY_IDX(cfg->aliases,i,mapcache_alias_entry*);
+    if(strcmp(alias_entry->endpoint,mapcache_alias))
+      continue;
+
+    mctx->config = alias_entry->cfg;
+    ctx->request = r;
+    mctx->connection_pool = alias_entry->cp;
+    init_apache_request_context(ctx);
+    return ctx;
+  }
+  ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r, "mapcache module bug? no mapcache_alias_entry found for %s",mapcache_alias);
+  return NULL;
 }
 
 static mapcache_context_apache_server* apache_server_context_create(server_rec *s, apr_pool_t *pool)
@@ -212,6 +244,73 @@ static mapcache_context_apache_server* apache_server_context_create(server_rec *
   return ctx;
 }
 
+/* read post body. code taken from "The apache modules book, Nick Kew" */
+static void read_post_body(mapcache_context_apache_request *ctx, mapcache_request_proxy *p) {
+  request_rec *r = ctx->request;
+  mapcache_context *mctx = (mapcache_context*)ctx;
+  int bytes,eos;
+  apr_bucket_brigade *bb, *bbin;
+  apr_bucket *b;
+  apr_status_t rv;
+  const char *clen = apr_table_get(r->headers_in, "Content-Length");
+  if(clen) {
+    bytes = strtol(clen, NULL, 0);
+    if(bytes >= p->rule->max_post_len) {
+      mctx->set_error(mctx, HTTP_REQUEST_ENTITY_TOO_LARGE, "post request too big");
+      return;
+    }
+  } else {
+    bytes = p->rule->max_post_len;
+  }
+
+  bb = apr_brigade_create(mctx->pool, r->connection->bucket_alloc);
+  bbin = apr_brigade_create(mctx->pool, r->connection->bucket_alloc);
+  p->post_len = 0;
+
+  do {
+    apr_bucket *nextb;
+    rv = ap_get_brigade(r->input_filters, bbin, AP_MODE_READBYTES, APR_BLOCK_READ, bytes);
+    if(rv != APR_SUCCESS) {
+      mctx->set_error(mctx, 500, "failed to read form input");
+      return;
+    }
+    for(b = APR_BRIGADE_FIRST(bbin); b != APR_BRIGADE_SENTINEL(bbin); b = nextb) {
+      nextb = APR_BUCKET_NEXT(b);
+      if(APR_BUCKET_IS_EOS(b)) {
+        eos = 1;
+      }
+      if(!APR_BUCKET_IS_METADATA(b)) {
+        if(b->length != (apr_size_t)(-1)) {
+          p->post_len += b->length;
+          if(p->post_len > p->rule->max_post_len) {
+            apr_bucket_delete(b);
+          }
+        }
+      }
+      if(p->post_len <= p->rule->max_post_len) {
+        APR_BUCKET_REMOVE(b);
+        APR_BRIGADE_INSERT_TAIL(bb, b);
+      }
+    }
+  } while (!eos);
+
+  if(p->post_len > p->rule->max_post_len) {
+    mctx->set_error(mctx, HTTP_REQUEST_ENTITY_TOO_LARGE, "request too big");
+    return;
+  }
+
+  p->post_buf = apr_palloc(mctx->pool, p->post_len+1);
+  
+  if(p->post_len> 0) {
+    rv = apr_brigade_flatten(bb, p->post_buf, &(p->post_len));
+    if(rv != APR_SUCCESS) {
+      mctx->set_error(mctx, 500, "error (flatten) reading form data");
+      return;
+    }
+  }
+  p->post_buf[p->post_len] = 0;
+}
+
 static int write_http_response(mapcache_context_apache_request *ctx, mapcache_http_response *response)
 {
   request_rec *r = ctx->request;
@@ -239,7 +338,7 @@ static int write_http_response(mapcache_context_apache_request *ctx, mapcache_ht
       }
     }
   }
-  if(response->data) {
+  if(response->data && response->data->size) {
     ap_set_content_length(r,response->data->size);
     ap_rwrite((void*)response->data->buf, response->data->size, r);
   }
@@ -251,10 +350,21 @@ static int write_http_response(mapcache_context_apache_request *ctx, mapcache_ht
 
 static void mod_mapcache_child_init(apr_pool_t *pool, server_rec *s)
 {
-  pchild = pool;
 #ifdef APR_HAS_THREADS
   apr_thread_mutex_create(&thread_mutex,APR_THREAD_MUTEX_DEFAULT,pool);
 #endif
+  for( ; s ; s=s->next) {
+    mapcache_server_cfg* cfg = ap_get_module_config(s->module_config, &mapcache_module);
+    int i,rv;
+    for(i=0;i<cfg->aliases->nelts;i++) {
+      mapcache_alias_entry *alias_entry = APR_ARRAY_IDX(cfg->aliases,i,mapcache_alias_entry*);
+      rv = mapcache_connection_pool_create(&(alias_entry->cp),pool);
+      ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, s, "creating a child process mapcache connection pool on server %s for alias %s", s->server_hostname, alias_entry->endpoint);
+      if(rv!=APR_SUCCESS) {
+        ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s, "failed to create mapcache connection pool");
+      }
+    }
+  }
 }
 
 static int mod_mapcache_request_handler(request_rec *r)
@@ -268,16 +378,25 @@ static int mod_mapcache_request_handler(request_rec *r)
   if (!r->handler || strcmp(r->handler, "mapcache")) {
     return DECLINED;
   }
-  if (r->method_number != M_GET) {
+  if (r->method_number != M_GET && r->method_number != M_POST) {
     return HTTP_METHOD_NOT_ALLOWED;
   }
 
 
   apache_ctx = apache_request_context_create(r);
+
+  if(!apache_ctx) {
+    return DECLINED;
+  }
+
   global_ctx = (mapcache_context*)apache_ctx;
+  global_ctx->supports_redirects = 1;
+  global_ctx->headers_in = r->headers_in;
 
   params = mapcache_http_parse_param_string(global_ctx, r->args);
 
+  //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "mapcache dispatch %s",r->path_info);
+
   mapcache_service_dispatch_request(global_ctx,&request,r->path_info,params,global_ctx->config);
   if(GC_HAS_ERROR(global_ctx) || !request) {
     return write_http_response(apache_ctx,
@@ -316,7 +435,45 @@ static int mod_mapcache_request_handler(request_rec *r)
     mapcache_request_get_tile *req_tile = (mapcache_request_get_tile*)request;
     http_response = mapcache_core_get_tile(global_ctx,req_tile);
   } else if( request->type == MAPCACHE_REQUEST_PROXY ) {
+    const char *buf;
     mapcache_request_proxy *req_proxy = (mapcache_request_proxy*)request;
+    if(r->method_number == M_POST) {
+      read_post_body(apache_ctx, req_proxy);
+      if(GC_HAS_ERROR(global_ctx)) {
+        return write_http_response(apache_ctx, mapcache_core_respond_to_error(global_ctx));
+      }
+      if(!req_proxy->headers) {
+        req_proxy->headers = apr_table_make(global_ctx->pool, 2);
+      }
+      apr_table_set(req_proxy->headers, "Content-Type", r->content_type);
+      if((buf = apr_table_get(r->headers_in,"X-Forwarded-For"))) {
+#if (AP_SERVER_MAJORVERSION_NUMBER == 2) && (AP_SERVER_MINORVERSION_NUMBER < 4)
+        apr_table_set(req_proxy->headers, "X-Forwarded-For", apr_psprintf(global_ctx->pool,"%s, %s", buf, r->connection->remote_ip));
+#else
+        apr_table_set(req_proxy->headers, "X-Forwarded-For", apr_psprintf(global_ctx->pool,"%s, %s", buf, r->connection->client_ip));
+#endif
+      } else {
+#if (AP_SERVER_MAJORVERSION_NUMBER == 2) && (AP_SERVER_MINORVERSION_NUMBER < 4)
+        apr_table_set(req_proxy->headers, "X-Forwarded-For", r->connection->remote_ip);
+#else
+        apr_table_set(req_proxy->headers, "X-Forwarded-For", r->connection->client_ip);
+#endif
+      }
+      if ((buf = apr_table_get(r->headers_in, "Host"))) {
+        const char *buf2;
+        if((buf2 = apr_table_get(r->headers_in,"X-Forwarded-Host"))) {
+          apr_table_set(req_proxy->headers, "X-Forwarded-Host", apr_psprintf(global_ctx->pool,"%s, %s",buf2,buf));
+        } else {
+          apr_table_set(req_proxy->headers, "X-Forwarded-Host", buf);
+        }
+      }
+
+      if ((buf = apr_table_get(r->headers_in, "X-Forwarded-Server"))) {
+        apr_table_set(req_proxy->headers, "X-Forwarded-Server", apr_psprintf(global_ctx->pool, "%s, %s", buf, r->server->server_hostname));
+      } else {
+        apr_table_set(req_proxy->headers, "X-Forwarded-Server", r->server->server_hostname);
+      }
+    }
     http_response = mapcache_core_proxy_request(global_ctx, req_proxy);
   } else if( request->type == MAPCACHE_REQUEST_GET_MAP) {
     mapcache_request_get_map *req_map = (mapcache_request_get_map*)request;
@@ -389,35 +546,35 @@ static int mapcache_alias_matches(const char *uri, const char *alias_fakename)
   return urip - uri;
 }
 
-static int mapcache_hook_intercept(request_rec *r)
+static int mapcache_hook_fixups(request_rec *r)
 {
-  mapcache_server_cfg *sconfig = ap_get_module_config(r->server->module_config, &mapcache_module);
-  apr_hash_index_t *entry;
+  int i;
+  //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "running mapcache fixup on %s (handler:%s,filename:%s)",r->uri,r->handler,r->filename);
+  if(!r->handler) {
+    mapcache_server_cfg *sconfig = ap_get_module_config(r->server->module_config, &mapcache_module);
+    mapcache_alias_entry *alias_entry;
 
-  if (!sconfig->aliases)
-    return DECLINED;
+    if (!sconfig || !sconfig->aliases)
+      return DECLINED;
 
-  if (r->uri[0] != '/' && r->uri[0])
-    return DECLINED;
+    if (r->uri[0] != '/' && r->uri[0])
+      return DECLINED;
 
-  entry = apr_hash_first(r->pool,sconfig->aliases);
+    /* loop through the entries to find one where the alias matches */
+    for(i=0; i<sconfig->aliases->nelts; i++) {
+      int l;
+      alias_entry = APR_ARRAY_IDX(sconfig->aliases,i,mapcache_alias_entry*);
+      //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "cheking mapcache alias %s against %s",r->uri,alias_entry->endpoint);
 
-  /* loop through the entries to find one where the alias matches */
-  while (entry) {
-    int l = 0;
-    const char *alias;
-    apr_ssize_t aliaslen;
-    mapcache_cfg *c;
-    apr_hash_this(entry,(const void**)&alias,&aliaslen,(void**)&c);
-
-    if((l=mapcache_alias_matches(r->uri, c->endpoint))>0) {
-      r->handler = "mapcache";
-      r->filename = c->configFile;
-      r->path_info = &(r->uri[l]);
-      return OK;
+      if((l=mapcache_alias_matches(r->uri, alias_entry->endpoint))>0) {
+        r->handler = apr_pstrdup(r->pool,"mapcache");
+        apr_table_set(r->notes,"mapcache_alias_entry",alias_entry->endpoint);
+        r->path_info = &(r->uri[l]);
+        //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "setting config %s for alias %s",alias_entry->configfile,alias_entry->endpoint);
+        //ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "using pathinfo %s from uri %s",r->path_info,r->uri);
+        return OK;
+      }
     }
-
-    entry = apr_hash_next(entry);
   }
 
   return DECLINED;
@@ -426,21 +583,17 @@ static int mapcache_hook_intercept(request_rec *r)
 
 static void mod_mapcache_register_hooks(apr_pool_t *p)
 {
-  static const char * const p1[] = { "mod_alias.c", "mod_rewrite.c", NULL };
-  static const char * const n1[]= { "mod_userdir.c",
-                                    "mod_vhost_alias.c", NULL
-                                  };
   ap_hook_child_init(mod_mapcache_child_init, NULL, NULL, APR_HOOK_MIDDLE);
   ap_hook_post_config(mod_mapcache_post_config, NULL, NULL, APR_HOOK_MIDDLE);
   ap_hook_handler(mod_mapcache_request_handler, NULL, NULL, APR_HOOK_MIDDLE);
-  ap_hook_translate_name(mapcache_hook_intercept, p1, n1, APR_HOOK_MIDDLE);
+  ap_hook_fixups(mapcache_hook_fixups, NULL, NULL, APR_HOOK_MIDDLE);
 
 }
 
 static void* mod_mapcache_create_server_conf(apr_pool_t *pool, server_rec *s)
 {
   mapcache_server_cfg *cfg = apr_pcalloc(pool, sizeof(mapcache_server_cfg));
-  cfg->aliases = NULL;
+  cfg->aliases = apr_array_make(pool,1,sizeof(mapcache_alias_entry*));
   return cfg;
 }
 
@@ -451,38 +604,69 @@ static void *mod_mapcache_merge_server_conf(apr_pool_t *p, void *base_, void *vh
   mapcache_server_cfg *vhost = (mapcache_server_cfg*)vhost_;
   mapcache_server_cfg *cfg = apr_pcalloc(p,sizeof(mapcache_server_cfg));
 
-  if (base->aliases && vhost->aliases) {
-    cfg->aliases = apr_hash_overlay(p, vhost->aliases,base->aliases);
-  } else if (vhost->aliases) {
-    cfg->aliases = apr_hash_copy(p,vhost->aliases);
-  } else if (base->aliases) {
-    cfg->aliases = apr_hash_copy(p,base->aliases);
+  cfg->aliases = apr_array_append(p, vhost->aliases,base->aliases);
+
+#if 0
+  {
+    mapcache_alias_entry *e;
+    int i;
+    fprintf(stderr,"#### merge called ####\n");
+    for(i=0;i<base->aliases->nelts;i++) {
+      e = APR_ARRAY_IDX(base->aliases,i,mapcache_alias_entry*);
+      fprintf(stderr,"merge base: have alias %s on %s\n",e->configfile,e->endpoint);
+    }
+    for(i=0;i<vhost->aliases->nelts;i++) {
+      e = APR_ARRAY_IDX(vhost->aliases,i,mapcache_alias_entry*);
+      fprintf(stderr,"merge vhosts: have alias %s on %s\n",e->configfile,e->endpoint);
+    }
+    for(i=0;i<cfg->aliases->nelts;i++) {
+      e = APR_ARRAY_IDX(cfg->aliases,i,mapcache_alias_entry*);
+      fprintf(stderr,"merge result: have alias %s on %s\n",e->configfile,e->endpoint);
+    }
   }
-  return vhost;
+#endif
+  return cfg;
 }
 
 static const char* mapcache_add_alias(cmd_parms *cmd, void *cfg, const char *alias, const char* configfile)
 {
-  mapcache_server_cfg *sconfig = ap_get_module_config(cmd->server->module_config, &mapcache_module);
-  mapcache_cfg *config = mapcache_configuration_create(cmd->pool);
-  mapcache_context *ctx = (mapcache_context*)apache_server_context_create(cmd->server,cmd->pool);
-  char *msg = NULL;
-  config->configFile = apr_pstrdup(cmd->pool,configfile);
-  config->endpoint = alias;
-  mapcache_configuration_parse(ctx,configfile,config,0);
+  mapcache_server_cfg *sconfig;
+  mapcache_alias_entry *alias_entry;
+  mapcache_context *ctx;
+  unsigned forbidden = NOT_IN_DIRECTORY|NOT_IN_FILES;
+  const char *err;
+
+#if (AP_SERVER_MAJORVERSION_NUMBER > 2) || (AP_SERVER_MINORVERSION_NUMBER >= 4)
+  forbidden |= NOT_IN_HTACCESS;
+#endif
+
+  err = ap_check_cmd_context(cmd, forbidden);
+  if (err) {
+    return err;
+  }
+
+  sconfig = ap_get_module_config(cmd->server->module_config, &mapcache_module);
+  if(!sconfig || !sconfig->aliases)
+    return "no mapcache module config, server bug?";
+
+  alias_entry = apr_pcalloc(cmd->pool,sizeof(mapcache_alias_entry));
+  ctx = (mapcache_context*)apache_server_context_create(cmd->server,cmd->pool);
+
+  alias_entry->cfg = mapcache_configuration_create(cmd->pool);
+  alias_entry->configfile = apr_pstrdup(cmd->pool,configfile);
+  alias_entry->endpoint = apr_pstrdup(cmd->pool,alias);
+  mapcache_configuration_parse(ctx,alias_entry->configfile,alias_entry->cfg,0);
   if(GC_HAS_ERROR(ctx)) {
     return ctx->get_error_message(ctx);
   }
-  mapcache_configuration_post_config(ctx, config);
+  mapcache_configuration_post_config(ctx, alias_entry->cfg);
   if(GC_HAS_ERROR(ctx)) {
     return ctx->get_error_message(ctx);
   }
-  ap_log_error(APLOG_MARK, APLOG_INFO, 0, cmd->server, "loaded mapcache configuration file from %s on alias %s", config->configFile, alias);
-  if(!sconfig->aliases) {
-    sconfig->aliases = apr_hash_make(cmd->pool);
-  }
-  apr_hash_set(sconfig->aliases,configfile,APR_HASH_KEY_STRING,config);
-  return msg;
+  APR_ARRAY_PUSH(sconfig->aliases,mapcache_alias_entry*) = alias_entry;
+  ap_log_error(APLOG_MARK, APLOG_INFO, 0, cmd->server, "loaded mapcache configuration file from %s on endpoint %s", alias_entry->configfile, alias_entry->endpoint);
+
+  return NULL;
 }
 
 
