@@ -39,6 +39,7 @@
 
 #include <apr_tables.h>
 #include <apr_hash.h>
+#include <apr_reslist.h>
 
 #include "util.h"
 #include "ezxml.h"
@@ -66,6 +67,14 @@
 #include <apr_memcache.h>
 #endif
 
+#ifdef USE_COUCHBASE
+#include <libcouchbase/couchbase.h>
+#endif
+
+#ifdef USE_RIAK
+#include <riack.h>
+#endif
+
 #define MAPCACHE_SUCCESS 0
 #define MAPCACHE_FAILURE 1
 #define MAPCACHE_TRUE 1
@@ -75,11 +84,17 @@
 #define MAPCACHE_TILESET_WRONG_EXTENT 4
 #define MAPCACHE_CACHE_MISS 5
 #define MAPCACHE_FILE_LOCKED 6
+#define MAPCACHE_CACHE_RELOAD 7
+
+#define MAPCACHE_MAX_NUM_TILES 1000
 
 #define MAPCACHE_USERAGENT "mod-mapcache/"MAPCACHE_VERSION
 
-#define MAPCACHE_LOCKFILE_PREFIX "_gc_lock"
-
+#if defined(_WIN32) && !defined(__CYGWIN__)
+#  define MS_DLL_EXPORT     __declspec(dllexport)
+#else
+#define  MS_DLL_EXPORT
+#endif
 
 
 typedef struct mapcache_image_format mapcache_image_format;
@@ -103,11 +118,19 @@ typedef struct mapcache_source_wms mapcache_source_wms;
 typedef struct mapcache_source_gdal mapcache_source_gdal;
 #endif
 typedef struct mapcache_cache_disk mapcache_cache_disk;
+typedef struct mapcache_cache_composite mapcache_cache_composite;
+typedef struct mapcache_cache_fallback mapcache_cache_fallback;
+typedef struct mapcache_cache_multitier mapcache_cache_multitier;
+typedef struct mapcache_cache_rest mapcache_cache_rest;
+typedef struct mapcache_cache_s3 mapcache_cache_s3;
+typedef struct mapcache_cache_azure mapcache_cache_azure;
+typedef struct mapcache_cache_google mapcache_cache_google;
 #ifdef USE_TIFF
 typedef struct mapcache_cache_tiff mapcache_cache_tiff;
 #endif
 typedef struct mapcache_http mapcache_http;
 typedef struct mapcache_request mapcache_request;
+typedef struct mapcache_request_image mapcache_request_image;
 typedef struct mapcache_request_proxy mapcache_request_proxy;
 typedef struct mapcache_request_get_capabilities mapcache_request_get_capabilities;
 typedef struct mapcache_request_get_capabilities_demo mapcache_request_get_capabilities_demo;
@@ -139,9 +162,12 @@ typedef struct mapcache_dimension_time mapcache_dimension_time;
 typedef struct mapcache_timedimension mapcache_timedimension;
 typedef struct mapcache_dimension_intervals mapcache_dimension_intervals;
 typedef struct mapcache_dimension_values mapcache_dimension_values;
+typedef struct mapcache_dimension_sqlite mapcache_dimension_sqlite;
 typedef struct mapcache_dimension_regex mapcache_dimension_regex;
 typedef struct mapcache_extent mapcache_extent;
 typedef struct mapcache_extent_i mapcache_extent_i;
+typedef struct mapcache_connection_pool mapcache_connection_pool;
+typedef struct mapcache_locker mapcache_locker;
 
 /** \defgroup utility Utility */
 /** @{ */
@@ -198,6 +224,18 @@ struct mapcache_context {
    */
   void (*clear_errors)(mapcache_context * ctx);
 
+  /**
+   * \brief clear current error and store it in mapcache_error
+   * \memberof mapcache_context
+   */
+  void (*pop_errors)(mapcache_context * ctx, void **error);
+
+  /**
+   * \brief restore error status from mapcache_error
+   * \memberof mapcache_context
+   */
+  void (*push_errors)(mapcache_context * ctx, void *error);
+
 
   /**
    * \brief log a message
@@ -208,7 +246,7 @@ struct mapcache_context {
   const char* (*get_instance_id)(mapcache_context * ctx);
   mapcache_context* (*clone)(mapcache_context *ctx);
   apr_pool_t *pool;
-  apr_pool_t *process_pool;
+  mapcache_connection_pool *connection_pool;
   void *threadlock;
   char *_contenttype;
   char *_errmsg;
@@ -216,12 +254,14 @@ struct mapcache_context {
   mapcache_cfg *config;
   mapcache_service *service;
   apr_table_t *exceptions;
+  int supports_redirects;
+  apr_table_t *headers_in;
 };
 
-void mapcache_context_init(mapcache_context *ctx);
-void mapcache_context_copy(mapcache_context *src, mapcache_context *dst);
+MS_DLL_EXPORT void mapcache_context_init(mapcache_context *ctx);
+MS_DLL_EXPORT void mapcache_context_copy(mapcache_context *src, mapcache_context *dst);
 
-#define GC_CHECK_ERROR_RETURN(ctx) (if(((mapcache_context*)ctx)->_errcode) return MAPCACHE_FAILURE;)
+#define GC_CHECK_ERROR_RETURN(ctx) if(((mapcache_context*)ctx)->_errcode) return MAPCACHE_FAILURE;
 #define GC_CHECK_ERROR(ctx) if(((mapcache_context*)ctx)->_errcode) return;
 #define GC_HAS_ERROR(ctx) (((mapcache_context*)ctx)->_errcode > 0)
 
@@ -299,6 +339,8 @@ mapcache_http* mapcache_http_clone(mapcache_context *ctx, mapcache_http *orig);
 struct mapcache_http {
   char *url; /**< the base url to request */
   apr_table_t *headers; /**< additional headers to add to the http request, eg, Referer */
+  char *post_body;
+  size_t post_len;
   int connection_timeout;
   int timeout;
   /* TODO: authentication */
@@ -355,7 +397,8 @@ struct mapcache_source_gdal {
 
 /** @{ */
 typedef enum {
-  MAPCACHE_CACHE_DISK
+  MAPCACHE_CACHE_DISK,
+  MAPCACHE_CACHE_REST
 #ifdef USE_MEMCACHE
   ,MAPCACHE_CACHE_MEMCACHE
 #endif
@@ -370,6 +413,13 @@ typedef enum {
 #endif
 #ifdef USE_TIFF
   ,MAPCACHE_CACHE_TIFF
+#endif
+  ,MAPCACHE_CACHE_COMPOSITE
+#ifdef USE_COUCHBASE
+  ,MAPCACHE_CACHE_COUCHBASE
+#endif
+#ifdef USE_RIAK
+  ,MAPCACHE_CACHE_RIAK
 #endif
 } mapcache_cache_type;
 
@@ -388,23 +438,23 @@ struct mapcache_cache {
    * \returns MAPCACHE_CACHE_MISS if the file does not exist on the disk
    * \memberof mapcache_cache
    */
-  int (*tile_get)(mapcache_context *ctx, mapcache_tile * tile);
+  int (*tile_get)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
   /**
    * delete tile from cache
    *
    * \memberof mapcache_cache
    */
-  void (*tile_delete)(mapcache_context *ctx, mapcache_tile * tile);
+  void (*tile_delete)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
-  int (*tile_exists)(mapcache_context *ctx, mapcache_tile * tile);
+  int (*tile_exists)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
 
   /**
    * set tile content to cache
    * \memberof mapcache_cache
    */
-  void (*tile_set)(mapcache_context *ctx, mapcache_tile * tile);
-  void (*tile_multi_set)(mapcache_context *ctx, mapcache_tile *tiles, int ntiles);
+  void (*tile_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile * tile);
+  void (*tile_multi_set)(mapcache_context *ctx, mapcache_cache *cache, mapcache_tile *tiles, int ntiles);
 
   void (*configuration_parse_xml)(mapcache_context *ctx, ezxml_t xml, mapcache_cache * cache, mapcache_cfg *config);
   void (*configuration_post_config)(mapcache_context *ctx, mapcache_cache * cache, mapcache_cfg *config);
@@ -425,7 +475,109 @@ struct mapcache_cache_disk {
    * Set filename for a given tile
    * \memberof mapcache_cache_disk
    */
-  void (*tile_key)(mapcache_context *ctx, mapcache_tile *tile, char **path);
+  void (*tile_key)(mapcache_context *ctx, mapcache_cache_disk *cache, mapcache_tile *tile, char **path);
+};
+
+typedef struct mapcache_cache_composite_cache_link mapcache_cache_composite_cache_link;
+struct mapcache_cache_composite_cache_link {
+  mapcache_cache *cache;
+  int minzoom;
+  int maxzoom;
+  apr_array_header_t *grids;
+  apr_array_header_t *dimensions; //TODO
+};
+
+struct mapcache_cache_composite {
+  mapcache_cache cache;
+  apr_array_header_t *cache_links;
+};
+
+struct mapcache_cache_fallback {
+  mapcache_cache cache;
+  apr_array_header_t *caches;
+};
+
+struct mapcache_cache_multitier {
+  mapcache_cache cache;
+  apr_array_header_t *caches;
+};
+
+
+typedef enum {
+  MAPCACHE_REST_METHOD_GET,
+  MAPCACHE_REST_METHOD_HEAD,
+  MAPCACHE_REST_METHOD_PUT,
+  MAPCACHE_REST_METHOD_POST,
+  MAPCACHE_REST_METHOD_DELETE
+} mapcache_rest_method;
+
+typedef enum {
+  MAPCACHE_REST_PROVIDER_NONE,
+  MAPCACHE_REST_PROVIDER_S3,
+  MAPCACHE_REST_PROVIDER_AZURE,
+  MAPCACHE_REST_PROVIDER_GOOGLE
+} mapcache_rest_provider;
+
+void sha256(const unsigned char *message, unsigned int len, unsigned char *digest);
+void hmac_sha256(const unsigned char *message, unsigned int message_len,
+          const unsigned char *key, unsigned int key_size,
+          unsigned char *mac, unsigned mac_size);
+void hmac_sha1(const char *message, unsigned int message_len,
+          const unsigned char *key, unsigned int key_size,
+          void *mac);
+void sha_hex_encode(unsigned char *sha, unsigned int sha_size);
+char *base64_encode(apr_pool_t *pool, const unsigned char *data, size_t input_length);
+
+typedef struct mapcache_rest_operation mapcache_rest_operation;
+struct mapcache_rest_operation {
+  apr_table_t *headers;
+  mapcache_rest_method method;
+  char *tile_url;
+  void (*add_headers)(mapcache_context *ctx, mapcache_cache_rest *pcache, mapcache_tile *tile, char *url, apr_table_t *headers);
+};
+
+typedef struct mapcache_rest_configuration mapcache_rest_configuration;
+struct mapcache_rest_configuration {
+  apr_table_t *common_headers;
+  char *tile_url;
+  mapcache_rest_operation has_tile;
+  mapcache_rest_operation get_tile;
+  mapcache_rest_operation set_tile;
+  mapcache_rest_operation multi_set_tile;
+  mapcache_rest_operation delete_tile;
+  void (*add_headers)(mapcache_context *ctx, mapcache_cache_rest *pcache, mapcache_tile *tile, char *url, apr_table_t *headers);
+};
+
+/**\class mapcache_cache_rest
+ * \brief a mapcache_cache on a 3rd party HTTP Rest API
+ * \implements mapcache_cache
+ */
+struct mapcache_cache_rest {
+  mapcache_cache cache;
+  mapcache_rest_configuration rest;
+  int use_redirects;
+  int retry_count;
+  mapcache_rest_provider provider;
+};
+
+struct mapcache_cache_s3 {
+  mapcache_cache_rest cache;
+  char *id;
+  char *secret;
+  char *region;
+};
+
+struct mapcache_cache_azure {
+  mapcache_cache_rest cache;
+  char *id;
+  char *secret;
+  char *container;
+};
+
+struct mapcache_cache_google {
+  mapcache_cache_rest cache;
+  char *access;
+  char *secret;
 };
 
 #ifdef USE_TIFF
@@ -436,6 +588,7 @@ struct mapcache_cache_tiff {
   int count_x;
   int count_y;
   mapcache_image_format_jpeg *format;
+  mapcache_locker *locker;
 };
 #endif
 
@@ -451,6 +604,8 @@ struct mapcache_cache_sqlite_stmt {
   char *sql;
 };
 
+struct sqlite_conn;
+
 struct mapcache_cache_sqlite {
   mapcache_cache cache;
   char *dbfile;
@@ -460,9 +615,11 @@ struct mapcache_cache_sqlite {
   mapcache_cache_sqlite_stmt set_stmt;
   mapcache_cache_sqlite_stmt delete_stmt;
   apr_table_t *pragmas;
-  void (*bind_stmt)(mapcache_context*ctx, void *stmt, mapcache_tile *tile);
+  void (*bind_stmt)(mapcache_context *ctx, void *stmt, mapcache_cache_sqlite *cache, mapcache_tile *tile);
   int n_prepared_statements;
   int detect_blank;
+  char *x_fmt,*y_fmt,*z_fmt,*inv_x_fmt,*inv_y_fmt,*div_x_fmt,*div_y_fmt,*inv_div_x_fmt,*inv_div_y_fmt;
+  int count_x, count_y;
 };
 
 /**
@@ -499,9 +656,17 @@ typedef struct mapcache_cache_memcache mapcache_cache_memcache;
  * \brief a mapcache_cache on memcached servers
  * \implements mapcache_cache
  */
+
+struct mapcache_cache_memcache_server {
+    char* host;
+    int port;
+};
+
 struct mapcache_cache_memcache {
   mapcache_cache cache;
-  apr_memcache_t *memcache;
+  int nservers;
+  struct mapcache_cache_memcache_server *servers;
+  int detect_blank;
 };
 
 /**
@@ -510,6 +675,49 @@ struct mapcache_cache_memcache {
 mapcache_cache* mapcache_cache_memcache_create(mapcache_context *ctx);
 #endif
 
+#ifdef USE_COUCHBASE
+typedef struct mapcache_cache_couchbase mapcache_cache_couchbase;
+
+/**\class mapcache_cache_couchbase
+ * \brief a mapcache_cache on couchbase servers
+ * \implements mapcache_cache
+ */
+struct mapcache_cache_couchbase {
+   mapcache_cache cache;
+//   apr_reslist_t *connection_pool;
+   char *host;
+   char *username;
+   char *password;
+   char *bucket;
+   mapcache_context *ctx;
+};
+
+/**
+ * \memberof mapcache_cache_couchbase
+ */
+mapcache_cache* mapcache_cache_couchbase_create(mapcache_context *ctx);
+#endif
+
+#ifdef USE_RIAK
+typedef struct mapcache_cache_riak mapcache_cache_riak;
+
+/**\class mapcache_cache_riak
+ * \brief a mapcache_cache for riak servers
+ * \implements mapcache_cache
+ */
+struct mapcache_cache_riak {
+   mapcache_cache cache;
+   char *host;
+   int port;
+   RIACK_STRING bucket;
+};
+
+/**
+ * \memberof mapcache_cache_riak
+ */
+mapcache_cache* mapcache_cache_riak_create(mapcache_context *ctx);
+#endif
+
 /** @} */
 
 
@@ -542,8 +750,13 @@ struct mapcache_request {
   mapcache_service *service;
 };
 
+struct mapcache_request_image {
+    mapcache_request request;
+    mapcache_image_format *format;
+};
+
 struct mapcache_request_get_tile {
-  mapcache_request request;
+  mapcache_request_image image_request;
 
   /**
    * a list of tiles requested by the client
@@ -557,8 +770,7 @@ struct mapcache_request_get_tile {
    * before being returned to the client
    */
   int ntiles;
-  mapcache_image_format *format;
-
+  int allow_redirect;
 };
 
 struct mapcache_http_response {
@@ -594,12 +806,11 @@ struct mapcache_request_get_feature_info {
 };
 
 struct mapcache_request_get_map {
-  mapcache_request request;
+  mapcache_request_image image_request;
   mapcache_map **maps;
   int nmaps;
   mapcache_getmap_strategy getmap_strategy;
   mapcache_resample_mode resample_mode;
-  mapcache_image_format *getmap_format;
 };
 
 struct mapcache_request_get_capabilities {
@@ -647,18 +858,22 @@ struct mapcache_request_get_capabilities_demo {
   mapcache_service *service;
 };
 
-struct mapcache_request_proxy {
-  mapcache_request request;
-  mapcache_http *http;
-  apr_table_t *params;
-  const char *pathinfo;
-};
-
 struct mapcache_forwarding_rule {
   char *name;
   mapcache_http *http;
   apr_array_header_t *match_params;  /* actually those are mapcache_dimensions */
   int append_pathinfo;
+  size_t max_post_len;
+};
+
+struct mapcache_request_proxy {
+  mapcache_request request;
+  mapcache_forwarding_rule *rule;
+  apr_table_t *params;
+  apr_table_t *headers;
+  const char *pathinfo;
+  char *post_buf;
+  size_t post_len;
 };
 
 
@@ -728,6 +943,7 @@ struct mapcache_service_wms {
   mapcache_getmap_strategy getmap_strategy;
   mapcache_resample_mode resample_mode;
   mapcache_image_format *getmap_format;
+  int allow_format_override; /* can the client specify which image format should be returned */
 };
 
 /**\class mapcache_service_kml
@@ -829,7 +1045,7 @@ mapcache_service* mapcache_service_demo_create(mapcache_context *ctx);
 /**
  * \brief return the request that corresponds to the given url
  */
-void mapcache_service_dispatch_request(mapcache_context *ctx,
+MS_DLL_EXPORT void mapcache_service_dispatch_request(mapcache_context *ctx,
                                        mapcache_request **request,
                                        char *pathinfo,
                                        apr_table_t *params,
@@ -926,25 +1142,22 @@ int mapcache_image_blank_color(mapcache_image* image);
  */
 int mapcache_image_has_alpha(mapcache_image *img);
 
+void mapcache_image_fill(mapcache_context *ctx, mapcache_image *image, const unsigned char *fill_color);
+
 /** @} */
 
 
 /** \defgroup http HTTP Request handling*/
 /** @{ */
 void mapcache_http_do_request(mapcache_context *ctx, mapcache_http *req, mapcache_buffer *data, apr_table_t *headers, long *http_code);
-void mapcache_http_do_request_with_params(mapcache_context *ctx, mapcache_http *req, apr_table_t *params,
-    mapcache_buffer *data, apr_table_t *headers, long *http_code);
 char* mapcache_http_build_url(mapcache_context *ctx, char *base, apr_table_t *params);
-apr_table_t *mapcache_http_parse_param_string(mapcache_context *ctx, char *args);
+MS_DLL_EXPORT apr_table_t *mapcache_http_parse_param_string(mapcache_context *ctx, char *args);
 /** @} */
 
 /** \defgroup configuration Configuration*/
 
 /** @{ */
 
-struct mapcache_server_cfg {
-  apr_hash_t *aliases; /**< list of mapcache configurations aliased to a server uri */
-};
 
 
 
@@ -954,12 +1167,72 @@ typedef enum {
   MAPCACHE_MODE_MIRROR_SPLIT
 } mapcache_mode;
 
+typedef enum {
+  MAPCACHE_LOCKER_DISK,
+  MAPCACHE_LOCKER_MEMCACHE,
+  MAPCACHE_LOCKER_FALLBACK
+} mapcache_lock_mode;
+
+typedef enum {
+    MAPCACHE_LOCK_AQUIRED,
+    MAPCACHE_LOCK_LOCKED,
+    MAPCACHE_LOCK_NOENT
+} mapcache_lock_result;
+
+
+struct mapcache_locker{
+  mapcache_lock_result (*aquire_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void **lock);
+  mapcache_lock_result (*ping_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void *lock);
+  void (*release_lock)(mapcache_context *ctx, mapcache_locker *self, char *resource, void *lock);
+
+  void (*parse_xml)(mapcache_context *ctx, mapcache_locker *self, ezxml_t node);
+  mapcache_lock_mode type;
+  double timeout;
+  double retry_interval; /* time to wait before checking again on a lock, in seconds */
+};
+
+typedef struct {
+  mapcache_locker locker;
+
+  /**
+   * directory where lock files will be placed.
+   * Must be readable and writable by the apache user.
+   * Must be placed on a network mounted shared directory if multiple mapcache instances
+   * need to be synchronized
+   */
+  const char *dir;
+} mapcache_locker_disk;
+
+mapcache_locker* mapcache_locker_disk_create(mapcache_context *ctx);
+
+#ifdef USE_MEMCACHE
+typedef struct {
+  char *host;
+  int port;
+} mapcache_locker_memcache_server;
+
+typedef struct {
+  mapcache_locker locker;
+  int nservers;
+  mapcache_locker_memcache_server *servers;
+} mapcache_locker_memcache;
+
+mapcache_locker* mapcache_locker_memcache_create(mapcache_context *ctx);
+#endif
+
+typedef struct {
+    mapcache_locker locker;
+    apr_array_header_t *lockers;
+} mapcache_locker_fallback;
+
+mapcache_locker* mapcache_locker_fallback_create(mapcache_context *ctx);
+
+void mapcache_config_parse_locker(mapcache_context *ctx, ezxml_t node, mapcache_locker **locker);
+
 /**
  * a configuration that will be served
  */
 struct mapcache_cfg {
-  char *configFile; /**< the filename from which this configuration was loaded */
-
   /**
    * a list of services that will be responded to
    */
@@ -1011,26 +1284,10 @@ struct mapcache_cfg {
 
   apr_table_t *metadata;
 
-  /**
-   * directory where lock files will be placed.
-   * Must be readable and writable by the apache user.
-   * Must be placed on a network mounted shared directory if multiple mapcache instances
-   * need to be synchronized
-   */
-  const char *lockdir;
-
-  /**
-   * time in nanoseconds to wait before rechecking for lockfile presence
-   */
-  apr_interval_time_t lock_retry_interval; /* time in nanoseconds to wait before rechecking for lockfile presence */
+  mapcache_locker *locker;
 
   int threaded_fetching;
 
-  /**
-   * the uri where the base of the service is mapped
-   */
-  const char *endpoint;
-
   /* for fastcgi only */
   int autoreload; /* should the modification time of the config file be recorded
                        and the file be reparsed if it is modified. */
@@ -1051,14 +1308,14 @@ struct mapcache_cfg {
  * @param pool
  * @return
  */
-void mapcache_configuration_parse(mapcache_context *ctx, const char *filename, mapcache_cfg *config, int cgi);
-void mapcache_configuration_post_config(mapcache_context *ctx, mapcache_cfg *config);
+MS_DLL_EXPORT void mapcache_configuration_parse(mapcache_context *ctx, const char *filename, mapcache_cfg *config, int cgi);
+MS_DLL_EXPORT void mapcache_configuration_post_config(mapcache_context *ctx, mapcache_cfg *config);
 void mapcache_configuration_parse_xml(mapcache_context *ctx, const char *filename, mapcache_cfg *config);
-mapcache_cfg* mapcache_configuration_create(apr_pool_t *pool);
+MS_DLL_EXPORT mapcache_cfg* mapcache_configuration_create(apr_pool_t *pool);
 mapcache_source* mapcache_configuration_get_source(mapcache_cfg *config, const char *key);
-mapcache_cache* mapcache_configuration_get_cache(mapcache_cfg *config, const char *key);
+MS_DLL_EXPORT mapcache_cache* mapcache_configuration_get_cache(mapcache_cfg *config, const char *key);
 mapcache_grid *mapcache_configuration_get_grid(mapcache_cfg *config, const char *key);
-mapcache_tileset* mapcache_configuration_get_tileset(mapcache_cfg *config, const char *key);
+MS_DLL_EXPORT mapcache_tileset* mapcache_configuration_get_tileset(mapcache_cfg *config, const char *key);
 mapcache_image_format *mapcache_configuration_get_image_format(mapcache_cfg *config, const char *key);
 void mapcache_configuration_add_image_format(mapcache_cfg *config, mapcache_image_format *format, const char * key);
 void mapcache_configuration_add_source(mapcache_cfg *config, mapcache_source *source, const char * key);
@@ -1096,6 +1353,14 @@ mapcache_source* mapcache_source_dummy_create(mapcache_context *ctx);
  */
 mapcache_cache* mapcache_cache_disk_create(mapcache_context *ctx);
 
+/**
+ * \memberof mapcache_cache_rest
+ */
+mapcache_cache* mapcache_cache_rest_create(mapcache_context *ctx);
+mapcache_cache* mapcache_cache_s3_create(mapcache_context *ctx);
+mapcache_cache* mapcache_cache_azure_create(mapcache_context *ctx);
+mapcache_cache* mapcache_cache_google_create(mapcache_context *ctx);
+
 #ifdef USE_TIFF
 /**
  * \memberof mapcache_cache_tiff
@@ -1103,6 +1368,10 @@ mapcache_cache* mapcache_cache_disk_create(mapcache_context *ctx);
 mapcache_cache* mapcache_cache_tiff_create(mapcache_context *ctx);
 #endif
 
+mapcache_cache* mapcache_cache_composite_create(mapcache_context *ctx);
+mapcache_cache* mapcache_cache_fallback_create(mapcache_context *ctx);
+mapcache_cache* mapcache_cache_multitier_create(mapcache_context *ctx);
+
 
 /** \defgroup tileset Tilesets*/
 /** @{ */
@@ -1125,6 +1394,8 @@ struct mapcache_tile {
    * \sa mapcache_image_format
    */
   mapcache_buffer *encoded_data;
+  char *redirect;
+  int allow_redirect;
   mapcache_image *raw_image;
   apr_time_t mtime; /**< last modification time */
   int expires; /**< time in seconds after which the tile should be rechecked for validity */
@@ -1168,7 +1439,7 @@ typedef enum {
   MAPCACHE_GRID_ORIGIN_BOTTOM_LEFT,
   MAPCACHE_GRID_ORIGIN_TOP_LEFT,
   MAPCACHE_GRID_ORIGIN_BOTTOM_RIGHT,
-  MAPCACHE_GRID_ORIGIN_TOP_RIGHT,
+  MAPCACHE_GRID_ORIGIN_TOP_RIGHT
 } mapcache_grid_origin;
 
 struct mapcache_grid {
@@ -1200,7 +1471,7 @@ struct mapcache_grid_link {
   mapcache_extent *restricted_extent;
   mapcache_extent_i *grid_limits;
   int minz,maxz;
-  
+
   /**
    * tiles above this zoom level will not be stored to the cache, but will be
    * dynamically generated (either by reconstructing from lower level tiles, or
@@ -1209,6 +1480,8 @@ struct mapcache_grid_link {
 
   int max_cached_zoom;
   mapcache_outofzoom_strategy outofzoom_strategy;
+
+  apr_array_header_t *intermediate_grids;
 };
 
 /**\class mapcache_tileset
@@ -1262,7 +1535,7 @@ struct mapcache_tileset {
   /**
    * the cache in which the tiles should be stored
    */
-  mapcache_cache *cache;
+  mapcache_cache *_cache;
 
   /**
    * the source from which tiles should be requested
@@ -1301,7 +1574,8 @@ void mapcache_tileset_get_map_tiles(mapcache_context *ctx, mapcache_tileset *til
                                     mapcache_grid_link *grid_link,
                                     mapcache_extent *bbox, int width, int height,
                                     int *ntiles,
-                                    mapcache_tile ***tiles);
+                                    mapcache_tile ***tiles,
+                                    mapcache_grid_link **effectively_used_grid_link);
 
 mapcache_image* mapcache_tileset_assemble_map_tiles(mapcache_context *ctx, mapcache_tileset *tileset,
     mapcache_grid_link *grid_link,
@@ -1338,14 +1612,14 @@ void mapcache_tileset_tile_validate(mapcache_context *ctx, mapcache_tile *tile);
  */
 void mapcache_tileset_get_level(mapcache_context *ctx, mapcache_tileset *tileset, double *resolution, int *level);
 
-void mapcache_grid_get_closest_level(mapcache_context *ctx, mapcache_grid_link *grid, double resolution, int *level);
-void mapcache_tileset_tile_get(mapcache_context *ctx, mapcache_tile *tile);
+mapcache_grid_link* mapcache_grid_get_closest_wms_level(mapcache_context *ctx, mapcache_grid_link *grid, double resolution, int *level);
+MS_DLL_EXPORT void mapcache_tileset_tile_get(mapcache_context *ctx, mapcache_tile *tile);
 
 /**
  * \brief delete tile from cache
  * @param whole_metatile delete all the other tiles from the metatile to
  */
-void mapcache_tileset_tile_delete(mapcache_context *ctx, mapcache_tile *tile, int whole_metatile);
+MS_DLL_EXPORT void mapcache_tileset_tile_delete(mapcache_context *ctx, mapcache_tile *tile, int whole_metatile);
 
 int mapcache_grid_is_bbox_aligned(mapcache_context *ctx, mapcache_grid *grid, mapcache_extent *bbox);
 
@@ -1356,7 +1630,7 @@ int mapcache_grid_is_bbox_aligned(mapcache_context *ctx, mapcache_grid *grid, ma
  * @param pool
  * @return
  */
-mapcache_tile* mapcache_tileset_tile_create(apr_pool_t *pool, mapcache_tileset *tileset, mapcache_grid_link *grid_link);
+MS_DLL_EXPORT mapcache_tile* mapcache_tileset_tile_create(apr_pool_t *pool, mapcache_tileset *tileset, mapcache_grid_link *grid_link);
 
 mapcache_tile* mapcache_tileset_tile_clone(apr_pool_t *pool, mapcache_tile *src);
 
@@ -1389,27 +1663,27 @@ void mapcache_tileset_configuration_check(mapcache_context *ctx, mapcache_tilese
 void mapcache_tileset_add_watermark(mapcache_context *ctx, mapcache_tileset *tileset, const char *filename);
 
 
-int mapcache_lock_or_wait_for_resource(mapcache_context *ctx, char *resource);
-void mapcache_unlock_resource(mapcache_context *ctx, char *resource);
+MS_DLL_EXPORT int mapcache_lock_or_wait_for_resource(mapcache_context *ctx, mapcache_locker *locker, char *resource, void **lock);
+MS_DLL_EXPORT void mapcache_unlock_resource(mapcache_context *ctx, mapcache_locker *locker, char *resource, void *lock);
 
-mapcache_metatile* mapcache_tileset_metatile_get(mapcache_context *ctx, mapcache_tile *tile);
-void mapcache_tileset_render_metatile(mapcache_context *ctx, mapcache_metatile *mt);
-char* mapcache_tileset_metatile_resource_key(mapcache_context *ctx, mapcache_metatile *mt);
+MS_DLL_EXPORT mapcache_metatile* mapcache_tileset_metatile_get(mapcache_context *ctx, mapcache_tile *tile);
+MS_DLL_EXPORT void mapcache_tileset_render_metatile(mapcache_context *ctx, mapcache_metatile *mt);
+MS_DLL_EXPORT char* mapcache_tileset_metatile_resource_key(mapcache_context *ctx, mapcache_metatile *mt);
 
 
 /** @} */
 
 
 
-mapcache_http_response* mapcache_core_get_capabilities(mapcache_context *ctx, mapcache_service *service, mapcache_request_get_capabilities *req_caps, char *url, char *path_info, mapcache_cfg *config);
-mapcache_http_response* mapcache_core_get_tile(mapcache_context *ctx, mapcache_request_get_tile *req_tile);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_get_capabilities(mapcache_context *ctx, mapcache_service *service, mapcache_request_get_capabilities *req_caps, char *url, char *path_info, mapcache_cfg *config);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_get_tile(mapcache_context *ctx, mapcache_request_get_tile *req_tile);
 
-mapcache_http_response* mapcache_core_get_map(mapcache_context *ctx, mapcache_request_get_map *req_map);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_get_map(mapcache_context *ctx, mapcache_request_get_map *req_map);
 
-mapcache_http_response* mapcache_core_get_featureinfo(mapcache_context *ctx, mapcache_request_get_feature_info *req_fi);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_get_featureinfo(mapcache_context *ctx, mapcache_request_get_feature_info *req_fi);
 
-mapcache_http_response* mapcache_core_proxy_request(mapcache_context *ctx, mapcache_request_proxy *req_proxy);
-mapcache_http_response* mapcache_core_respond_to_error(mapcache_context *ctx);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_proxy_request(mapcache_context *ctx, mapcache_request_proxy *req_proxy);
+MS_DLL_EXPORT mapcache_http_response* mapcache_core_respond_to_error(mapcache_context *ctx);
 
 
 /* in grid.c */
@@ -1418,7 +1692,7 @@ mapcache_grid* mapcache_grid_create(apr_pool_t *pool);
 const char* mapcache_grid_get_crs(mapcache_context *ctx, mapcache_grid *grid);
 const char* mapcache_grid_get_srs(mapcache_context *ctx, mapcache_grid *grid);
 
-void mapcache_grid_get_extent(mapcache_context *ctx, mapcache_grid *grid,
+MS_DLL_EXPORT void mapcache_grid_get_extent(mapcache_context *ctx, mapcache_grid *grid,
                               int x, int y, int z, mapcache_extent *bbox);
 /**
  * \brief compute x y value for given lon/lat (dx/dy) and given zoomlevel
@@ -1430,7 +1704,7 @@ void mapcache_grid_get_extent(mapcache_context *ctx, mapcache_grid *grid,
  * @param x
  * @param y
  */
-void mapcache_grid_get_xy(mapcache_context *ctx, mapcache_grid *grid, double dx, double dy, int z, int *x, int *y);
+MS_DLL_EXPORT void mapcache_grid_get_xy(mapcache_context *ctx, mapcache_grid *grid, double dx, double dy, int z, int *x, int *y);
 
 double mapcache_grid_get_resolution(mapcache_extent *bbox, int sx, int sy);
 double mapcache_grid_get_horizontal_resolution(mapcache_extent *bbox, int width);
@@ -1450,12 +1724,12 @@ int mapcache_grid_get_level(mapcache_context *ctx, mapcache_grid *grid, double *
  * \param extent
  * \param tolerance the number of tiles around the given extent that can be requested without returning an error.
  */
-void mapcache_grid_compute_limits(const mapcache_grid *grid, const mapcache_extent *extent, mapcache_extent_i *limits, int tolerance);
+MS_DLL_EXPORT void mapcache_grid_compute_limits(const mapcache_grid *grid, const mapcache_extent *extent, mapcache_extent_i *limits, int tolerance);
 
 /* in util.c */
-int mapcache_util_extract_int_list(mapcache_context *ctx, const char* args, const char *sep, int **numbers,
+MS_DLL_EXPORT int mapcache_util_extract_int_list(mapcache_context *ctx, const char* args, const char *sep, int **numbers,
                                    int *numbers_count);
-int mapcache_util_extract_double_list(mapcache_context *ctx, const char* args, const char *sep, double **numbers,
+MS_DLL_EXPORT int mapcache_util_extract_double_list(mapcache_context *ctx, const char* args, const char *sep, double **numbers,
                                       int *numbers_count);
 char *mapcache_util_str_replace(apr_pool_t *pool, const char *string, const char *substr,
                                 const char *replacement );
@@ -1474,6 +1748,7 @@ char* mapcache_util_get_tile_dimkey(mapcache_context *ctx, mapcache_tile *tile,
 
 char* mapcache_util_get_tile_key(mapcache_context *ctx, mapcache_tile *tile, char *stemplate,
                                  char* sanitized_chars, char *sanitize_to);
+void mapcache_make_parent_dirs(mapcache_context *ctx, char *filename);
 
 /**\defgroup imageio Image IO */
 /** @{ */
@@ -1484,6 +1759,7 @@ char* mapcache_util_get_tile_key(mapcache_context *ctx, mapcache_tile *tile, cha
 typedef enum {
   MAPCACHE_COMPRESSION_BEST, /**< best but slowest compression*/
   MAPCACHE_COMPRESSION_FAST, /**< fast compression*/
+  MAPCACHE_COMPRESSION_DISABLE, /**< no compression*/
   MAPCACHE_COMPRESSION_DEFAULT /**< default compression*/
 } mapcache_compression_type;
 
@@ -1535,7 +1811,7 @@ struct mapcache_image_format_mixed {
   mapcache_image_format *opaque;
 };
 
-mapcache_buffer* mapcache_empty_png_decode(mapcache_context *ctx, const unsigned char *hex_color, int *is_empty);
+mapcache_buffer* mapcache_empty_png_decode(mapcache_context *ctx, int width, int height, const unsigned char *hex_color, int *is_empty);
 
 
 mapcache_image_format* mapcache_imageio_create_mixed_format(apr_pool_t *pool,
@@ -1658,7 +1934,8 @@ typedef enum {
   MAPCACHE_DIMENSION_VALUES,
   MAPCACHE_DIMENSION_REGEX,
   MAPCACHE_DIMENSION_INTERVALS,
-  MAPCACHE_DIMENSION_TIME
+  MAPCACHE_DIMENSION_TIME,
+  MAPCACHE_DIMENSION_SQLITE
 } mapcache_dimension_type;
 
 struct mapcache_dimension {
@@ -1667,6 +1944,7 @@ struct mapcache_dimension {
   char *unit;
   apr_table_t *metadata;
   char *default_value;
+  int skip_validation;
 
   /**
    * \brief validate the given value
@@ -1683,7 +1961,7 @@ struct mapcache_dimension {
    *
    * \returns a list of character strings that will be included in the capabilities <dimension> element
    */
-  const char** (*print_ogc_formatted_values)(mapcache_context *context, mapcache_dimension *dimension);
+  apr_array_header_t*  (*print_ogc_formatted_values)(mapcache_context *context, mapcache_dimension *dimension);
 
   /**
    * \brief parse the value given in the configuration
@@ -1698,6 +1976,13 @@ struct mapcache_dimension_values {
   int case_sensitive;
 };
 
+struct mapcache_dimension_sqlite {
+  mapcache_dimension dimension;
+  char *dbfile;
+  char *validate_query;
+  char *list_query;
+};
+
 struct mapcache_dimension_regex {
   mapcache_dimension dimension;
   char *regex_string;
@@ -1721,6 +2006,7 @@ struct mapcache_dimension_time {
 };
 
 mapcache_dimension* mapcache_dimension_values_create(apr_pool_t *pool);
+mapcache_dimension* mapcache_dimension_sqlite_create(apr_pool_t *pool);
 mapcache_dimension* mapcache_dimension_regex_create(apr_pool_t *pool);
 mapcache_dimension* mapcache_dimension_intervals_create(apr_pool_t *pool);
 mapcache_dimension* mapcache_dimension_time_create(apr_pool_t *pool);
@@ -1734,13 +2020,13 @@ typedef enum {
   MAPCACHE_TIMEDIMENSION_SOURCE_SQLITE
 } mapcache_timedimension_source_type;
 
-apr_array_header_t* mapcache_timedimension_get_entries_for_value(mapcache_context *ctx, mapcache_timedimension *timedimesnion,
+MS_DLL_EXPORT apr_array_header_t* mapcache_timedimension_get_entries_for_value(mapcache_context *ctx, mapcache_timedimension *timedimesnion,
         mapcache_tileset *tileset, mapcache_grid *grid, mapcache_extent *extent, const char *value);
 
 struct mapcache_timedimension {
   mapcache_timedimension_assembly_type assembly_type;
   void (*configuration_parse_xml)(mapcache_context *context, mapcache_timedimension *dim, ezxml_t node);
-  apr_array_header_t* (*get_entries_for_interval)(mapcache_context *ctx, mapcache_timedimension *dim, mapcache_tileset *tileset, 
+  apr_array_header_t* (*get_entries_for_interval)(mapcache_context *ctx, mapcache_timedimension *dim, mapcache_tileset *tileset,
         mapcache_grid *grid, mapcache_extent *extent, time_t start, time_t end);
   apr_array_header_t* (*get_all_entries)(mapcache_context *ctx, mapcache_timedimension *dim, mapcache_tileset *tileset);
   char *default_value;
@@ -1759,6 +2045,26 @@ mapcache_timedimension* mapcache_timedimension_sqlite_create(apr_pool_t *pool);
 
 int mapcache_is_axis_inverted(const char *srs);
 
+typedef struct mapcache_pooled_connection_container mapcache_pooled_connection_container;
+typedef struct mapcache_pooled_connection mapcache_pooled_connection;
+typedef struct mapcache_pooled_connection_private_data mapcache_pooled_connection_private_data;
+
+struct mapcache_pooled_connection {
+    mapcache_pooled_connection_private_data *private;
+    void *connection;
+};
+
+typedef void (*mapcache_connection_constructor)(mapcache_context *ctx, void **connection, void *params);
+typedef void (*mapcache_connection_destructor)(void *connection);
+
+MS_DLL_EXPORT apr_status_t mapcache_connection_pool_create(mapcache_connection_pool **cp, apr_pool_t *server_pool);
+mapcache_pooled_connection* mapcache_connection_pool_get_connection(mapcache_context *ctx, char *key,
+        mapcache_connection_constructor constructor,
+        mapcache_connection_destructor destructor,
+        void *params);
+void mapcache_connection_pool_invalidate_connection(mapcache_context *ctx, mapcache_pooled_connection *connection);
+void mapcache_connection_pool_release_connection(mapcache_context *ctx, mapcache_pooled_connection *connection);
+
 #endif /* MAPCACHE_H_ */
 /* vim: ts=2 sts=2 et sw=2
 */
