@@ -47,17 +47,20 @@
 
 #include <sqlite3.h>
 
-static apr_hash_t *ro_connection_pools = NULL;
-static apr_hash_t *rw_connection_pools = NULL;
+struct sqlite_conn_params {
+  mapcache_cache_sqlite *cache;
+  char *dbfile;
+  int readonly;
+};
 
 struct sqlite_conn {
   sqlite3 *handle;
-  int readonly;
   int nstatements;
   sqlite3_stmt **prepared_statements;
-  char *errmsg;
 };
 
+#define SQLITE_CONN(pooled_connection) ((struct sqlite_conn*)((pooled_connection)->connection))
+
 #define HAS_TILE_STMT_IDX 0
 #define GET_TILE_STMT_IDX 1
 #define SQLITE_SET_TILE_STMT_IDX 2
@@ -70,8 +73,9 @@ struct sqlite_conn {
 #define MBTILES_DEL_TILE_STMT1_IDX 7
 #define MBTILES_DEL_TILE_STMT2_IDX 8
 
+static void _mapcache_cache_sqlite_filename_for_tile(mapcache_context *ctx, mapcache_cache_sqlite *dcache, mapcache_tile *tile, char **path);
 
-static int _sqlite_set_pragmas(apr_pool_t *pool, mapcache_cache_sqlite* cache, struct sqlite_conn *conn)
+static void _sqlite_set_pragmas(mapcache_context *ctx, mapcache_cache_sqlite* cache, struct sqlite_conn *conn)
 {
   if (cache->pragmas && !apr_is_empty_table(cache->pragmas)) {
     const apr_array_header_t *elts = apr_table_elts(cache->pragmas);
@@ -80,7 +84,7 @@ static int _sqlite_set_pragmas(apr_pool_t *pool, mapcache_cache_sqlite* cache, s
     char *pragma_stmt;
     for (i = 0; i < elts->nelts; i++) {
       apr_table_entry_t entry = APR_ARRAY_IDX(elts, i, apr_table_entry_t);
-      pragma_stmt = apr_psprintf(pool,"PRAGMA %s=%s",entry.key,entry.val);
+      pragma_stmt = apr_psprintf(ctx->pool,"PRAGMA %s=%s",entry.key,entry.val);
       do {
         ret = sqlite3_exec(conn->handle, pragma_stmt, 0, 0, NULL);
         if (ret != SQLITE_OK && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
@@ -88,78 +92,61 @@ static int _sqlite_set_pragmas(apr_pool_t *pool, mapcache_cache_sqlite* cache, s
         }
       } while (ret == SQLITE_BUSY || ret == SQLITE_LOCKED);
       if (ret != SQLITE_OK) {
-        conn->errmsg = apr_psprintf(pool,"failed to execute pragma statement %s",pragma_stmt);
-        return MAPCACHE_FAILURE;
+        ctx->set_error(ctx, 500, "failed to execute pragma statement %s",pragma_stmt);
+        return;
       }
     }
   }
-  return MAPCACHE_SUCCESS;
+  return;
 }
 
-static apr_status_t _sqlite_reslist_get_rw_connection(void **conn_, void *params, apr_pool_t *pool)
+
+
+static void mapcache_sqlite_release_conn(mapcache_context *ctx, mapcache_pooled_connection *conn) {
+  mapcache_connection_pool_release_connection(ctx,conn);
+}
+
+void mapcache_sqlite_connection_constructor(mapcache_context *ctx, void **conn_, void *params)
 {
   int ret;
   int flags;  
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) params;
-  struct sqlite_conn *conn = apr_pcalloc(pool, sizeof (struct sqlite_conn));
+  struct sqlite_conn_params *sq_params = (struct sqlite_conn_params*)params;
+  struct sqlite_conn *conn = calloc(1, sizeof (struct sqlite_conn));
   *conn_ = conn;
-  flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_NOMUTEX | SQLITE_OPEN_CREATE;
-  ret = sqlite3_open_v2(cache->dbfile, &conn->handle, flags, NULL);
+  if(sq_params->readonly) {
+    flags = SQLITE_OPEN_READONLY | SQLITE_OPEN_NOMUTEX;
+  } else {
+    flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_NOMUTEX | SQLITE_OPEN_CREATE;
+    mapcache_make_parent_dirs(ctx,sq_params->dbfile);
+    GC_CHECK_ERROR(ctx);
+  }
+  ret = sqlite3_open_v2(sq_params->dbfile, &conn->handle, flags, NULL);
   if (ret != SQLITE_OK) {
-    conn->errmsg = apr_psprintf(pool,"sqlite backend failed to open db %s: %s", cache->dbfile, sqlite3_errmsg(conn->handle));
-    return APR_EGENERAL;
+    ctx->set_error(ctx,500,"sqlite backend failed to open db %s: %s", sq_params->dbfile, sqlite3_errmsg(conn->handle));
+    return;
   }
   sqlite3_busy_timeout(conn->handle, 300000);
   do {
-    ret = sqlite3_exec(conn->handle, cache->create_stmt.sql, 0, 0, NULL);
+    ret = sqlite3_exec(conn->handle, sq_params->cache->create_stmt.sql, 0, 0, NULL);
     if (ret != SQLITE_OK && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
       break;
     }
   } while (ret == SQLITE_BUSY || ret == SQLITE_LOCKED);
   if (ret != SQLITE_OK) {
-    conn->errmsg = apr_psprintf(pool, "sqlite backend failed to create db schema on %s: %s", cache->dbfile, sqlite3_errmsg(conn->handle));
+    ctx->set_error(ctx,500, "sqlite backend failed to create db schema on %s: %s", sq_params->dbfile, sqlite3_errmsg(conn->handle));
     sqlite3_close(conn->handle);
-    return APR_EGENERAL;
+    return;
   }
-  conn->readonly = 0;
-  ret = _sqlite_set_pragmas(pool, cache, conn);
-  if(ret != MAPCACHE_SUCCESS) {
+  _sqlite_set_pragmas(ctx, sq_params->cache, conn);
+  if(GC_HAS_ERROR(ctx)) {
     sqlite3_close(conn->handle);
-    return APR_EGENERAL;
+    return;
   }
-  conn->prepared_statements = calloc(cache->n_prepared_statements,sizeof(sqlite3_stmt*));
-  conn->nstatements = cache->n_prepared_statements;
-
-  return APR_SUCCESS;
+  conn->prepared_statements = calloc(sq_params->cache->n_prepared_statements,sizeof(sqlite3_stmt*));
+  conn->nstatements = sq_params->cache->n_prepared_statements;
 }
 
-static apr_status_t _sqlite_reslist_get_ro_connection(void **conn_, void *params, apr_pool_t *pool)
-{
-  int ret;
-  int flags;  
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) params;
-  struct sqlite_conn *conn = apr_pcalloc(pool, sizeof (struct sqlite_conn));
-  *conn_ = conn;
-  flags = SQLITE_OPEN_READONLY | SQLITE_OPEN_NOMUTEX;
-  ret = sqlite3_open_v2(cache->dbfile, &conn->handle, flags, NULL);
-  
-  if (ret != SQLITE_OK) {
-    return APR_EGENERAL;
-  }
-  sqlite3_busy_timeout(conn->handle, 300000);
-  conn->readonly = 1;
-
-  ret = _sqlite_set_pragmas(pool,cache, conn);
-  if (ret != MAPCACHE_SUCCESS) {
-    sqlite3_close(conn->handle);
-    return APR_EGENERAL;
-  }
-  conn->prepared_statements = calloc(cache->n_prepared_statements,sizeof(sqlite3_stmt*));
-  conn->nstatements = cache->n_prepared_statements;
-  return APR_SUCCESS;
-}
-
-static apr_status_t _sqlite_reslist_free_connection(void *conn_, void *params, apr_pool_t *pool)
+void mapcache_sqlite_connection_destructor(void *conn_)
 {
   struct sqlite_conn *conn = (struct sqlite_conn*) conn_;
   int i;
@@ -170,109 +157,106 @@ static apr_status_t _sqlite_reslist_free_connection(void *conn_, void *params, a
   }
   free(conn->prepared_statements);
   sqlite3_close(conn->handle);
-  return APR_SUCCESS;
 }
 
-static struct sqlite_conn* _sqlite_get_conn(mapcache_context *ctx, mapcache_tile* tile, int readonly) {
-  apr_status_t rv;
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
-  struct sqlite_conn *conn = NULL;
-  apr_reslist_t *pool = NULL;
-  apr_hash_t *pool_container;
-  if (readonly) {
-    pool_container = ro_connection_pools;
+static mapcache_pooled_connection* mapcache_sqlite_get_conn(mapcache_context *ctx, mapcache_cache_sqlite *cache, mapcache_tile *tile, int readonly) {
+  struct sqlite_conn_params params;
+  mapcache_pooled_connection *pc;
+  char *key;
+  _mapcache_cache_sqlite_filename_for_tile(ctx,cache,tile,&params.dbfile);
+  params.cache = cache;
+  params.readonly = readonly;
+  if(!strstr(cache->dbfile,"{")) {
+    key = apr_pstrcat(ctx->pool,readonly?"ro_":"rw_",cache->cache.name,NULL);
   } else {
-    pool_container = rw_connection_pools;
-  }
-  if(!pool_container || NULL == (pool = apr_hash_get(pool_container,cache->cache.name, APR_HASH_KEY_STRING)) ) {
-#ifdef APR_HAS_THREADS
-    if(ctx->threadlock)
-      apr_thread_mutex_lock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-    if(!ro_connection_pools) {
-      ro_connection_pools = apr_hash_make(ctx->process_pool);
-      rw_connection_pools = apr_hash_make(ctx->process_pool);
-    }
-
-    /* probably doesn't exist, unless the previous mutex locked us, so we check */
-    pool = apr_hash_get(ro_connection_pools,cache->cache.name, APR_HASH_KEY_STRING);
-    if(!pool) {
-      /* there where no existing connection pools, create them*/
-      rv = apr_reslist_create(&pool,
-                              0 /* min */,
-                              10 /* soft max */,
-                              200 /* hard max */,
-                              60*1000000 /*60 seconds, ttl*/,
-                              _sqlite_reslist_get_ro_connection, /* resource constructor */
-                              _sqlite_reslist_free_connection, /* resource destructor */
-                              cache, ctx->process_pool);
-      if(rv != APR_SUCCESS) {
-        ctx->set_error(ctx,500,"failed to create bdb ro connection pool");
-#ifdef APR_HAS_THREADS
-        if(ctx->threadlock)
-          apr_thread_mutex_unlock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-        return NULL;
-      }
-      apr_hash_set(ro_connection_pools,cache->cache.name,APR_HASH_KEY_STRING,pool);
-      rv = apr_reslist_create(&pool,
-                              0 /* min */,
-                              1 /* soft max */,
-                              1 /* hard max */,
-                              60*1000000 /*60 seconds, ttl*/,
-                              _sqlite_reslist_get_rw_connection, /* resource constructor */
-                              _sqlite_reslist_free_connection, /* resource destructor */
-                              cache, ctx->process_pool);
-      if(rv != APR_SUCCESS) {
-        ctx->set_error(ctx,500,"failed to create bdb rw connection pool");
-#ifdef APR_HAS_THREADS
-        if(ctx->threadlock)
-          apr_thread_mutex_unlock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-        return NULL;
-      }
-      apr_hash_set(rw_connection_pools,cache->cache.name,APR_HASH_KEY_STRING,pool);
-    }
-#ifdef APR_HAS_THREADS
-    if(ctx->threadlock)
-      apr_thread_mutex_unlock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-    if(readonly)
-      pool = apr_hash_get(ro_connection_pools,cache->cache.name, APR_HASH_KEY_STRING);
-    else
-      pool = apr_hash_get(rw_connection_pools,cache->cache.name, APR_HASH_KEY_STRING);
-    assert(pool);
-  }
-  rv = apr_reslist_acquire(pool, (void **) &conn);
-  if (rv != APR_SUCCESS) {
-    ctx->set_error(ctx, 500, "failed to aquire connection to sqlite backend: %s", (conn && conn->errmsg)?conn->errmsg:"unknown error");
-    return NULL;
-  }
-  return conn;
-}
-
-static void _sqlite_release_conn(mapcache_context *ctx, mapcache_tile *tile, struct sqlite_conn *conn)
-{
-  apr_reslist_t *pool;
-  apr_hash_t *pool_container;
-  if(conn->readonly) {
-    pool_container = ro_connection_pools;
-  } else {
-    pool_container = rw_connection_pools;
-  }
-  pool = apr_hash_get(pool_container,tile->tileset->cache->name, APR_HASH_KEY_STRING);
-
-  if (GC_HAS_ERROR(ctx)) {
-    apr_reslist_invalidate(pool, (void*) conn);
-  } else {
-    apr_reslist_release(pool, (void*) conn);
+    key = apr_pstrcat(ctx->pool,readonly?"ro_":"rw_",params.dbfile,NULL);
   }
+  pc = mapcache_connection_pool_get_connection(ctx,key,mapcache_sqlite_connection_constructor,mapcache_sqlite_connection_destructor,&params);
+  return pc;
 }
 
 
+/**
+ * \brief return sqlite db filename for given tile
+ *
+ * \param tile the tile to get the key from
+ * \param path pointer to a char* that will contain the filename
+ * \param r
+ * \private \memberof mapcache_cache_sqlite
+ */
+static void _mapcache_cache_sqlite_filename_for_tile(mapcache_context *ctx, mapcache_cache_sqlite *dcache, mapcache_tile *tile, char **path)
+{
+  *path = dcache->dbfile;
+
+  if(strstr(*path,"{")) {
+    /*
+     * generic template substitutions
+     */
+    if(strstr(*path,"{tileset}"))
+      *path = mapcache_util_str_replace(ctx->pool,*path, "{tileset}",
+              tile->tileset->name);
+    if(strstr(*path,"{grid}"))
+      *path = mapcache_util_str_replace(ctx->pool,*path, "{grid}",
+              tile->grid_link->grid->name);
+    if(tile->dimensions && strstr(*path,"{dim}")) {
+      char *dimstring="";
+      const apr_array_header_t *elts = apr_table_elts(tile->dimensions);
+      int i = elts->nelts;
+      while(i--) {
+        apr_table_entry_t *entry = &(APR_ARRAY_IDX(elts,i,apr_table_entry_t));
+        const char *dimval = mapcache_util_str_sanitize(ctx->pool,entry->val,"/.",'#');
+        dimstring = apr_pstrcat(ctx->pool,dimstring,"#",dimval,NULL);
+      }
+      *path = mapcache_util_str_replace(ctx->pool,*path, "{dim}", dimstring);
+    }
+    
+    
+    while(strstr(*path,"{z}"))
+      *path = mapcache_util_str_replace(ctx->pool,*path, "{z}",
+              apr_psprintf(ctx->pool,dcache->z_fmt,tile->z));
+    
+    if(dcache->count_x > 0) {
+      while(strstr(*path,"{div_x}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{div_x}",
+                apr_psprintf(ctx->pool,dcache->div_x_fmt,tile->x/dcache->count_x));
+      while(strstr(*path,"{inv_div_x}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{inv_div_x}",
+                apr_psprintf(ctx->pool,dcache->inv_div_x_fmt,(tile->grid_link->grid->levels[tile->z]->maxx - tile->x - 1)/dcache->count_x));
+      while(strstr(*path,"{x}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{x}",
+                apr_psprintf(ctx->pool,dcache->x_fmt,tile->x/dcache->count_x*dcache->count_x));
+      while(strstr(*path,"{inv_x}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{inv_x}",
+                apr_psprintf(ctx->pool,dcache->inv_x_fmt,(tile->grid_link->grid->levels[tile->z]->maxx - tile->x - 1)/dcache->count_x*dcache->count_x));
+    }
+    
+    if(dcache->count_y > 0) {
+      while(strstr(*path,"{div_y}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{div_y}",
+                apr_psprintf(ctx->pool,dcache->div_y_fmt,tile->y/dcache->count_y));
+      while(strstr(*path,"{inv_div_y}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{inv_div_y}",
+                apr_psprintf(ctx->pool,dcache->inv_div_y_fmt,(tile->grid_link->grid->levels[tile->z]->maxy - tile->y - 1)/dcache->count_y));
+      while(strstr(*path,"{y}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{y}",
+                apr_psprintf(ctx->pool,dcache->y_fmt,tile->y/dcache->count_y*dcache->count_y));
+      while(strstr(*path,"{inv_y}"))
+        *path = mapcache_util_str_replace(ctx->pool,*path, "{inv_y}",
+                apr_psprintf(ctx->pool,dcache->inv_y_fmt,(tile->grid_link->grid->levels[tile->z]->maxy - tile->y - 1)/dcache->count_y*dcache->count_y));
+      
+    }
+  }
+
+  if(!*path) {
+    ctx->set_error(ctx,500, "failed to allocate tile key");
+  }
+}
+
+
+
 /**
  * \brief apply appropriate tile properties to the sqlite statement */
-static void _bind_sqlite_params(mapcache_context *ctx, void *vstmt, mapcache_tile *tile)
+static void _bind_sqlite_params(mapcache_context *ctx, void *vstmt, mapcache_cache_sqlite *cache, mapcache_tile *tile)
 {
   sqlite3_stmt *stmt = vstmt;
   int paramidx;
@@ -311,8 +295,7 @@ static void _bind_sqlite_params(mapcache_context *ctx, void *vstmt, mapcache_til
   paramidx = sqlite3_bind_parameter_index(stmt, ":data");
   if (paramidx) {
     int written = 0;
-    if(((mapcache_cache_sqlite*)tile->tileset->cache)->detect_blank && tile->grid_link->grid->tile_sx == 256 &&
-            tile->grid_link->grid->tile_sy == 256) {
+    if(cache->detect_blank) {
       if(!tile->raw_image) {
         tile->raw_image = mapcache_imageio_decode(ctx, tile->encoded_data);
         GC_CHECK_ERROR(ctx);
@@ -339,7 +322,7 @@ static void _bind_sqlite_params(mapcache_context *ctx, void *vstmt, mapcache_til
   }
 }
 
-static void _bind_mbtiles_params(mapcache_context *ctx, void *vstmt, mapcache_tile *tile)
+static void _bind_mbtiles_params(mapcache_context *ctx, void *vstmt, mapcache_cache_sqlite *cache, mapcache_tile *tile)
 {
   sqlite3_stmt *stmt = vstmt;
   int paramidx;
@@ -389,26 +372,29 @@ static void _bind_mbtiles_params(mapcache_context *ctx, void *vstmt, mapcache_ti
 
 }
 
-static int _mapcache_cache_sqlite_has_tile(mapcache_context *ctx, mapcache_tile *tile)
+static int _mapcache_cache_sqlite_has_tile(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, tile, 1);
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) pcache;
+  mapcache_pooled_connection *pc;
+  struct sqlite_conn *conn;
   sqlite3_stmt *stmt;
   int ret;
+  pc = mapcache_sqlite_get_conn(ctx,cache,tile,1);
   if (GC_HAS_ERROR(ctx)) {
-    if(conn) _sqlite_release_conn(ctx, tile, conn);
+    if(pc) mapcache_sqlite_release_conn(ctx, pc);
     if(!tile->tileset->read_only && tile->tileset->source) {
       /* not an error in this case, as the db file may not have been created yet */
       ctx->clear_errors(ctx);
     }
     return MAPCACHE_FALSE;
   }
+  conn = SQLITE_CONN(pc);
   stmt = conn->prepared_statements[HAS_TILE_STMT_IDX];
   if(!stmt) {
     sqlite3_prepare(conn->handle, cache->exists_stmt.sql, -1, &conn->prepared_statements[HAS_TILE_STMT_IDX], NULL);
     stmt = conn->prepared_statements[HAS_TILE_STMT_IDX];
   }
-  cache->bind_stmt(ctx, stmt, tile);
+  cache->bind_stmt(ctx, stmt, cache, tile);
   ret = sqlite3_step(stmt);
   if (ret != SQLITE_DONE && ret != SQLITE_ROW) {
     ctx->set_error(ctx, 500, "sqlite backend failed on has_tile: %s", sqlite3_errmsg(conn->handle));
@@ -419,46 +405,53 @@ static int _mapcache_cache_sqlite_has_tile(mapcache_context *ctx, mapcache_tile
     ret = MAPCACHE_TRUE;
   }
   sqlite3_reset(stmt);
-  _sqlite_release_conn(ctx, tile, conn);
+  mapcache_sqlite_release_conn(ctx, pc);
   return ret;
 }
 
-static void _mapcache_cache_sqlite_delete(mapcache_context *ctx, mapcache_tile *tile)
+static void _mapcache_cache_sqlite_delete(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, tile, 0);
-  sqlite3_stmt *stmt = conn->prepared_statements[SQLITE_DEL_TILE_STMT_IDX];
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) pcache;
+  mapcache_pooled_connection *pc;
+  struct sqlite_conn *conn;
+  sqlite3_stmt *stmt;
   int ret;
+  pc = mapcache_sqlite_get_conn(ctx,cache,tile,0);
   if (GC_HAS_ERROR(ctx)) {
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return;
   }
+  conn = SQLITE_CONN(pc);
+  stmt = conn->prepared_statements[SQLITE_DEL_TILE_STMT_IDX];
   if(!stmt) {
     sqlite3_prepare(conn->handle, cache->delete_stmt.sql, -1, &conn->prepared_statements[SQLITE_DEL_TILE_STMT_IDX], NULL);
     stmt = conn->prepared_statements[SQLITE_DEL_TILE_STMT_IDX];
   }
-  cache->bind_stmt(ctx, stmt, tile);
+  cache->bind_stmt(ctx, stmt, cache, tile);
   ret = sqlite3_step(stmt);
   if (ret != SQLITE_DONE && ret != SQLITE_ROW) {
     ctx->set_error(ctx, 500, "sqlite backend failed on delete: %s", sqlite3_errmsg(conn->handle));
   }
   sqlite3_reset(stmt);
-  _sqlite_release_conn(ctx, tile, conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
 
-static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_tile *tile)
+static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, tile, 0);
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) pcache;
+  mapcache_pooled_connection *pc;
+  struct sqlite_conn *conn;
   sqlite3_stmt *stmt1,*stmt2,*stmt3;
   int ret;
   const char *tile_id;
   size_t tile_id_size;
+  pc = mapcache_sqlite_get_conn(ctx,cache,tile,0);
   if (GC_HAS_ERROR(ctx)) {
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return;
   }
+  conn = SQLITE_CONN(pc);
   stmt1 = conn->prepared_statements[MBTILES_DEL_TILE_SELECT_STMT_IDX];
   stmt2 = conn->prepared_statements[MBTILES_DEL_TILE_STMT1_IDX];
   stmt3 = conn->prepared_statements[MBTILES_DEL_TILE_STMT2_IDX];
@@ -474,19 +467,19 @@ static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_tile
   /* first extract tile_id from the tile we will delete. We need this because we do not know
    * if the tile is empty or not.
    * If it is empty, we will not delete the image blob data from the images table */
-  cache->bind_stmt(ctx, stmt1, tile);
+  cache->bind_stmt(ctx, stmt1, cache, tile);
   do {
     ret = sqlite3_step(stmt1);
     if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
       ctx->set_error(ctx, 500, "sqlite backend failed on mbtile del 1: %s", sqlite3_errmsg(conn->handle));
       sqlite3_reset(stmt1);
-      _sqlite_release_conn(ctx, tile, conn);
+      mapcache_sqlite_release_conn(ctx, pc);
       return;
     }
   } while (ret == SQLITE_BUSY || ret == SQLITE_LOCKED);
   if (ret == SQLITE_DONE) { /* tile does not exist, ignore */
     sqlite3_reset(stmt1);
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return;
   } else {
     tile_id = (const char*) sqlite3_column_text(stmt1, 0);
@@ -495,13 +488,13 @@ static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_tile
 
 
   /* delete the tile from the "map" table */
-  cache->bind_stmt(ctx,stmt2, tile);
+  cache->bind_stmt(ctx,stmt2, cache, tile);
   ret = sqlite3_step(stmt2);
   if (ret != SQLITE_DONE && ret != SQLITE_ROW) {
     ctx->set_error(ctx, 500, "sqlite backend failed on mbtile del 2: %s", sqlite3_errmsg(conn->handle));
     sqlite3_reset(stmt1);
     sqlite3_reset(stmt2);
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return;
   }
 
@@ -517,7 +510,7 @@ static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_tile
       sqlite3_reset(stmt1);
       sqlite3_reset(stmt2);
       sqlite3_reset(stmt3);
-      _sqlite_release_conn(ctx, tile, conn);
+      mapcache_sqlite_release_conn(ctx, pc);
       return;
     }
   }
@@ -525,15 +518,14 @@ static void _mapcache_cache_mbtiles_delete(mapcache_context *ctx, mapcache_tile
   sqlite3_reset(stmt1);
   sqlite3_reset(stmt2);
   sqlite3_reset(stmt3);
-  _sqlite_release_conn(ctx, tile, conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
 
 
-static void _single_mbtile_set(mapcache_context *ctx, mapcache_tile *tile, struct sqlite_conn *conn)
+static void _single_mbtile_set(mapcache_context *ctx, mapcache_cache_sqlite *cache, mapcache_tile *tile, struct sqlite_conn *conn)
 {
   sqlite3_stmt *stmt1,*stmt2;
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)tile->tileset->cache;
   int ret;
   if(!tile->raw_image) {
     tile->raw_image = mapcache_imageio_decode(ctx, tile->encoded_data);
@@ -552,8 +544,8 @@ static void _single_mbtile_set(mapcache_context *ctx, mapcache_tile *tile, struc
       stmt1 = conn->prepared_statements[MBTILES_SET_EMPTY_TILE_STMT1_IDX];
       stmt2 = conn->prepared_statements[MBTILES_SET_EMPTY_TILE_STMT2_IDX];
     }
-    cache->bind_stmt(ctx, stmt1, tile);
-    cache->bind_stmt(ctx, stmt2, tile);
+    cache->bind_stmt(ctx, stmt1, cache, tile);
+    cache->bind_stmt(ctx, stmt2, cache, tile);
   } else {
     stmt1 = conn->prepared_statements[MBTILES_SET_TILE_STMT1_IDX];
     stmt2 = conn->prepared_statements[MBTILES_SET_TILE_STMT2_IDX];
@@ -567,8 +559,8 @@ static void _single_mbtile_set(mapcache_context *ctx, mapcache_tile *tile, struc
       stmt1 = conn->prepared_statements[MBTILES_SET_TILE_STMT1_IDX];
       stmt2 = conn->prepared_statements[MBTILES_SET_TILE_STMT2_IDX];
     }
-    cache->bind_stmt(ctx, stmt1, tile);
-    cache->bind_stmt(ctx, stmt2, tile);
+    cache->bind_stmt(ctx, stmt1, cache, tile);
+    cache->bind_stmt(ctx, stmt2, cache, tile);
   }
   do {
     ret = sqlite3_step(stmt1);
@@ -596,47 +588,49 @@ static void _single_mbtile_set(mapcache_context *ctx, mapcache_tile *tile, struc
   sqlite3_reset(stmt2);
 }
 
-static int _mapcache_cache_sqlite_get(mapcache_context *ctx, mapcache_tile *tile)
+static int _mapcache_cache_sqlite_get(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) pcache;
   struct sqlite_conn *conn;
   sqlite3_stmt *stmt;
   int ret;
-  conn = _sqlite_get_conn(ctx, tile, 1);
+  mapcache_pooled_connection *pc = mapcache_sqlite_get_conn(ctx,cache,tile,1);
   if (GC_HAS_ERROR(ctx)) {
-    if(conn) _sqlite_release_conn(ctx, tile, conn);
     if(tile->tileset->read_only || !tile->tileset->source) {
+      mapcache_sqlite_release_conn(ctx, pc);
       return MAPCACHE_FAILURE;
     } else {
       /* not an error in this case, as the db file may not have been created yet */
       ctx->clear_errors(ctx);
+      mapcache_sqlite_release_conn(ctx, pc);
       return MAPCACHE_CACHE_MISS;
     }
   }
+  conn = SQLITE_CONN(pc);
   stmt = conn->prepared_statements[GET_TILE_STMT_IDX];
   if(!stmt) {
     sqlite3_prepare(conn->handle, cache->get_stmt.sql, -1, &conn->prepared_statements[GET_TILE_STMT_IDX], NULL);
     stmt = conn->prepared_statements[GET_TILE_STMT_IDX];
   }
-  cache->bind_stmt(ctx, stmt, tile);
+  cache->bind_stmt(ctx, stmt, cache, tile);
   do {
     ret = sqlite3_step(stmt);
     if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
       ctx->set_error(ctx, 500, "sqlite backend failed on get: %s", sqlite3_errmsg(conn->handle));
       sqlite3_reset(stmt);
-      _sqlite_release_conn(ctx, tile, conn);
+      mapcache_sqlite_release_conn(ctx, pc);
       return MAPCACHE_FAILURE;
     }
   } while (ret == SQLITE_BUSY || ret == SQLITE_LOCKED);
   if (ret == SQLITE_DONE) {
     sqlite3_reset(stmt);
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return MAPCACHE_CACHE_MISS;
   } else {
     const void *blob = sqlite3_column_blob(stmt, 0);
     int size = sqlite3_column_bytes(stmt, 0);
     if(size>0 && ((char*)blob)[0] == '#') {
-      tile->encoded_data = mapcache_empty_png_decode(ctx,blob,&tile->nodata);
+      tile->encoded_data = mapcache_empty_png_decode(ctx,tile->grid_link->grid->tile_sx, tile->grid_link->grid->tile_sy ,blob,&tile->nodata);
     } else {
       tile->encoded_data = mapcache_buffer_create(size, ctx->pool);
       memcpy(tile->encoded_data->buf, blob, size);
@@ -647,14 +641,13 @@ static int _mapcache_cache_sqlite_get(mapcache_context *ctx, mapcache_tile *tile
       apr_time_ansi_put(&(tile->mtime), mtime);
     }
     sqlite3_reset(stmt);
-    _sqlite_release_conn(ctx, tile, conn);
+    mapcache_sqlite_release_conn(ctx, pc);
     return MAPCACHE_SUCCESS;
   }
 }
 
-static void _single_sqlitetile_set(mapcache_context *ctx, mapcache_tile *tile, struct sqlite_conn *conn)
+static void _single_sqlitetile_set(mapcache_context *ctx, mapcache_cache_sqlite *cache, mapcache_tile *tile, struct sqlite_conn *conn)
 {
-  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*) tile->tileset->cache;
   sqlite3_stmt *stmt = conn->prepared_statements[SQLITE_SET_TILE_STMT_IDX];
   int ret;
 
@@ -662,7 +655,7 @@ static void _single_sqlitetile_set(mapcache_context *ctx, mapcache_tile *tile, s
     sqlite3_prepare(conn->handle, cache->set_stmt.sql, -1, &conn->prepared_statements[SQLITE_SET_TILE_STMT_IDX], NULL);
     stmt = conn->prepared_statements[SQLITE_SET_TILE_STMT_IDX];
   }
-  cache->bind_stmt(ctx, stmt, tile);
+  cache->bind_stmt(ctx, stmt, cache, tile);
   do {
     ret = sqlite3_step(stmt);
     if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
@@ -676,29 +669,41 @@ static void _single_sqlitetile_set(mapcache_context *ctx, mapcache_tile *tile, s
   sqlite3_reset(stmt);
 }
 
-static void _mapcache_cache_sqlite_set(mapcache_context *ctx, mapcache_tile *tile)
+static void _mapcache_cache_sqlite_set(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, tile, 0);
-  GC_CHECK_ERROR(ctx);
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)pcache;
+  struct sqlite_conn *conn;
+  mapcache_pooled_connection *pc = mapcache_sqlite_get_conn(ctx,cache,tile,0);
+  if (GC_HAS_ERROR(ctx)) {
+    mapcache_sqlite_release_conn(ctx, pc);
+    return;
+  }
+  conn = SQLITE_CONN(pc);
   sqlite3_exec(conn->handle, "BEGIN TRANSACTION", 0, 0, 0);
-  _single_sqlitetile_set(ctx,tile,conn);
+  _single_sqlitetile_set(ctx,cache, tile,conn);
   if (GC_HAS_ERROR(ctx)) {
     sqlite3_exec(conn->handle, "ROLLBACK TRANSACTION", 0, 0, 0);
   } else {
     sqlite3_exec(conn->handle, "END TRANSACTION", 0, 0, 0);
   }
-  _sqlite_release_conn(ctx, tile, conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
-static void _mapcache_cache_sqlite_multi_set(mapcache_context *ctx, mapcache_tile *tiles, int ntiles)
+static void _mapcache_cache_sqlite_multi_set(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tiles, int ntiles)
 {
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, &tiles[0], 0);
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)pcache;
   int i;
-  GC_CHECK_ERROR(ctx);
+  struct sqlite_conn *conn;
+  mapcache_pooled_connection *pc = mapcache_sqlite_get_conn(ctx,cache,&tiles[0],0);
+  if (GC_HAS_ERROR(ctx)) {
+    mapcache_sqlite_release_conn(ctx, pc);
+    return;
+  }
+  conn = SQLITE_CONN(pc);
   sqlite3_exec(conn->handle, "BEGIN TRANSACTION", 0, 0, 0);
   for (i = 0; i < ntiles; i++) {
     mapcache_tile *tile = &tiles[i];
-    _single_sqlitetile_set(ctx,tile,conn);
+    _single_sqlitetile_set(ctx,cache, tile,conn);
     if(GC_HAS_ERROR(ctx)) break;
   }
   if (GC_HAS_ERROR(ctx)) {
@@ -706,34 +711,42 @@ static void _mapcache_cache_sqlite_multi_set(mapcache_context *ctx, mapcache_til
   } else {
     sqlite3_exec(conn->handle, "END TRANSACTION", 0, 0, 0);
   }
-  _sqlite_release_conn(ctx, &tiles[0], conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
-static void _mapcache_cache_mbtiles_set(mapcache_context *ctx, mapcache_tile *tile)
+static void _mapcache_cache_mbtiles_set(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tile)
 {
-  struct sqlite_conn *conn = _sqlite_get_conn(ctx, tile, 0);
-  GC_CHECK_ERROR(ctx);
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)pcache;
+  struct sqlite_conn *conn;
+  mapcache_pooled_connection *pc = mapcache_sqlite_get_conn(ctx,cache,tile,0);
+  if (GC_HAS_ERROR(ctx)) {
+    mapcache_sqlite_release_conn(ctx, pc);
+    return;
+  }
+  conn = SQLITE_CONN(pc);
   if(!tile->raw_image) {
     tile->raw_image = mapcache_imageio_decode(ctx, tile->encoded_data);
     if(GC_HAS_ERROR(ctx)) {
-      _sqlite_release_conn(ctx, tile, conn);
+      mapcache_sqlite_release_conn(ctx, pc);
       return;
     }
   }
   sqlite3_exec(conn->handle, "BEGIN TRANSACTION", 0, 0, 0);
-  _single_mbtile_set(ctx,tile,conn);
+  _single_mbtile_set(ctx, cache, tile,conn);
   if (GC_HAS_ERROR(ctx)) {
     sqlite3_exec(conn->handle, "ROLLBACK TRANSACTION", 0, 0, 0);
   } else {
     sqlite3_exec(conn->handle, "END TRANSACTION", 0, 0, 0);
   }
-  _sqlite_release_conn(ctx, tile, conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
-static void _mapcache_cache_mbtiles_multi_set(mapcache_context *ctx, mapcache_tile *tiles, int ntiles)
+static void _mapcache_cache_mbtiles_multi_set(mapcache_context *ctx, mapcache_cache *pcache, mapcache_tile *tiles, int ntiles)
 {
-  struct sqlite_conn *conn = NULL;
   int i;
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)pcache;
+  mapcache_pooled_connection *pc;
+  struct sqlite_conn *conn;
 
   /* decode/encode image data before going into the sqlite write lock */
   for (i = 0; i < ntiles; i++) {
@@ -748,12 +761,17 @@ static void _mapcache_cache_mbtiles_multi_set(mapcache_context *ctx, mapcache_ti
       GC_CHECK_ERROR(ctx);
     }
   }
-  conn = _sqlite_get_conn(ctx, &tiles[0], 0);
+  pc = mapcache_sqlite_get_conn(ctx,cache,&tiles[0],0);
+  if (GC_HAS_ERROR(ctx)) {
+    mapcache_sqlite_release_conn(ctx, pc);
+    return;
+  }
+  conn = SQLITE_CONN(pc);
 
   sqlite3_exec(conn->handle, "BEGIN TRANSACTION", 0, 0, 0);
   for (i = 0; i < ntiles; i++) {
     mapcache_tile *tile = &tiles[i];
-    _single_mbtile_set(ctx,tile,conn);
+    _single_mbtile_set(ctx,cache,tile,conn);
     if(GC_HAS_ERROR(ctx)) break;
   }
   if (GC_HAS_ERROR(ctx)) {
@@ -761,32 +779,69 @@ static void _mapcache_cache_mbtiles_multi_set(mapcache_context *ctx, mapcache_ti
   } else {
     sqlite3_exec(conn->handle, "END TRANSACTION", 0, 0, 0);
   }
-  _sqlite_release_conn(ctx, &tiles[0], conn);
+  mapcache_sqlite_release_conn(ctx, pc);
 }
 
-static void _mapcache_cache_sqlite_configuration_parse_xml(mapcache_context *ctx, ezxml_t node, mapcache_cache *cache, mapcache_cfg *config)
+static void _mapcache_cache_sqlite_configuration_parse_xml(mapcache_context *ctx, ezxml_t node, mapcache_cache *pcache, mapcache_cfg *config)
 {
   ezxml_t cur_node;
-  mapcache_cache_sqlite *dcache;
+  mapcache_cache_sqlite *cache;
   sqlite3_initialize();
   sqlite3_config(SQLITE_CONFIG_MULTITHREAD);
-  dcache = (mapcache_cache_sqlite*) cache;
+  cache = (mapcache_cache_sqlite*) pcache;
   if ((cur_node = ezxml_child(node, "base")) != NULL) {
     ctx->set_error(ctx, 500, "sqlite config <base> not supported anymore, use <dbfile>");
     return;
   }
   if ((cur_node = ezxml_child(node, "dbname_template")) != NULL) {
-    ctx->set_error(ctx, 500, "sqlite config <dbname_template> not supported anymore, use <dbfile>");
+    ctx->set_error(ctx, 500, "sqlite config <dbname_template> not supported anymore, use a \"multi-sqlite3\" cache type");
     return;
   }
   if ((cur_node = ezxml_child(node, "dbfile")) != NULL) {
-    dcache->dbfile = apr_pstrdup(ctx->pool, cur_node->txt);
+    char *fmt;
+    cache->dbfile = apr_pstrdup(ctx->pool, cur_node->txt);
+    fmt = (char*)ezxml_attr(cur_node,"x_fmt");
+    if(fmt && *fmt) {
+      cache->x_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"y_fmt");
+    if(fmt && *fmt) {
+      cache->y_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"z_fmt");
+    if(fmt && *fmt) {
+      cache->z_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"inv_x_fmt");
+    if(fmt && *fmt) {
+      cache->inv_x_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"inv_y_fmt");
+    if(fmt && *fmt) {
+      cache->inv_y_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"div_x_fmt");
+    if(fmt && *fmt) {
+      cache->div_x_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"div_y_fmt");
+    if(fmt && *fmt) {
+      cache->div_y_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"inv_div_x_fmt");
+    if(fmt && *fmt) {
+      cache->inv_div_x_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
+    fmt = (char*)ezxml_attr(cur_node,"inv_div_y_fmt");
+    if(fmt && *fmt) {
+      cache->inv_div_y_fmt = apr_pstrdup(ctx->pool,fmt);
+    }
   }
   
-  dcache->detect_blank = 0;
+  cache->detect_blank = 0;
   if ((cur_node = ezxml_child(node, "detect_blank")) != NULL) {
-    if(!strcasecmp(cur_node->txt,"true")) {
-      dcache->detect_blank = 1;
+    if(strcasecmp(cur_node->txt,"false")) {
+      cache->detect_blank = 1;
     }
   }
 
@@ -796,20 +851,53 @@ static void _mapcache_cache_sqlite_configuration_parse_xml(mapcache_context *ctx
     }
   }
   if ((cur_node = ezxml_child(node, "pragma")) != NULL) {
-    dcache->pragmas = apr_table_make(ctx->pool,1);
+    cache->pragmas = apr_table_make(ctx->pool,1);
     while(cur_node) {
       char *name = (char*)ezxml_attr(cur_node,"name");
       if(!name || !cur_node->txt || !strlen(cur_node->txt)) {
         ctx->set_error(ctx,500,"<pragma> missing name attribute");
         return;
       }
-      apr_table_set(dcache->pragmas,name,cur_node->txt);
+      apr_table_set(cache->pragmas,name,cur_node->txt);
       cur_node = cur_node->next;
     }
   }
-  if (!dcache->dbfile) {
-    ctx->set_error(ctx, 500, "sqlite cache \"%s\" is missing <dbfile> entry", cache->name);
-    return;
+  if ((cur_node = ezxml_child(node, "queries")) != NULL) {
+    ezxml_t query_node;
+    if ((query_node = ezxml_child(cur_node, "exists")) != NULL) {
+      cache->exists_stmt.sql = apr_pstrdup(ctx->pool,query_node->txt);
+    }
+    if ((query_node = ezxml_child(cur_node, "get")) != NULL) {
+      cache->get_stmt.sql = apr_pstrdup(ctx->pool,query_node->txt);
+    }
+    if ((query_node = ezxml_child(cur_node, "set")) != NULL) {
+      cache->set_stmt.sql = apr_pstrdup(ctx->pool,query_node->txt);
+    }
+    if ((query_node = ezxml_child(cur_node, "delete")) != NULL) {
+      cache->delete_stmt.sql = apr_pstrdup(ctx->pool,query_node->txt);
+    }
+    if ((query_node = ezxml_child(cur_node, "create")) != NULL) {
+      cache->create_stmt.sql = apr_pstrdup(ctx->pool,query_node->txt);
+    }
+  }
+  
+  cur_node = ezxml_child(node,"xcount");
+  if(cur_node && cur_node->txt && *cur_node->txt) {
+    char *endptr;
+    cache->count_x = (int)strtol(cur_node->txt,&endptr,10);
+    if(*endptr != 0) {
+      ctx->set_error(ctx,400,"failed to parse xcount value %s for sqlite cache %s", cur_node->txt,cache->cache.name);
+      return;
+    }
+  }
+  cur_node = ezxml_child(node,"ycount");
+  if(cur_node && cur_node->txt && *cur_node->txt) {
+    char *endptr;
+    cache->count_y = (int)strtol(cur_node->txt,&endptr,10);
+    if(*endptr != 0) {
+      ctx->set_error(ctx,400,"failed to parse ycount value %s for sqlite cache %s", cur_node->txt,cache->cache.name);
+      return;
+    }
   }
 }
 
@@ -817,18 +905,24 @@ static void _mapcache_cache_sqlite_configuration_parse_xml(mapcache_context *ctx
  * \private \memberof mapcache_cache_sqlite
  */
 static void _mapcache_cache_sqlite_configuration_post_config(mapcache_context *ctx,
-    mapcache_cache *cache, mapcache_cfg *cfg)
+    mapcache_cache *pcache, mapcache_cfg *cfg)
 {
+  mapcache_cache_sqlite *cache = (mapcache_cache_sqlite*)pcache;
+  if (!cache->dbfile) {
+    ctx->set_error(ctx, 500, "sqlite cache \"%s\" is missing <dbfile> entry", pcache->name);
+    return;
+  }
 }
 
 /**
  * \private \memberof mapcache_cache_sqlite
  */
 static void _mapcache_cache_mbtiles_configuration_post_config(mapcache_context *ctx,
-    mapcache_cache *cache, mapcache_cfg *cfg)
+    mapcache_cache *pcache, mapcache_cfg *cfg)
 {
   /* check that only one tileset/grid references this cache, as mbtiles does
    not support multiple tilesets/grids per cache */
+#ifdef FIXME
   int nrefs = 0;
   apr_hash_index_t *tileseti = apr_hash_first(ctx->pool,cfg->tilesets);
   while(tileseti) {
@@ -849,12 +943,9 @@ static void _mapcache_cache_mbtiles_configuration_post_config(mapcache_context *
     }
     tileseti = apr_hash_next(tileseti);
   }
-
+#endif
 }
 
-/**
- * \brief creates and initializes a mapcache_sqlite_cache
- */
 mapcache_cache* mapcache_cache_sqlite_create(mapcache_context *ctx)
 {
   mapcache_cache_sqlite *cache = apr_pcalloc(ctx->pool, sizeof (mapcache_cache_sqlite));
@@ -884,9 +975,15 @@ mapcache_cache* mapcache_cache_sqlite_create(mapcache_context *ctx)
   cache->n_prepared_statements = 4;
   cache->bind_stmt = _bind_sqlite_params;
   cache->detect_blank = 1;
-  return (mapcache_cache*) cache;
+  cache->x_fmt = cache->y_fmt = cache->z_fmt
+          = cache->inv_x_fmt = cache->inv_y_fmt
+          = cache->div_x_fmt = cache->div_y_fmt
+          = cache->inv_div_x_fmt = cache->inv_div_y_fmt = apr_pstrdup(ctx->pool,"%d");
+  cache->count_x = cache->count_y = -1;
+  return (mapcache_cache*)cache;
 }
 
+
 /**
  * \brief creates and initializes a mapcache_sqlite_cache
  */
