@@ -96,6 +96,7 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
     char *name = (char*)ezxml_attr(dimension_node,"name");
     char *type = (char*)ezxml_attr(dimension_node,"type");
     char *unit = (char*)ezxml_attr(dimension_node,"unit");
+    char *skip_validation = (char*)ezxml_attr(dimension_node,"skip_validation");
     char *default_value = (char*)ezxml_attr(dimension_node,"default");
 
     mapcache_dimension *dimension = NULL;
@@ -112,6 +113,8 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
         dimension = mapcache_dimension_regex_create(ctx->pool);
       } else if(!strcmp(type,"intervals")) {
         dimension = mapcache_dimension_intervals_create(ctx->pool);
+      } else if(!strcmp(type,"sqlite")) {
+        dimension = mapcache_dimension_sqlite_create(ctx->pool);
       } else if(!strcmp(type,"time")) {
         ctx->set_error(ctx,501,"time dimension type not implemented yet");
         return;
@@ -130,6 +133,10 @@ void parseDimensions(mapcache_context *ctx, ezxml_t node, mapcache_tileset *tile
     if(unit && *unit) {
       dimension->unit = apr_pstrdup(ctx->pool,unit);
     }
+    
+    if(skip_validation && !strcmp(skip_validation,"true")) {
+      dimension->skip_validation = MAPCACHE_TRUE;
+    }
 
     if(default_value && *default_value) {
       dimension->default_value = apr_pstrdup(ctx->pool,default_value);
@@ -387,6 +394,8 @@ void parseFormat(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
         compression = MAPCACHE_COMPRESSION_FAST;
       } else if(!strcmp(cur_node->txt, "best")) {
         compression = MAPCACHE_COMPRESSION_BEST;
+      } else if(!strcmp(cur_node->txt, "none")) {
+        compression = MAPCACHE_COMPRESSION_DISABLE;
       } else {
         ctx->set_error(ctx, 400, "unknown compression type %s for format \"%s\"", cur_node->txt, name);
         return;
@@ -496,6 +505,20 @@ void parseCache(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
   }
   if(!strcmp(type,"disk")) {
     cache = mapcache_cache_disk_create(ctx);
+  } else if(!strcmp(type,"fallback")) {
+    cache = mapcache_cache_fallback_create(ctx);
+  } else if(!strcmp(type,"multitier")) {
+    cache = mapcache_cache_multitier_create(ctx);
+  } else if(!strcmp(type,"composite")) {
+    cache = mapcache_cache_composite_create(ctx);
+  } else if(!strcmp(type,"rest")) {
+    cache = mapcache_cache_rest_create(ctx);
+  } else if(!strcmp(type,"s3")) {
+    cache = mapcache_cache_s3_create(ctx);
+  } else if(!strcmp(type,"azure")) {
+    cache = mapcache_cache_azure_create(ctx);
+  } else if(!strcmp(type,"google")) {
+    cache = mapcache_cache_google_create(ctx);
   } else if(!strcmp(type,"bdb")) {
 #ifdef USE_BDB
     cache = mapcache_cache_bdb_create(ctx);
@@ -537,6 +560,18 @@ void parseCache(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
 #else
     ctx->set_error(ctx,400, "failed to add cache \"%s\": tiff support is not available on this build",name);
     return;
+#endif
+  } else if(!strcmp(type,"couchbase")) {
+#ifdef USE_COUCHBASE
+    cache = mapcache_cache_couchbase_create(ctx);
+#else
+    ctx->set_error(ctx, 400, "failed to add cache \"%s\": couchbase support is not available on this build", name);
+#endif
+  } else if(!strcmp(type,"riak")) {
+#ifdef USE_RIAK
+    cache = mapcache_cache_riak_create(ctx);
+#else
+    ctx->set_error(ctx, 400, "failed to add cache \"%s\": riak support is not available on this build", name);
 #endif
   } else {
     ctx->set_error(ctx, 400, "unknown cache type %s for cache \"%s\"", type, name);
@@ -632,6 +667,7 @@ void parseTileset(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
     gridlink->maxz = grid->nlevels;
     gridlink->grid_limits = (mapcache_extent_i*)apr_pcalloc(ctx->pool,grid->nlevels*sizeof(mapcache_extent_i));
     gridlink->outofzoom_strategy = MAPCACHE_OUTOFZOOM_NOTCONFIGURED;
+    gridlink->intermediate_grids = apr_array_make(ctx->pool,1,sizeof(mapcache_grid_link*));
 
     restrictedExtent = (char*)ezxml_attr(cur_node,"restricted_extent");
     if(restrictedExtent) {
@@ -667,6 +703,11 @@ void parseTileset(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
         return;
       }
     }
+    sTolerance = (char*)ezxml_attr(cur_node,"use_wms_intermediate_resolutions");
+    if(sTolerance && !strcmp(sTolerance,"true")) {
+      mapcache_grid_link *intermediate_gridlink = apr_pcalloc(ctx->pool,sizeof(mapcache_grid_link));
+      APR_ARRAY_PUSH(gridlink->intermediate_grids,mapcache_grid_link*) = intermediate_gridlink;
+    }
 
 
     mapcache_grid_compute_limits(grid,extent,gridlink->grid_limits,tolerance);
@@ -741,6 +782,40 @@ void parseTileset(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
     if(!havewgs84bbox && !strcasecmp(grid->srs,"EPSG:4326")) {
       tileset->wgs84bbox = *extent;
     }
+
+    if(gridlink->intermediate_grids->nelts > 0) {
+      double factor = 0.5, unitheight,unitwidth;
+      int i;
+      mapcache_grid_link *igl = APR_ARRAY_IDX(gridlink->intermediate_grids, 0, mapcache_grid_link*);
+      igl->restricted_extent = gridlink->restricted_extent;
+      igl->minz = gridlink->minz;
+      igl->max_cached_zoom = gridlink->max_cached_zoom - 1;
+      igl->maxz = gridlink->maxz - 1;
+      igl->outofzoom_strategy = gridlink->outofzoom_strategy;
+      igl->grid = mapcache_grid_create(ctx->pool);
+      igl->grid->extent = gridlink->grid->extent;
+      igl->grid->name = apr_psprintf(ctx->pool,"%s_intermediate_%g",gridlink->grid->name,factor);
+      igl->grid->nlevels = gridlink->grid->nlevels - 1;
+      igl->grid->origin = gridlink->grid->origin;
+      igl->grid->srs = gridlink->grid->srs;
+      igl->grid->srs_aliases = gridlink->grid->srs_aliases;
+      igl->grid->unit = gridlink->grid->unit;
+      igl->grid->tile_sx = gridlink->grid->tile_sx + gridlink->grid->tile_sx * factor;
+      igl->grid->tile_sy = gridlink->grid->tile_sy + gridlink->grid->tile_sy * factor;
+      igl->grid->levels = (mapcache_grid_level**)apr_pcalloc(ctx->pool, igl->grid->nlevels*sizeof(mapcache_grid_level*));
+      for(i=0; i<igl->grid->nlevels; i++) {
+        mapcache_grid_level *level = (mapcache_grid_level*)apr_pcalloc(ctx->pool,sizeof(mapcache_grid_level));
+        level->resolution = gridlink->grid->levels[i]->resolution + (gridlink->grid->levels[i+1]->resolution - gridlink->grid->levels[i]->resolution) * factor;
+        unitheight = igl->grid->tile_sy * level->resolution;
+        unitwidth = igl->grid->tile_sx * level->resolution;
+        
+        level->maxy = ceil((igl->grid->extent.maxy-igl->grid->extent.miny - 0.01* unitheight)/unitheight);
+        level->maxx = ceil((igl->grid->extent.maxx-igl->grid->extent.minx - 0.01* unitwidth)/unitwidth);
+        igl->grid->levels[i] = level;
+      }
+      igl->grid_limits = (mapcache_extent_i*)apr_pcalloc(ctx->pool,igl->grid->nlevels*sizeof(mapcache_extent_i));
+      mapcache_grid_compute_limits(igl->grid,extent,igl->grid_limits,tolerance);
+    }
     APR_ARRAY_PUSH(tileset->grid_links,mapcache_grid_link*) = gridlink;
   }
 
@@ -762,7 +837,7 @@ void parseTileset(mapcache_context *ctx, ezxml_t node, mapcache_cfg *config)
                      " but it is not configured", name, cur_node->txt);
       return;
     }
-    tileset->cache = cache;
+    tileset->_cache = cache;
   }
 
   if ((cur_node = ezxml_child(node,"source")) != NULL) {
@@ -910,6 +985,7 @@ void parseServices(mapcache_context *ctx, ezxml_t root, mapcache_cfg *config)
 }
 
 
+
 void mapcache_configuration_parse_xml(mapcache_context *ctx, const char *filename, mapcache_cfg *config)
 {
   ezxml_t doc, node;
@@ -1078,20 +1154,35 @@ void mapcache_configuration_parse_xml(mapcache_context *ctx, const char *filenam
     }
   }
 
-  if((node = ezxml_child(doc,"lock_dir")) != NULL) {
-    config->lockdir = apr_pstrdup(ctx->pool, node->txt);
+  if((node = ezxml_child(doc,"locker")) != NULL) {
+    mapcache_config_parse_locker(ctx,node,&config->locker);
+    GC_CHECK_ERROR(ctx);
   } else {
-    config->lockdir = apr_pstrdup(ctx->pool,"/tmp");
-  }
-
-  if((node = ezxml_child(doc,"lock_retry")) != NULL) {
-    char *endptr;
-    config->lock_retry_interval = (unsigned int)strtol(node->txt,&endptr,10);
-    if(*endptr != 0 || config->lock_retry_interval < 0) {
-      ctx->set_error(ctx, 400, "failed to parse lock_retry microseconds \"%s\". Expecting a positive integer",
-                     node->txt);
-      return;
+    /* backwards compatibility */
+    int micro_retry;
+    mapcache_locker_disk *ldisk;
+    config->locker = mapcache_locker_disk_create(ctx);
+    ldisk = (mapcache_locker_disk*)config->locker;
+    if((node = ezxml_child(doc,"lock_dir")) != NULL) {
+      ldisk->dir = apr_pstrdup(ctx->pool, node->txt);
+    } else {
+      ldisk->dir = apr_pstrdup(ctx->pool,"/tmp");
     }
+
+    if((node = ezxml_child(doc,"lock_retry")) != NULL) {
+      char *endptr;
+      micro_retry = strtol(node->txt,&endptr,10);
+      if(*endptr != 0 || micro_retry <= 0) {
+        ctx->set_error(ctx, 400, "failed to parse lock_retry microseconds \"%s\". Expecting a positive integer",
+            node->txt);
+        return;
+      }
+    } else {
+      /* default retry interval is 1/100th of a second, i.e. 10000 microseconds */
+      micro_retry = 10000;
+    }
+    config->locker->retry_interval = micro_retry / 1000000.0;
+    config->locker->timeout=120;
   }
 
   if((node = ezxml_child(doc,"threaded_fetching")) != NULL) {
