@@ -196,10 +196,13 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
 {
   int expires = 0;
   mapcache_http_response *response;
-  int i,is_empty=1 /* response image is initially empty */;
   char *timestr;
-  mapcache_image *base=NULL;
-  mapcache_image_format *format = NULL;
+  mapcache_image *base;
+  mapcache_image_format *format;
+  mapcache_image_format_type t;
+  int i,is_empty=1; /* response image is initially empty */;
+  base=NULL;
+  format = NULL;
 
 #ifdef DEBUG
   if(req_tile->ntiles ==0) {
@@ -209,11 +212,22 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
 #endif
   response = mapcache_http_response_create(ctx->pool);
 
+  if(ctx->supports_redirects && req_tile->ntiles == 1) {
+    req_tile->tiles[0]->allow_redirect = 1;
+  }
+
 
   mapcache_prefetch_tiles(ctx,req_tile->tiles,req_tile->ntiles);
   if(GC_HAS_ERROR(ctx))
     return NULL;
 
+  if(req_tile->tiles[0]->redirect) {
+    response->code = 302;
+    apr_table_set(response->headers,"Location",req_tile->tiles[0]->redirect);
+    response->data = mapcache_buffer_create(0, ctx->pool);
+    return response;
+  }
+
   /* loop through tiles, and eventually merge them vertically together */
   for(i=0; i<req_tile->ntiles; i++) {
     mapcache_tile *tile = req_tile->tiles[i]; /* shortcut */
@@ -222,7 +236,7 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
     if(tile->expires && (tile->expires < expires || expires == 0)) {
       expires = tile->expires;
     }
-    
+
     if(tile->nodata) {
       /* treat the special case where the cache explicitely stated that the
        tile was empty, and we don't have any vertical merging to do */
@@ -233,8 +247,8 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
       }
       continue;
     }
-    
-    /* treat the most common case: 
+
+    /* treat the most common case:
      - we have a single tile request (i.e. isempty is true)
      - the cache returned the encoded image
      */
@@ -271,7 +285,7 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
       mapcache_image_merge(ctx, base, tile->raw_image);
     } else {
       /* we don't need to merge onto an existing tile and don't have access to the tile's encoded data.
-       * 
+       *
        * we don't encode the tile's raw image data just yet because we might need to merge another one on top
        * of it later.
        */
@@ -283,8 +297,8 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
   if(!response->data) {
     /* we need to encode the raw image data*/
     if(base) {
-      if(req_tile->format) {
-        format = req_tile->format;
+      if(req_tile->image_request.format) {
+        format = req_tile->image_request.format;
       } else {
         format = req_tile->tiles[0]->tileset->format;
         if(!format) {
@@ -296,20 +310,20 @@ mapcache_http_response *mapcache_core_get_tile(mapcache_context *ctx, mapcache_r
         return NULL;
       }
     } else {
+      unsigned char empty[5] = {'#',0,0,0,0};
 #ifdef DEBUG
       if(!is_empty) {
         ctx->set_error(ctx,500,"BUG: no image data to encode, but tile not marked as empty");
         return NULL;
       }
 #endif
-      unsigned char empty[5] = {'#',0,0,0,0};
-      response->data = mapcache_empty_png_decode(ctx,empty,&is_empty); /* is_empty is unchanged and left to 1 */
+      response->data = mapcache_empty_png_decode(ctx,req_tile->tiles[0]->grid_link->grid->tile_sx, req_tile->tiles[0]->grid_link->grid->tile_sy, empty,&is_empty); /* is_empty is unchanged and left to 1 */
       format = mapcache_configuration_get_image_format(ctx->config,"PNG8");
     }
   }
-  
+
   /* compute the content-type */
-  mapcache_image_format_type t = mapcache_imageio_header_sniff(ctx,response->data);
+  t = mapcache_imageio_header_sniff(ctx,response->data);
   if(t == GC_PNG)
     apr_table_set(response->headers,"Content-Type","image/png");
   else if(t == GC_JPEG)
@@ -334,15 +348,18 @@ mapcache_map* mapcache_assemble_maps(mapcache_context *ctx, mapcache_map **maps,
   mapcache_tile ***maptiles;
   int *nmaptiles;
   mapcache_tile **tiles;
+  mapcache_grid_link **effectively_used_grid_links;
   mapcache_map *basemap = NULL;
   int ntiles = 0;
   int i;
   maptiles = apr_pcalloc(ctx->pool,nmaps*sizeof(mapcache_tile**));
   nmaptiles = apr_pcalloc(ctx->pool,nmaps*sizeof(int));
+  effectively_used_grid_links = apr_pcalloc(ctx->pool,nmaps*sizeof(mapcache_grid_link*));
   for(i=0; i<nmaps; i++) {
     mapcache_tileset_get_map_tiles(ctx,maps[i]->tileset,maps[i]->grid_link,
                                    &maps[i]->extent, maps[i]->width, maps[i]->height,
-                                   &(nmaptiles[i]), &(maptiles[i]));
+                                   &(nmaptiles[i]), &(maptiles[i]), &(effectively_used_grid_links[i]));
+    if(GC_HAS_ERROR(ctx)) return NULL;
     ntiles += nmaptiles[i];
   }
   tiles = apr_pcalloc(ctx->pool,ntiles * sizeof(mapcache_tile*));
@@ -379,7 +396,7 @@ mapcache_map* mapcache_assemble_maps(mapcache_context *ctx, mapcache_map **maps,
       }
     }
     if(hasdata) {
-      maps[i]->raw_image = mapcache_tileset_assemble_map_tiles(ctx,maps[i]->tileset,maps[i]->grid_link,
+      maps[i]->raw_image = mapcache_tileset_assemble_map_tiles(ctx,maps[i]->tileset,effectively_used_grid_links[i],
                            &maps[i]->extent, maps[i]->width, maps[i]->height,
                            nmaptiles[i], maptiles[i],
                            mode);
@@ -468,7 +485,7 @@ mapcache_http_response *mapcache_core_get_map(mapcache_context *ctx, mapcache_re
   }
 
   if(basemap->raw_image) {
-    format = req_map->getmap_format; /* always defined, defaults to JPEG */
+    format = req_map->image_request.format; /* always defined, defaults to JPEG */
     response->data = format->write(ctx,basemap->raw_image,format);
     if(GC_HAS_ERROR(ctx)) {
       return NULL;
@@ -516,16 +533,21 @@ mapcache_http_response *mapcache_core_proxy_request(mapcache_context *ctx, mapca
   mapcache_http *http;
   mapcache_http_response *response = mapcache_http_response_create(ctx->pool);
   response->data = mapcache_buffer_create(30000,ctx->pool);
-  http = req_proxy->http;
+  http = mapcache_http_clone(ctx, req_proxy->rule->http);
   if(req_proxy->pathinfo) {
-    http = mapcache_http_clone(ctx,http);
     if( (*(req_proxy->pathinfo)) == '/' ||
         http->url[strlen(http->url)-1] == '/')
       http->url = apr_pstrcat(ctx->pool,http->url,req_proxy->pathinfo,NULL);
     else
       http->url = apr_pstrcat(ctx->pool,http->url,"/",req_proxy->pathinfo,NULL);
   }
-  mapcache_http_do_request_with_params(ctx,http,req_proxy->params,response->data,response->headers,&response->code);
+  http->url = mapcache_http_build_url(ctx,http->url,req_proxy->params);
+  http->post_body = req_proxy->post_buf;
+  http->post_len = req_proxy->post_len;
+  if(req_proxy->headers) {
+    apr_table_overlap(http->headers, req_proxy->headers, APR_OVERLAP_TABLES_SET);
+  }
+  mapcache_http_do_request(ctx,http, response->data,response->headers,&response->code);
   if(response->code !=0 && GC_HAS_ERROR(ctx)) {
     /* the http request was successful, but the server returned an error */
     ctx->clear_errors(ctx);
