@@ -35,14 +35,27 @@
 #include <time.h>
 #ifdef USE_SQLITE
 #include <sqlite3.h>
-#include <apr_reslist.h>
-#include <apr_hash.h>
-#ifdef APR_HAS_THREADS
-#include <apr_thread_mutex.h>
-#endif
+#include <float.h>
 #endif
 
+#ifndef HAVE_TIMEGM
+time_t timegm(struct tm *tm)
+{
+  time_t t, tdiff;
+  struct tm in, gtime, ltime;
 
+  memcpy(&in, tm, sizeof(in));
+  t = mktime(&in);
+
+  memcpy(&gtime, gmtime(&t), sizeof(gtime));
+  memcpy(&ltime, localtime(&t), sizeof(ltime));
+  gtime.tm_isdst = ltime.tm_isdst;
+  tdiff = t - mktime(&gtime);
+
+  memcpy(&in, tm, sizeof(in));
+  return mktime(&in) + tdiff;
+}
+#endif
 
 static int _mapcache_dimension_intervals_validate(mapcache_context *ctx, mapcache_dimension *dim, char **value)
 {
@@ -69,16 +82,15 @@ static int _mapcache_dimension_intervals_validate(mapcache_context *ctx, mapcach
   return MAPCACHE_FAILURE;
 }
 
-static const char** _mapcache_dimension_intervals_print(mapcache_context *ctx, mapcache_dimension *dim)
+static apr_array_header_t* _mapcache_dimension_intervals_print(mapcache_context *ctx, mapcache_dimension *dim)
 {
   mapcache_dimension_intervals *dimension = (mapcache_dimension_intervals*)dim;
-  const char **ret = (const char**)apr_pcalloc(ctx->pool,(dimension->nintervals+1)*sizeof(const char*));
+  apr_array_header_t *ret = apr_array_make(ctx->pool,dimension->nintervals,sizeof(char*));
   int i;
   for(i=0; i<dimension->nintervals; i++) {
     mapcache_interval *interval = &dimension->intervals[i];
-    ret[i] = apr_psprintf(ctx->pool,"%g/%g/%g",interval->start,interval->end,interval->resolution);
+    APR_ARRAY_PUSH(ret,char*) = apr_psprintf(ctx->pool,"%g/%g/%g",interval->start,interval->end,interval->resolution);
   }
-  ret[i]=NULL;
   return ret;
 }
 
@@ -149,12 +161,11 @@ static int _mapcache_dimension_regex_validate(mapcache_context *ctx, mapcache_di
   return MAPCACHE_FAILURE;
 }
 
-static const char** _mapcache_dimension_regex_print(mapcache_context *ctx, mapcache_dimension *dim)
+static apr_array_header_t* _mapcache_dimension_regex_print(mapcache_context *ctx, mapcache_dimension *dim)
 {
   mapcache_dimension_regex *dimension = (mapcache_dimension_regex*)dim;
-  const char **ret = (const char**)apr_pcalloc(ctx->pool,2*sizeof(const char*));
-  ret[0]=dimension->regex_string;
-  ret[1]=NULL;
+  apr_array_header_t *ret = apr_array_make(ctx->pool,1,sizeof(char*));
+  APR_ARRAY_PUSH(ret,char*) = apr_pstrdup(ctx->pool,dimension->regex_string);
   return ret;
 }
 
@@ -212,15 +223,14 @@ static int _mapcache_dimension_values_validate(mapcache_context *ctx, mapcache_d
   return MAPCACHE_FAILURE;
 }
 
-static const char** _mapcache_dimension_values_print(mapcache_context *ctx, mapcache_dimension *dim)
+static apr_array_header_t* _mapcache_dimension_values_print(mapcache_context *ctx, mapcache_dimension *dim)
 {
   mapcache_dimension_values *dimension = (mapcache_dimension_values*)dim;
-  const char **ret = (const char**)apr_pcalloc(ctx->pool,(dimension->nvalues+1)*sizeof(const char*));
+  apr_array_header_t *ret = apr_array_make(ctx->pool,dimension->nvalues,sizeof(char*));
   int i;
   for(i=0; i<dimension->nvalues; i++) {
-    ret[i] = dimension->values[i];
+    APR_ARRAY_PUSH(ret,char*) = apr_pstrdup(ctx->pool,dimension->values[i]);
   }
-  ret[i]=NULL;
   return ret;
 }
 
@@ -308,105 +318,202 @@ mapcache_dimension* mapcache_dimension_regex_create(apr_pool_t *pool)
 
 #ifdef USE_SQLITE
 
-static apr_hash_t *time_connection_pools = NULL;
-
-struct sqlite_time_conn {
+struct sqlite_dimension_conn {
   sqlite3 *handle;
-  sqlite3_stmt *prepared_statement;
-  char *errmsg;
+  sqlite3_stmt **prepared_statements;
+  int n_statements;
 };
 
-static apr_status_t _sqlite_time_reslist_get_ro_connection(void **conn_, void *params, apr_pool_t *pool)
+void mapcache_sqlite_dimension_connection_constructor(mapcache_context *ctx, void **conn_, void *params)
 {
   int ret;
-  int flags;  
-  mapcache_timedimension_sqlite *dim = (mapcache_timedimension_sqlite*) params;
-  struct sqlite_time_conn *conn = apr_pcalloc(pool, sizeof (struct sqlite_time_conn));
+  int flags;
+  char *dbfile = (char*) params;
+  struct sqlite_dimension_conn *conn = calloc(1, sizeof (struct sqlite_dimension_conn));
   *conn_ = conn;
   flags = SQLITE_OPEN_READONLY | SQLITE_OPEN_NOMUTEX;
-  ret = sqlite3_open_v2(dim->dbfile, &conn->handle, flags, NULL);
-  
+  ret = sqlite3_open_v2(dbfile, &conn->handle, flags, NULL);
+
   if (ret != SQLITE_OK) {
-    return APR_EGENERAL;
+    ctx->set_error(ctx,500,"failed to open sqlite dimension dbfile (%s): %s",dbfile,sqlite3_errmsg(conn->handle));
+    sqlite3_close(conn->handle);
+    *conn_=NULL;
+    return;
   }
   sqlite3_busy_timeout(conn->handle, 300000);
-  return APR_SUCCESS;
 }
 
-static apr_status_t _sqlite_time_reslist_free_connection(void *conn_, void *params, apr_pool_t *pool)
+void mapcache_sqlite_dimension_connection_destructor(void *conn_)
 {
-  struct sqlite_time_conn *conn = (struct sqlite_time_conn*) conn_;
-  if(conn->prepared_statement) {
-    sqlite3_finalize(conn->prepared_statement);
+  struct sqlite_dimension_conn *conn = (struct sqlite_dimension_conn*) conn_;
+  while(conn->n_statements) {
+    conn->n_statements--;
+    if(conn->prepared_statements[conn->n_statements]) {
+      sqlite3_finalize(conn->prepared_statements[conn->n_statements]);
+    }
   }
+  free(conn->prepared_statements);
   sqlite3_close(conn->handle);
-  return APR_SUCCESS;
+  free(conn);
 }
 
-static struct sqlite_time_conn* _sqlite_time_get_conn(mapcache_context *ctx, mapcache_timedimension_sqlite *dim) {
-  apr_status_t rv;
-  struct sqlite_time_conn *conn = NULL;
-  apr_reslist_t *pool = NULL;
-  if(!time_connection_pools || NULL == (pool = apr_hash_get(time_connection_pools,dim->timedimension.key, APR_HASH_KEY_STRING)) ) {
-    
-#ifdef APR_HAS_THREADS
-    if(ctx->threadlock)
-      apr_thread_mutex_lock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-    
-    if(!time_connection_pools) {
-      time_connection_pools = apr_hash_make(ctx->process_pool);
-    }
-
-    /* probably doesn't exist, unless the previous mutex locked us, so we check */
-    pool = apr_hash_get(time_connection_pools,dim->timedimension.key, APR_HASH_KEY_STRING);
-    if(!pool) {
-      /* there where no existing connection pools, create them*/
-      rv = apr_reslist_create(&pool,
-                              0 /* min */,
-                              10 /* soft max */,
-                              200 /* hard max */,
-                              60*1000000 /*60 seconds, ttl*/,
-                              _sqlite_time_reslist_get_ro_connection, /* resource constructor */
-                              _sqlite_time_reslist_free_connection, /* resource destructor */
-                              dim, ctx->process_pool);
-      if(rv != APR_SUCCESS) {
-        ctx->set_error(ctx,500,"failed to create sqlite time connection pool");
-#ifdef APR_HAS_THREADS
-        if(ctx->threadlock)
-          apr_thread_mutex_unlock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-        return NULL;
-      }
-      apr_hash_set(time_connection_pools,dim->timedimension.key,APR_HASH_KEY_STRING,pool);
-    }
-#ifdef APR_HAS_THREADS
-    if(ctx->threadlock)
-      apr_thread_mutex_unlock((apr_thread_mutex_t*)ctx->threadlock);
-#endif
-    pool = apr_hash_get(time_connection_pools,dim->timedimension.key, APR_HASH_KEY_STRING);
-    assert(pool);
-  }
-  rv = apr_reslist_acquire(pool, (void **) &conn);
-  if (rv != APR_SUCCESS) {
-    ctx->set_error(ctx, 500, "failed to aquire connection to time dimension sqlite backend: %s", (conn && conn->errmsg)?conn->errmsg:"unknown error");
-    return NULL;
-  }
-  return conn;
+static mapcache_pooled_connection* _sqlite_time_dimension_get_conn(mapcache_context *ctx, mapcache_timedimension_sqlite *dim) {
+  mapcache_pooled_connection *pc = mapcache_connection_pool_get_connection(ctx,dim->timedimension.key,
+        mapcache_sqlite_dimension_connection_constructor,
+        mapcache_sqlite_dimension_connection_destructor, dim->dbfile);
+  return pc;
+}
+static mapcache_pooled_connection* _sqlite_dimension_get_conn(mapcache_context *ctx, mapcache_dimension_sqlite *dim) {
+  mapcache_pooled_connection *pc = mapcache_connection_pool_get_connection(ctx,dim->dimension.name,
+        mapcache_sqlite_dimension_connection_constructor,
+        mapcache_sqlite_dimension_connection_destructor, dim->dbfile);
+  return pc;
 }
 
-static void _sqlite_time_release_conn(mapcache_context *ctx, mapcache_timedimension_sqlite *sdim, struct sqlite_time_conn *conn)
+static void _sqlite_dimension_release_conn(mapcache_context *ctx, mapcache_pooled_connection *pc)
 {
-  apr_reslist_t *pool;
-  pool = apr_hash_get(time_connection_pools,sdim->timedimension.key, APR_HASH_KEY_STRING);
-
-  if (GC_HAS_ERROR(ctx)) {
-    apr_reslist_invalidate(pool, (void*) conn);
+  if(GC_HAS_ERROR(ctx)) {
+    mapcache_connection_pool_invalidate_connection(ctx,pc);
   } else {
-    apr_reslist_release(pool, (void*) conn);
+    mapcache_connection_pool_release_connection(ctx,pc);
   }
 }
 
+
+static int _mapcache_dimension_sqlite_validate(mapcache_context *ctx, mapcache_dimension *dim, char **value)
+{
+  mapcache_dimension_sqlite *dimension = (mapcache_dimension_sqlite*)dim;
+  struct sqlite_dimension_conn *conn = NULL;
+  int sqliteret,paramidx,ret=MAPCACHE_FAILURE;
+  mapcache_pooled_connection *pc;
+  pc = _sqlite_dimension_get_conn(ctx,dimension);
+  if (GC_HAS_ERROR(ctx)) {
+    return ret;
+  }
+  conn = pc->connection;
+  if(!conn->prepared_statements) {
+    conn->prepared_statements = calloc(2,sizeof(sqlite3_stmt*));
+    conn->n_statements = 2;
+  }
+  if(!conn->prepared_statements[0]) {
+    sqliteret = sqlite3_prepare_v2(conn->handle, dimension->validate_query, -1, &conn->prepared_statements[0], NULL);
+    if(sqliteret != SQLITE_OK) {
+      ctx->set_error(ctx, 500, "sqlite dimension backend failed on preparing query: %s", sqlite3_errmsg(conn->handle));
+      goto cleanup;
+    }
+  }
+
+  paramidx = sqlite3_bind_parameter_index(conn->prepared_statements[0], ":dim");
+  if (paramidx) {
+    sqliteret = sqlite3_bind_text(conn->prepared_statements[0], paramidx, *value, -1, SQLITE_STATIC);
+    if(sqliteret != SQLITE_OK) {
+      ctx->set_error(ctx,400, "sqlite dimension failed to bind :dim : %s", sqlite3_errmsg(conn->handle));
+      goto cleanup;
+    }
+  }
+  do {
+    sqliteret = sqlite3_step(conn->prepared_statements[0]);
+    if (sqliteret != SQLITE_DONE && sqliteret != SQLITE_ROW && sqliteret != SQLITE_BUSY && sqliteret != SQLITE_LOCKED) {
+      ctx->set_error(ctx, 500, "sqlite dimension backend failed on query : %s (%d)", sqlite3_errmsg(conn->handle), sqliteret);
+      goto cleanup;
+    }
+    if(sqliteret == SQLITE_ROW) {
+      const char* dim_modified = (const char*) sqlite3_column_text(conn->prepared_statements[0], 0);
+      if(strcmp(dim_modified, *value)) {
+        *value = apr_pstrdup(ctx->pool, dim_modified);
+      }
+      ret = MAPCACHE_SUCCESS;
+    }
+  } while (sqliteret == SQLITE_ROW || sqliteret == SQLITE_BUSY || sqliteret == SQLITE_LOCKED);
+
+cleanup:
+  if(conn->prepared_statements[0]) {
+    sqlite3_reset(conn->prepared_statements[0]);
+  }
+  _sqlite_dimension_release_conn(ctx,pc);
+
+  return ret;
+}
+
+static apr_array_header_t* _mapcache_dimension_sqlite_print(mapcache_context *ctx, mapcache_dimension *dim)
+{
+  mapcache_dimension_sqlite *dimension = (mapcache_dimension_sqlite*)dim;
+  struct sqlite_dimension_conn *conn = NULL;
+  int sqliteret;
+  apr_array_header_t *ret = apr_array_make(ctx->pool,0,sizeof(char*));
+  mapcache_pooled_connection *pc;
+  pc = _sqlite_dimension_get_conn(ctx,dimension);
+  if (GC_HAS_ERROR(ctx)) {
+    return ret;
+  }
+  conn = pc->connection;
+  if(!conn->prepared_statements) {
+    conn->prepared_statements = calloc(2,sizeof(sqlite3_stmt*));
+    conn->n_statements = 2;
+  }
+
+  if(!conn->prepared_statements[1]) {
+    sqliteret = sqlite3_prepare_v2(conn->handle, dimension->list_query, -1, &conn->prepared_statements[1], NULL);
+    if(sqliteret != SQLITE_OK) {
+      ctx->set_error(ctx, 500, "sqlite dimension backend failed on preparing query: %s", sqlite3_errmsg(conn->handle));
+      goto cleanup;
+    }
+  }
+  do {
+    sqliteret = sqlite3_step(conn->prepared_statements[1]);
+    if (sqliteret != SQLITE_DONE && sqliteret != SQLITE_ROW && sqliteret != SQLITE_BUSY && sqliteret != SQLITE_LOCKED) {
+      ctx->set_error(ctx, 500, "sqlite dimension backend failed on query : %s (%d)", sqlite3_errmsg(conn->handle), sqliteret);
+      goto cleanup;
+    }
+    if(sqliteret == SQLITE_ROW) {
+      const char* sqdim = (const char*) sqlite3_column_text(conn->prepared_statements[1], 0);
+      APR_ARRAY_PUSH(ret,char*) = apr_pstrdup(ctx->pool,sqdim);
+    }
+  } while (sqliteret == SQLITE_ROW || sqliteret == SQLITE_BUSY || sqliteret == SQLITE_LOCKED);
+
+cleanup:
+  if(conn->prepared_statements[1]) {
+    sqlite3_reset(conn->prepared_statements[1]);
+  }
+  _sqlite_dimension_release_conn(ctx,pc);
+
+  return ret;
+}
+
+
+static void _mapcache_dimension_sqlite_parse_xml(mapcache_context *ctx, mapcache_dimension *dim,
+    ezxml_t node)
+{
+  mapcache_dimension_sqlite *dimension;
+  ezxml_t child;
+
+  dimension = (mapcache_dimension_sqlite*)dim;
+
+  child = ezxml_child(node,"dbfile");
+  if(child) {
+    dimension->dbfile = apr_pstrdup(ctx->pool, child->txt);
+  } else {
+    ctx->set_error(ctx,400,"sqlite dimension \"%s\" has no <dbfile> node", dim->name);
+    return;
+  }
+  child = ezxml_child(node,"validate_query");
+  if(child) {
+    dimension->validate_query = apr_pstrdup(ctx->pool, child->txt);
+  } else {
+    ctx->set_error(ctx,400,"sqlite dimension \"%s\" has no <validate_query> node", dim->name);
+    return;
+  }
+  child = ezxml_child(node,"list_query");
+  if(child) {
+    dimension->list_query = apr_pstrdup(ctx->pool, child->txt);
+  } else {
+    ctx->set_error(ctx,400,"sqlite dimension \"%s\" has no <list_query> node", dim->name);
+    return;
+  }
+
+}
+
+
 static void _bind_sqlite_timedimension_params(mapcache_context *ctx, sqlite3_stmt *stmt,
         sqlite3 *handle, mapcache_tileset *tileset, mapcache_grid *grid, mapcache_extent *extent,
         time_t start, time_t end)
@@ -432,38 +539,36 @@ static void _bind_sqlite_timedimension_params(mapcache_context *ctx, sqlite3_stm
     }
   }
 
-  if(extent) {
-    paramidx = sqlite3_bind_parameter_index(stmt, ":minx");
-    if (paramidx) {
-      ret = sqlite3_bind_double(stmt, paramidx, extent->minx);
-      if(ret != SQLITE_OK) {
-        ctx->set_error(ctx,400, "failed to bind :minx %s", sqlite3_errmsg(handle));
-        return;
-      }
+  paramidx = sqlite3_bind_parameter_index(stmt, ":minx");
+  if (paramidx) {
+    ret = sqlite3_bind_double(stmt, paramidx, extent?extent->minx:-DBL_MAX);
+    if(ret != SQLITE_OK) {
+      ctx->set_error(ctx,400, "failed to bind :minx %s", sqlite3_errmsg(handle));
+      return;
     }
-    paramidx = sqlite3_bind_parameter_index(stmt, ":miny");
-    if (paramidx) {
-      ret = sqlite3_bind_double(stmt, paramidx, extent->miny);
-      if(ret != SQLITE_OK) {
-        ctx->set_error(ctx,400, "failed to bind :miny %s", sqlite3_errmsg(handle));
-        return;
-      }
+  }
+  paramidx = sqlite3_bind_parameter_index(stmt, ":miny");
+  if (paramidx) {
+    ret = sqlite3_bind_double(stmt, paramidx, extent?extent->miny:-DBL_MAX);
+    if(ret != SQLITE_OK) {
+      ctx->set_error(ctx,400, "failed to bind :miny %s", sqlite3_errmsg(handle));
+      return;
     }
-    paramidx = sqlite3_bind_parameter_index(stmt, ":maxx");
-    if (paramidx) {
-      ret = sqlite3_bind_double(stmt, paramidx, extent->maxx);
-      if(ret != SQLITE_OK) {
-        ctx->set_error(ctx,400, "failed to bind :maxx %s", sqlite3_errmsg(handle));
-        return;
-      }
+  }
+  paramidx = sqlite3_bind_parameter_index(stmt, ":maxx");
+  if (paramidx) {
+    ret = sqlite3_bind_double(stmt, paramidx, extent?extent->maxx:DBL_MAX);
+    if(ret != SQLITE_OK) {
+      ctx->set_error(ctx,400, "failed to bind :maxx %s", sqlite3_errmsg(handle));
+      return;
     }
-    paramidx = sqlite3_bind_parameter_index(stmt, ":maxy");
-    if (paramidx) {
-      ret = sqlite3_bind_double(stmt, paramidx, extent->maxy);
-      if(ret != SQLITE_OK) {
-        ctx->set_error(ctx,400, "failed to bind :maxy %s", sqlite3_errmsg(handle));
-        return;
-      }
+  }
+  paramidx = sqlite3_bind_parameter_index(stmt, ":maxy");
+  if (paramidx) {
+    ret = sqlite3_bind_double(stmt, paramidx, extent?extent->maxy:DBL_MAX);
+    if(ret != SQLITE_OK) {
+      ctx->set_error(ctx,400, "failed to bind :maxy %s", sqlite3_errmsg(handle));
+      return;
     }
   }
 
@@ -490,47 +595,48 @@ apr_array_header_t *_mapcache_timedimension_sqlite_get_entries(mapcache_context
         mapcache_tileset *tileset, mapcache_grid *grid, mapcache_extent *extent, time_t start, time_t end) {
   mapcache_timedimension_sqlite *sdim = (mapcache_timedimension_sqlite*)dim;
   int ret;
-  sqlite3_stmt *stmt;
   apr_array_header_t *time_ids = NULL;
-  struct sqlite_time_conn *conn = _sqlite_time_get_conn(ctx, sdim);
+  mapcache_pooled_connection *pc;
+  struct sqlite_dimension_conn *conn;
+  pc = _sqlite_time_dimension_get_conn(ctx,sdim);
   if (GC_HAS_ERROR(ctx)) {
-    if(conn) _sqlite_time_release_conn(ctx, sdim, conn);
     return NULL;
   }
-  stmt = conn->prepared_statement;
-  if(!stmt) {
-    ret = sqlite3_prepare_v2(conn->handle, sdim->query, -1, &conn->prepared_statement, NULL);
+  conn = pc->connection;
+  if(!conn->prepared_statements) {
+    conn->prepared_statements = calloc(1,sizeof(sqlite3_stmt*));
+    conn->n_statements = 1;
+  }
+  if(!conn->prepared_statements[0]) {
+    ret = sqlite3_prepare_v2(conn->handle, sdim->query, -1, &conn->prepared_statements[0], NULL);
     if(ret != SQLITE_OK) {
       ctx->set_error(ctx, 500, "time sqlite backend failed on preparing query: %s", sqlite3_errmsg(conn->handle));
-      _sqlite_time_release_conn(ctx, sdim, conn);
+      _sqlite_dimension_release_conn(ctx, pc);
       return NULL;
     }
-    stmt = conn->prepared_statement;
   }
-  
-  _bind_sqlite_timedimension_params(ctx,stmt,conn->handle,tileset,grid,extent,start,end);
+
+  _bind_sqlite_timedimension_params(ctx,conn->prepared_statements[0],conn->handle,tileset,grid,extent,start,end);
   if(GC_HAS_ERROR(ctx)) {
-    sqlite3_reset(stmt);
-    _sqlite_time_release_conn(ctx, sdim, conn);
+    _sqlite_dimension_release_conn(ctx, pc);
     return NULL;
   }
-  
+
   time_ids = apr_array_make(ctx->pool,0,sizeof(char*));
   do {
-    ret = sqlite3_step(stmt);
+    ret = sqlite3_step(conn->prepared_statements[0]);
     if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_BUSY && ret != SQLITE_LOCKED) {
       ctx->set_error(ctx, 500, "sqlite backend failed on timedimension query : %s (%d)", sqlite3_errmsg(conn->handle), ret);
-      sqlite3_reset(stmt);
-      _sqlite_time_release_conn(ctx, sdim, conn);
+      _sqlite_dimension_release_conn(ctx, pc);
       return NULL;
     }
     if(ret == SQLITE_ROW) {
-      const char* time_id = (const char*) sqlite3_column_text(stmt, 0);
+      const char* time_id = (const char*) sqlite3_column_text(conn->prepared_statements[0], 0);
       APR_ARRAY_PUSH(time_ids,char*) = apr_pstrdup(ctx->pool,time_id);
     }
   } while (ret == SQLITE_ROW || ret == SQLITE_BUSY || ret == SQLITE_LOCKED);
-  sqlite3_reset(stmt);
-  _sqlite_time_release_conn(ctx, sdim, conn);
+  sqlite3_reset(conn->prepared_statements[0]);
+  _sqlite_dimension_release_conn(ctx, pc);
   return time_ids;
 }
 
@@ -554,7 +660,7 @@ typedef enum {
 void _mapcache_timedimension_sqlite_parse_xml(mapcache_context *ctx, mapcache_timedimension *dim, ezxml_t node) {
   mapcache_timedimension_sqlite *sdim = (mapcache_timedimension_sqlite*)dim;
   ezxml_t child;
-  
+
   child = ezxml_child(node,"dbfile");
   if(child && child->txt && *child->txt) {
     sdim->dbfile = apr_pstrdup(ctx->pool,child->txt);
@@ -573,8 +679,8 @@ void _mapcache_timedimension_sqlite_parse_xml(mapcache_context *ctx, mapcache_ti
 #endif
 
 char *mapcache_ogc_strptime(const char *value, struct tm *ts, mapcache_time_interval_t *ti) {
-  memset (ts, '\0', sizeof (*ts));
   char *valueptr;
+  memset (ts, '\0', sizeof (*ts));
   valueptr = strptime(value,"%Y-%m-%dT%H:%M:%SZ",ts);
   *ti = MAPCACHE_TINTERVAL_SECOND;
   if(valueptr) return valueptr;
@@ -611,10 +717,15 @@ apr_array_header_t* mapcache_timedimension_get_entries_for_value(mapcache_contex
     ctx->set_error(ctx,400,"failed to parse time %s",value);
     return NULL;
   }
-  
-  if(*valueptr == '/') {
+
+  if(*valueptr == '/' || (*valueptr == '-' && *(valueptr+1) == '-')) {
     /* we have a second (end) time */
-    valueptr++;
+    if (*valueptr == '/') {
+      valueptr++;
+    }
+    else {
+      valueptr += 2;
+    }
     valueptr = mapcache_ogc_strptime(valueptr,&tm_end,&tie);
     if(!valueptr) {
       ctx->set_error(ctx,400,"failed to parse end time in %s",value);
@@ -657,6 +768,22 @@ apr_array_header_t* mapcache_timedimension_get_entries_for_value(mapcache_contex
   /* end loop */
 }
 
+mapcache_dimension* mapcache_dimension_sqlite_create(apr_pool_t *pool)
+{
+#ifdef USE_SQLITE
+  mapcache_dimension_sqlite *dimension = apr_pcalloc(pool, sizeof(mapcache_dimension_sqlite));
+  dimension->dimension.type = MAPCACHE_DIMENSION_SQLITE;
+  dimension->dbfile = NULL;
+  dimension->dimension.validate = _mapcache_dimension_sqlite_validate;
+  dimension->dimension.configuration_parse_xml = _mapcache_dimension_sqlite_parse_xml;
+  dimension->dimension.print_ogc_formatted_values = _mapcache_dimension_sqlite_print;
+  return (mapcache_dimension*)dimension;
+#else
+  return NULL;
+#endif
+}
+
+
 #ifdef USE_SQLITE
 mapcache_timedimension* mapcache_timedimension_sqlite_create(apr_pool_t *pool) {
   mapcache_timedimension_sqlite *dim = apr_pcalloc(pool, sizeof(mapcache_timedimension_sqlite));
