@@ -85,6 +85,9 @@ int mapcache_image_has_alpha(mapcache_image *img)
 void mapcache_image_merge(mapcache_context *ctx, mapcache_image *base, mapcache_image *overlay)
 {
   int starti,startj;
+  pixman_image_t *si;
+  pixman_image_t *bi;
+  pixman_transform_t transform;
 #ifndef USE_PIXMAN
   int i,j;
   unsigned char *browptr, *orowptr, *bptr, *optr;
@@ -97,11 +100,10 @@ void mapcache_image_merge(mapcache_context *ctx, mapcache_image *base, mapcache_
   starti = (base->h - overlay->h)/2;
   startj = (base->w - overlay->w)/2;
 #ifdef USE_PIXMAN
-  pixman_image_t *si = pixman_image_create_bits(PIXMAN_a8r8g8b8,overlay->w,overlay->h,
+  si = pixman_image_create_bits(PIXMAN_a8r8g8b8,overlay->w,overlay->h,
                        (uint32_t*)overlay->data,overlay->stride);
-  pixman_image_t *bi = pixman_image_create_bits(PIXMAN_a8r8g8b8,base->w,base->h,
+  bi = pixman_image_create_bits(PIXMAN_a8r8g8b8,base->w,base->h,
                        (uint32_t*)base->data,base->stride);
-  pixman_transform_t transform;
   pixman_transform_init_translate(&transform,
                                   pixman_int_to_fixed(-startj),
                                   pixman_int_to_fixed(-starti));
@@ -359,5 +361,24 @@ int mapcache_image_blank_color(mapcache_image* image)
     return MAPCACHE_FALSE;
 }
 
+
+void mapcache_image_fill(mapcache_context *ctx, mapcache_image *image, const unsigned char *fill_color) {
+#if 0 && defined(USE_PIXMAN)
+  pixman_fill((uint32_t*)image->data, image->stride, 32, 0, 0, image->w, image->h, *((int*)fill_color) );
+#else
+  int r,c;
+  unsigned char *pixptr;
+  for(r=0;r<image->h;r++) {
+    pixptr = image->data + image->stride * r;
+    for(c=0;c<image->w;c++) {
+      pixptr[0]=fill_color[0];
+      pixptr[1]=fill_color[1];
+      pixptr[2]=fill_color[2];
+      pixptr[3]=fill_color[3];
+      pixptr+=4;
+    }
+  }
+#endif
+}
 /* vim: ts=2 sts=2 et sw=2
 */
