@@ -45,6 +45,9 @@ typedef unsigned long int uint64_t;
 #ifndef Z_BEST_COMPRESSION
 #define Z_BEST_COMPRESSION 9
 #endif
+#ifndef Z_NO_COMPRESSION
+#define Z_NO_COMPRESSION 0
+#endif
 
 /* Table of CRCs of all 8-bit messages. */
 #if __BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__
@@ -146,48 +149,152 @@ static unsigned long crc(unsigned char *buf, int len)
   return update_crc(0xffffffffL, buf, len) ^ 0xffffffffL;
 }
 
-static unsigned char empty_png[] = {
+static unsigned char empty_png_256[] = {
   0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,0x00,0x0d,0x49,0x48,0x44,0x52
   ,0x00,0x00,0x01,0x00,0x00,0x00,0x01,0x00,0x01,0x03,0x00,0x00,0x00,0x66,0xbc,0x3a
-  ,0x25,0x00,0x00,0x00,0x03,0x50,0x4c,0x54,0x45,0x73,0x91,0xad,0x31,0xf0,0x8f,0xdd
-  ,0x00,0x00,0x00,0x01,0x74,0x52,0x4e,0x53,0xff,0x6d,0xe4,0x37,0xeb,0x00,0x00,0x00
+  ,0x25,0x00,0x00,0x00,0x03,
+  /*PLTE*/0x50,0x4c,0x54,0x45,
+  /*RGB*/0x73,0x91,0xad,
+  /*PLTE-CRC*/0x31,0xf0,0x8f,0xdd
+  ,0x00,0x00,0x00,0x01,
+  /*tRNS*/0x74,0x52,0x4e,0x53,
+  /*A (fully opaque by default)*/0xff,
+  /*tRNS-CRC*/0x6d,0xe4,0x37,0xeb,
+  0x00,0x00,0x00
   ,0x1f,0x49,0x44,0x41,0x54,0x68,0xde,0xed,0xc1,0x01,0x0d,0x00,0x00,0x00,0xc2,0xa0
   ,0xf7,0x4f,0x6d,0x0e,0x37,0xa0,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xbe,0x0d
   ,0x21,0x00,0x00,0x01,0x7f,0x19,0x9c,0xa7,0x00,0x00,0x00,0x00,0x49,0x45,0x4e,0x44
   ,0xae,0x42,0x60,0x82
 };
+static size_t plte_offset_256= 0x25;
+static size_t trns_offset_256 = 0x34;
 
-static size_t plte_offset = 0x25;
-static size_t trns_offset = 0x34;
 
-mapcache_buffer* mapcache_empty_png_decode(mapcache_context *ctx, const unsigned char *hex_color, int *is_empty) {
+unsigned char empty_png_384[] = {
+  0x89, 0x50, 0x4e, 0x47, 0x0d, 0x0a, 0x1a, 0x0a, 0x00, 0x00, 0x00, 0x0d,
+  0x49, 0x48, 0x44, 0x52, 0x00, 0x00, 0x01, 0x80, 0x00, 0x00, 0x01, 0x80,
+  0x01, 0x03, 0x00, 0x00, 0x00, 0x9e, 0x09, 0x27, 0xfc, 0x00, 0x00, 0x00,
+  0x03, 
+  /*PLTE*/0x50,0x4c,0x54,0x45,
+  /*RGB*/0xcd,0x44,0x44,
+  /*PLTE-CRC*/0xdb,0xa9,0x37,0x41
+  ,0x00,0x00,0x00,0x01,
+  /*tRNS*/0x74,0x52,0x4e,0x53,
+  /*A (fully opaque by default)*/0xff,
+  /*tRNS-CRC*/0x6d,0xe4,0x37,0xeb,
+  0x00,0x00,0x00,
+  0x28, 0x49, 0x44, 0x41, 0x54, 0x78, 0x5e, 0xed,
+  0xc1, 0x01, 0x01, 0x00, 0x00, 0x00, 0x82, 0x20, 0xff, 0xaf, 0x6e, 0x48,
+  0x40, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
+  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x7c, 0x19, 0x49, 0x80, 0x00,
+  0x01, 0x28, 0x77, 0xfb, 0x61, 0x00, 0x00, 0x00, 0x00, 0x49, 0x45, 0x4e,
+  0x44, 0xae, 0x42, 0x60, 0x82
+};
+static size_t plte_offset_384= 0x25;
+static size_t trns_offset_384 = 0x34;
+
+
+static unsigned char empty_png_512[] = {
+  0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,0x00,0x0d,0x49,0x48,0x44,0x52
+  ,0x00,0x00,0x02,0x00,0x00,0x00,0x02,0x00,0x01,0x03,0x00,0x00,0x00,0xce,0xb6,0x46
+  ,0xb9,0x00,0x00,0x00,0x03,
+  /*PLTE*/0x50,0x4c,0x54,0x45,
+  /*RGB*/0xcd,0x44,0x44,
+  /*PLTE-CRC*/0xdb,0xa9,0x37,0x41
+  ,0x00,0x00,0x00,0x01,
+  /*tRNS*/0x74,0x52,0x4e,0x53,
+  /*A (fully opaque by default)*/0xff,
+  /*tRNS-CRC*/0x6d,0xe4,0x37,0xeb,
+  0x00,0x00,0x00
+  ,0x36,0x49,0x44,0x41,0x54,0x78,0x5e,0xed,0xc1,0x01,0x01,0x00,0x00,0x00,0x82,0x20
+  ,0xff,0xaf,0x6e,0x48,0x40,0x01,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
+  ,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00
+  ,0x00,0x00,0x00,0x00,0x00,0x7c,0x1b,0x82,0x00,0x00,0x01,0x9a,0x38,0x6a,0xc7,0x00
+  ,0x00,0x00,0x00,0x49,0x45,0x4e,0x44,0xae,0x42,0x60,0x82
+};
+static size_t plte_offset_512= 0x25;
+static size_t trns_offset_512 = 0x34;
+
+
+mapcache_buffer* mapcache_empty_png_decode(mapcache_context *ctx, int width, int height, const unsigned char *hex_color, int *is_empty) {
   int chunkcrc;
   unsigned char *dd;
-  mapcache_buffer *encoded_data = mapcache_buffer_create(sizeof(empty_png)+4,ctx->pool);
-  dd = encoded_data->buf;
-  memcpy(dd,empty_png,sizeof(empty_png));
-  dd[plte_offset+4] = hex_color[3]; //r
-  dd[plte_offset+5] = hex_color[2]; //g
-  dd[plte_offset+6] = hex_color[1]; //b
-  chunkcrc = crc(dd+plte_offset,7);
-  dd[plte_offset+7] = (unsigned char)((chunkcrc >> 24) & 0xff);
-  dd[plte_offset+8] = (unsigned char)((chunkcrc >> 16) & 0xff);
-  dd[plte_offset+9] = (unsigned char)((chunkcrc >> 8) & 0xff);
-  dd[plte_offset+10] = (unsigned char)(chunkcrc & 0xff);
-  if(hex_color[4] != 255) {
-    dd[trns_offset+4] = hex_color[4];
-    chunkcrc = crc(dd+trns_offset,5);
-    dd[trns_offset+5] = (unsigned char)((chunkcrc >> 24) & 0xff);
-    dd[trns_offset+6] = (unsigned char)((chunkcrc >> 16) & 0xff);
-    dd[trns_offset+7] = (unsigned char)((chunkcrc >> 8) & 0xff);
-    dd[trns_offset+8] = (unsigned char)(chunkcrc & 0xff);
+  mapcache_buffer *encoded_data;
+  if(width == 256 && height == 256) {
+    encoded_data = mapcache_buffer_create(sizeof(empty_png_256)+4,ctx->pool);
+    dd = encoded_data->buf;
+    memcpy(dd,empty_png_256,sizeof(empty_png_256));
+    dd[plte_offset_256+4] = hex_color[3]; //r
+    dd[plte_offset_256+5] = hex_color[2]; //g
+    dd[plte_offset_256+6] = hex_color[1]; //b
+    chunkcrc = crc(dd+plte_offset_256,7);
+    dd[plte_offset_256+7] = (unsigned char)((chunkcrc >> 24) & 0xff);
+    dd[plte_offset_256+8] = (unsigned char)((chunkcrc >> 16) & 0xff);
+    dd[plte_offset_256+9] = (unsigned char)((chunkcrc >> 8) & 0xff);
+    dd[plte_offset_256+10] = (unsigned char)(chunkcrc & 0xff);
+    if(hex_color[4] != 255) {
+      dd[trns_offset_256+4] = hex_color[4];
+      chunkcrc = crc(dd+trns_offset_256,5);
+      dd[trns_offset_256+5] = (unsigned char)((chunkcrc >> 24) & 0xff);
+      dd[trns_offset_256+6] = (unsigned char)((chunkcrc >> 16) & 0xff);
+      dd[trns_offset_256+7] = (unsigned char)((chunkcrc >> 8) & 0xff);
+      dd[trns_offset_256+8] = (unsigned char)(chunkcrc & 0xff);
+    }
+    encoded_data->size = sizeof(empty_png_256);
+  } else if(width == 384 && height == 384) {
+    encoded_data = mapcache_buffer_create(sizeof(empty_png_384)+4,ctx->pool);
+    dd = encoded_data->buf;
+    memcpy(dd,empty_png_384,sizeof(empty_png_384));
+    dd[plte_offset_384+4] = hex_color[3]; //r
+    dd[plte_offset_384+5] = hex_color[2]; //g
+    dd[plte_offset_384+6] = hex_color[1]; //b
+    chunkcrc = crc(dd+plte_offset_384,7);
+    dd[plte_offset_384+7] = (unsigned char)((chunkcrc >> 24) & 0xff);
+    dd[plte_offset_384+8] = (unsigned char)((chunkcrc >> 16) & 0xff);
+    dd[plte_offset_384+9] = (unsigned char)((chunkcrc >> 8) & 0xff);
+    dd[plte_offset_384+10] = (unsigned char)(chunkcrc & 0xff);
+    if(hex_color[4] != 255) {
+      dd[trns_offset_384+4] = hex_color[4];
+      chunkcrc = crc(dd+trns_offset_384,5);
+      dd[trns_offset_384+5] = (unsigned char)((chunkcrc >> 24) & 0xff);
+      dd[trns_offset_384+6] = (unsigned char)((chunkcrc >> 16) & 0xff);
+      dd[trns_offset_384+7] = (unsigned char)((chunkcrc >> 8) & 0xff);
+      dd[trns_offset_384+8] = (unsigned char)(chunkcrc & 0xff);
+    }
+    encoded_data->size = sizeof(empty_png_384);
+  } else if(width == 512 && height == 512) {
+    encoded_data = mapcache_buffer_create(sizeof(empty_png_512)+4,ctx->pool);
+    dd = encoded_data->buf;
+    memcpy(dd,empty_png_512,sizeof(empty_png_512));
+    dd[plte_offset_512+4] = hex_color[3]; //r
+    dd[plte_offset_512+5] = hex_color[2]; //g
+    dd[plte_offset_512+6] = hex_color[1]; //b
+    chunkcrc = crc(dd+plte_offset_512,7);
+    dd[plte_offset_512+7] = (unsigned char)((chunkcrc >> 24) & 0xff);
+    dd[plte_offset_512+8] = (unsigned char)((chunkcrc >> 16) & 0xff);
+    dd[plte_offset_512+9] = (unsigned char)((chunkcrc >> 8) & 0xff);
+    dd[plte_offset_512+10] = (unsigned char)(chunkcrc & 0xff);
+    if(hex_color[4] != 255) {
+      dd[trns_offset_512+4] = hex_color[4];
+      chunkcrc = crc(dd+trns_offset_512,5);
+      dd[trns_offset_512+5] = (unsigned char)((chunkcrc >> 24) & 0xff);
+      dd[trns_offset_512+6] = (unsigned char)((chunkcrc >> 16) & 0xff);
+      dd[trns_offset_512+7] = (unsigned char)((chunkcrc >> 8) & 0xff);
+      dd[trns_offset_512+8] = (unsigned char)(chunkcrc & 0xff);
+    }
+    encoded_data->size = sizeof(empty_png_512);
+  } else {
+    /* here for compatibility reasons, although this should not be used in production as it is cpu-heavy */
+    mapcache_image *rgba = mapcache_image_create_with_data(ctx,width,height);
+    mapcache_image_format *format = mapcache_configuration_get_image_format(ctx->config,"PNG8");
+    mapcache_image_fill(ctx,rgba,hex_color+1);
+    encoded_data = format->write(ctx,rgba,format);
   }
   if(hex_color[4] == 0) {
     *is_empty = 1;
   } else {
     *is_empty = 0;
   }
-  encoded_data->size = sizeof(empty_png);
   return encoded_data;
 }
 
@@ -217,14 +324,10 @@ void _mapcache_imageio_png_flush_func(png_structp png_ptr)
   // do nothing
 }
 
-#ifndef _WIN32
-static inline int premultiply (int color,int alpha)
-#else
-static __inline int premultiply (int color,int alpha)
-#endif
-{
-  int temp = (alpha * color) + 0x80;
-  return ((temp + (temp >> 8)) >> 8);
+#define PREMULTIPLY(out,color,alpha)\
+{\
+  int temp = ((alpha) * (color)) + 0x80;\
+  out =((temp + (temp >> 8)) >> 8);\
 }
 
 
@@ -318,9 +421,9 @@ void _mapcache_imageio_png_decode_to_image(mapcache_context *ctx, mapcache_buffe
         pixptr[1] = 0;
         pixptr[2] = 0;
       } else {
-        pixptr[0] = premultiply(pixel[2],alpha);
-        pixptr[1] = premultiply(pixel[1],alpha);
-        pixptr[2] = premultiply(pixel[0],alpha);
+        PREMULTIPLY(pixptr[0],pixel[2],alpha);
+        PREMULTIPLY(pixptr[1],pixel[1],alpha);
+        PREMULTIPLY(pixptr[2],pixel[0],alpha);
       }
       pixptr += 4;
     }
@@ -410,6 +513,9 @@ mapcache_buffer* _mapcache_imageio_png_encode(mapcache_context *ctx, mapcache_im
     png_set_compression_level (png_ptr, Z_BEST_COMPRESSION);
   else if(compression == MAPCACHE_COMPRESSION_FAST)
     png_set_compression_level (png_ptr, Z_BEST_SPEED);
+  else if(compression == MAPCACHE_COMPRESSION_DISABLE)
+    png_set_compression_level (png_ptr, Z_NO_COMPRESSION);
+    
   png_set_filter(png_ptr,0,PNG_FILTER_NONE);
 
   info_ptr = png_create_info_struct(png_ptr);
@@ -1314,6 +1420,8 @@ mapcache_buffer* _mapcache_imageio_png_q_encode( mapcache_context *ctx, mapcache
     png_set_compression_level (png_ptr, Z_BEST_COMPRESSION);
   else if(compression == MAPCACHE_COMPRESSION_FAST)
     png_set_compression_level (png_ptr, Z_BEST_SPEED);
+  else if(compression == MAPCACHE_COMPRESSION_DISABLE)
+    png_set_compression_level (png_ptr, Z_NO_COMPRESSION);
   png_set_filter(png_ptr,0,PNG_FILTER_NONE);
   info_ptr = png_create_info_struct(png_ptr);
   if (!info_ptr) {
@@ -1413,7 +1521,7 @@ mapcache_image_format* mapcache_imageio_create_png_q_format(apr_pool_t *pool, ch
   mapcache_image_format_png_q *format = apr_pcalloc(pool, sizeof(mapcache_image_format_png_q));
   format->format.format.name = name;
   format->format.format.extension = apr_pstrdup(pool,"png");
-  format->format.format.mime_type = apr_pstrdup(pool,"image/png");
+  format->format.format.mime_type = apr_pstrdup(pool,"image/png; mode=8bit");
   format->format.compression_level = compression;
   format->format.format.write = _mapcache_imageio_png_q_encode;
   format->format.format.create_empty_image = _mapcache_imageio_png_create_empty;
