@@ -200,7 +200,7 @@ void _create_capabilities_wms(mapcache_context *ctx, mapcache_request_get_capabi
     layerxml = ezxml_add_child(toplayer,"Layer",0);
     ezxml_set_attr(layerxml, "cascaded", "1");
     ezxml_set_attr(layerxml, "queryable", (tileset->source && tileset->source->info_formats)?"1":"0");
-    
+
     ezxml_set_txt(ezxml_add_child(layerxml,"Name",0),tileset->name);
     tsxml = ezxml_add_child(vendorxml, "TileSet",0);
 
@@ -228,8 +228,9 @@ void _create_capabilities_wms(mapcache_context *ctx, mapcache_request_get_capabi
 
     if(tileset->dimensions) {
       for(i=0; i<tileset->dimensions->nelts; i++) {
-        const char **value;
-        char *dimval;
+        apr_array_header_t *values;
+        int value_idx;
+        char *dimval = NULL;
         mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
         ezxml_t dimxml = ezxml_add_child(layerxml,"Dimension",0);
         ezxml_set_attr(dimxml,"name",dimension->name);
@@ -238,12 +239,14 @@ void _create_capabilities_wms(mapcache_context *ctx, mapcache_request_get_capabi
         if(dimension->unit) {
           ezxml_set_attr(dimxml,"units",dimension->unit);
         }
-        value = dimension->print_ogc_formatted_values(ctx,dimension);
-        dimval = apr_pstrdup(ctx->pool,*value);
-        value++;
-        while(*value) {
-          dimval = apr_pstrcat(ctx->pool,dimval,",",*value,NULL);
-          value++;
+        values = dimension->print_ogc_formatted_values(ctx,dimension);
+        for(value_idx=0;value_idx<values->nelts;value_idx++) {
+          char *idval = APR_ARRAY_IDX(values,value_idx,char*);
+          if(dimval) {
+            dimval = apr_pstrcat(ctx->pool,dimval,",",idval,NULL);
+          } else {
+            dimval = apr_pstrdup(ctx->pool,idval);
+          }
         }
         ezxml_set_txt(dimxml,dimval);
       }
@@ -483,6 +486,7 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
       mapcache_grid_link *main_grid_link = NULL;
       mapcache_tileset *main_tileset = NULL;
       mapcache_request_type type;
+      mapcache_image_format *imf;
 
       /* count the number of layers that are requested.
        * if we are in combined-mirror mode, then there is
@@ -545,24 +549,44 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
         type = MAPCACHE_REQUEST_GET_MAP;
       }
 
+      imf = wms_service->getmap_format;
+      if(wms_service->allow_format_override) {
+        str = apr_table_get(params,"FORMAT");
+        if(strcmp(str,imf->name) && strcmp(str,imf->mime_type)) {
+          apr_hash_index_t *hi;
+          for (hi = apr_hash_first(ctx->pool, ctx->config->image_formats); hi; hi = apr_hash_next(hi)) {
+            apr_hash_this(hi, NULL, NULL, (void**)&imf);
+            if(!strcmp(imf->name, str) || (imf->mime_type && !strcmp(imf->mime_type, str))) {
+              break;
+            }
+          }
+          if(!hi) { /* did not find any matching format for given mimetype or name */
+            errcode = 404;
+            errmsg = apr_psprintf(ctx->pool,"received wms request with invalid format %s", str);
+            goto proxies;
+          }
+        }
+      }
+
 
       if(type == MAPCACHE_REQUEST_GET_TILE) {
         tile_req = apr_pcalloc(ctx->pool, sizeof(mapcache_request_get_tile));
-        tile_req->request.type = MAPCACHE_REQUEST_GET_TILE;
         tile_req->tiles = apr_pcalloc(ctx->pool, count*sizeof(mapcache_tile*));
-        tile_req->format = wms_service->getmap_format;
+        tile_req->image_request.format = imf;
         *request = (mapcache_request*)tile_req;
+        (*request)->type = MAPCACHE_REQUEST_GET_TILE;
       } else {
         map_req = apr_pcalloc(ctx->pool, sizeof(mapcache_request_get_map));
-        map_req->request.type = MAPCACHE_REQUEST_GET_MAP;
         map_req->maps = apr_pcalloc(ctx->pool, count*sizeof(mapcache_map*));
         map_req->getmap_strategy = wms_service->getmap_strategy;
         map_req->resample_mode = wms_service->resample_mode;
-        map_req->getmap_format = wms_service->getmap_format;
+        map_req->image_request.format = imf;
         *request = (mapcache_request*)map_req;
+        (*request)->type = MAPCACHE_REQUEST_GET_MAP;
       }
       nallocated = count;
 
+
       /*
        * loop through all the layers to verify that they reference the requested grid,
        * and to extract any dimensions if configured
@@ -636,17 +660,36 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
           const char *value;
           if(tileset->dimensions) {
             for(i=0; i<tileset->dimensions->nelts; i++) {
+              char *dim_name;
               mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
-              if((value = (char*)apr_table_get(params,dimension->name)) != NULL) {
+              if(!strcasecmp(dimension->name,"TIME") || !strcasecmp(dimension->name,"ELEVATION")) {
+                dim_name = dimension->name;
+              } else {
+                dim_name = apr_pstrcat(ctx->pool, "dim_", dimension->name, NULL);
+              }
+              if((value = (char*)apr_table_get(params,dim_name)) == NULL) {
+                if(strcasecmp(dimension->name,"TIME") && strcasecmp(dimension->name,"ELEVATION")) {
+                  /* also test for the dimension without the DIM_ prefix if the latter was not found in the KVP params */
+                  dim_name = dimension->name;
+                  value = (char*)apr_table_get(params,dimension->name);
+                }
+              }
+
+              if(value) {
                 char *tmpval = apr_pstrdup(ctx->pool,value);
-                int ok = dimension->validate(ctx,dimension,&tmpval);
-                GC_CHECK_ERROR(ctx);
+                int ok;
+                if(dimension->skip_validation) {
+                  ok = MAPCACHE_SUCCESS;
+                } else {
+                  ok = dimension->validate(ctx,dimension,&tmpval);
+                  GC_CHECK_ERROR(ctx);
+                }
                 if(ok == MAPCACHE_SUCCESS)
                   apr_table_setn(dimtable,dimension->name,tmpval);
                 else {
                   errcode = 400;
                   errmsg = apr_psprintf(ctx->pool, "dimension \"%s\" value \"%s\" fails to validate",
-                                        dimension->name, value);
+                                        dim_name, value);
                   goto proxies;
                 }
               }
@@ -672,9 +715,9 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
             /* we need to create more tile/map entries */
               if(timedim_selected->nelts > 1) {
                 /* apr pools have no realloc */
+                mapcache_tile** tmptiles;
                 nallocated = nallocated + timedim_selected->nelts - 1;
-                mapcache_tile** tmptiles =
-                        apr_palloc(ctx->pool, nallocated * sizeof(mapcache_tile*));
+                tmptiles = apr_palloc(ctx->pool, nallocated * sizeof(mapcache_tile*));
                 for(i=0;i<tile_req->ntiles;i++) {
                   tmptiles[i] = tile_req->tiles[i];
                 }
@@ -695,9 +738,9 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
             /* we need to create more tile/map entries */
               if(timedim_selected->nelts > 1) {
                 /* apr pools have no realloc */
+                mapcache_map** tmpmaps;
                 nallocated = nallocated + timedim_selected->nelts - 1;
-                mapcache_map** tmpmaps =
-                        apr_palloc(ctx->pool, nallocated * sizeof(mapcache_map*));
+                tmpmaps = apr_palloc(ctx->pool, nallocated * sizeof(mapcache_map*));
                 for(i=0;i<map_req->nmaps;i++) {
                   tmpmaps[i] = map_req->maps[i];
                 }
@@ -713,7 +756,7 @@ void _mapcache_service_wms_parse_request(mapcache_context *ctx, mapcache_service
                 apr_table_set(map_req->maps[map_req->nmaps-1]->dimensions,tileset->timedimension->key,
                         APR_ARRAY_IDX(timedim_selected,i,char*));
               }
-              
+
             }
           }
         }
@@ -876,7 +919,7 @@ proxies:
         *request = (mapcache_request*)req_proxy;
         (*request)->service = this;
         (*request)->type = MAPCACHE_REQUEST_PROXY;
-        req_proxy->http = rule->http;
+        req_proxy->rule = rule;
         req_proxy->params = params;
         if(rule->append_pathinfo) {
           req_proxy->pathinfo = pathinfo;
@@ -916,32 +959,42 @@ void _configuration_parse_wms_xml(mapcache_context *ctx, ezxml_t node, mapcache_
 
   for( rule_node = ezxml_child(node,"forwarding_rule"); rule_node; rule_node = rule_node->next) {
     mapcache_forwarding_rule *rule;
-    ezxml_t http_node;
-    ezxml_t pathinfo_node;
-    ezxml_t param_node;
+    ezxml_t node;
     char *name = (char*)ezxml_attr(rule_node,"name");
     if(!name) name = "(null)";
     rule = apr_pcalloc(ctx->pool, sizeof(mapcache_forwarding_rule));
     rule->name = apr_pstrdup(ctx->pool,name);
     rule->match_params = apr_array_make(ctx->pool,1,sizeof(mapcache_dimension*));
+    rule->max_post_len = 10485760; /* 10 megabytes by default */
 
-    pathinfo_node = ezxml_child(rule_node,"append_pathinfo");
-    if(pathinfo_node && !strcasecmp(pathinfo_node->txt,"true")) {
+    node = ezxml_child(rule_node,"append_pathinfo");
+    if(node && !strcasecmp(node->txt,"true")) {
       rule->append_pathinfo = 1;
     } else {
       rule->append_pathinfo = 0;
     }
-    http_node = ezxml_child(rule_node,"http");
-    if(!http_node) {
+
+    node = ezxml_child(rule_node,"max_post_length");
+    if(node) {
+      char *endptr;
+      rule->max_post_len= (size_t)strtol(node->txt,&endptr,10);
+      if(*endptr != 0 || rule->max_post_len <= 0) {
+        ctx->set_error(ctx,500,"rule \"%s\" cannot have a negative or null <max_post_length>",name);
+        return;
+      }
+    }
+
+    node = ezxml_child(rule_node,"http");
+    if(!node) {
       ctx->set_error(ctx,500,"rule \"%s\" does not contain an <http> block",name);
       return;
     }
-    rule->http = mapcache_http_configuration_parse_xml(ctx,http_node);
+    rule->http = mapcache_http_configuration_parse_xml(ctx,node);
     GC_CHECK_ERROR(ctx);
 
-    for(param_node = ezxml_child(rule_node,"param"); param_node; param_node = param_node->next) {
-      char *name = (char*)ezxml_attr(param_node,"name");
-      char *type = (char*)ezxml_attr(param_node,"type");
+    for(node = ezxml_child(rule_node,"param"); node; node = node->next) {
+      char *name = (char*)ezxml_attr(node,"name");
+      char *type = (char*)ezxml_attr(node,"type");
 
       mapcache_dimension *dimension = NULL;
 
@@ -966,7 +1019,7 @@ void _configuration_parse_wms_xml(mapcache_context *ctx, ezxml_t node, mapcache_
 
       dimension->name = apr_pstrdup(ctx->pool,name);
 
-      dimension->configuration_parse_xml(ctx,dimension,param_node);
+      dimension->configuration_parse_xml(ctx,dimension,node);
       GC_CHECK_ERROR(ctx);
 
       APR_ARRAY_PUSH(rule->match_params,mapcache_dimension*) = dimension;
@@ -986,11 +1039,16 @@ void _configuration_parse_wms_xml(mapcache_context *ctx, ezxml_t node, mapcache_
 
   wms->getmap_format = mapcache_configuration_get_image_format(cfg,"JPEG");
   if ((rule_node = ezxml_child(node,"format")) != NULL) {
+    const char *attr;
     wms->getmap_format = mapcache_configuration_get_image_format(cfg,rule_node->txt);
     if(!wms->getmap_format) {
       ctx->set_error(ctx,400, "unknown <format> %s for wms service", rule_node->txt);
       return;
     }
+    attr = ezxml_attr(rule_node,"allow_client_override");
+    if(attr && !strcmp(attr,"true")) {
+      wms->allow_format_override = 1;
+    }
   }
 
   if ((rule_node = ezxml_child(node,"resample_mode")) != NULL) {
@@ -1064,6 +1122,7 @@ mapcache_service* mapcache_service_wms_create(mapcache_context *ctx)
   service->getmap_strategy = MAPCACHE_GETMAP_ASSEMBLE;
   service->resample_mode = MAPCACHE_RESAMPLE_BILINEAR;
   service->getmap_format = NULL;
+  service->allow_format_override = 0;
   return (mapcache_service*)service;
 }
 
