@@ -348,6 +348,15 @@ void _create_capabilities_wmts(mapcache_context *ctx, mapcache_request_get_capab
       ezxml_set_txt(ezxml_add_child(layer,"ows:Abstract",0),abstract);
     }
 
+    if(tileset->wgs84bbox.minx != tileset->wgs84bbox.maxx) {
+      ezxml_t bbox = ezxml_add_child(layer,"ows:WGS84BoundingBox",0);
+      ezxml_set_txt(ezxml_add_child(bbox,"ows:LowerCorner",0),
+                    apr_psprintf(ctx->pool,"%f %f",tileset->wgs84bbox.minx, tileset->wgs84bbox.miny));
+      ezxml_set_txt(ezxml_add_child(bbox,"ows:UpperCorner",0),
+                    apr_psprintf(ctx->pool,"%f %f",tileset->wgs84bbox.maxx, tileset->wgs84bbox.maxy));
+    }
+
+
     ezxml_set_txt(ezxml_add_child(layer,"ows:Identifier",0),tileset->name);
 
     style = ezxml_add_child(layer,"Style",0);
@@ -363,8 +372,8 @@ void _create_capabilities_wmts(mapcache_context *ctx, mapcache_request_get_capab
 
     if(tileset->dimensions) {
       for(i=0; i<tileset->dimensions->nelts; i++) {
-        const char **values;
-        const char **value;
+        apr_array_header_t *values;
+        int value_idx;
         mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
         ezxml_t dim = ezxml_add_child(layer,"Dimension",0);
         ezxml_set_txt(ezxml_add_child(dim,"ows:Identifier",0),dimension->name);
@@ -374,10 +383,9 @@ void _create_capabilities_wmts(mapcache_context *ctx, mapcache_request_get_capab
           ezxml_set_txt(ezxml_add_child(dim,"UOM",0),dimension->unit);
         }
         values = dimension->print_ogc_formatted_values(ctx,dimension);
-        value = values;
-        while(*value) {
-          ezxml_set_txt(ezxml_add_child(dim,"Value",0),*value);
-          value++;
+        for(value_idx=0;value_idx<values->nelts;value_idx++) {
+          char *idval = APR_ARRAY_IDX(values,value_idx,char*);
+          ezxml_set_txt(ezxml_add_child(dim,"Value",0),idval);
         }
         dimensionstemplate = apr_pstrcat(ctx->pool,dimensionstemplate,"{",dimension->name,"}/",NULL);
       }
@@ -390,13 +398,6 @@ void _create_capabilities_wmts(mapcache_context *ctx, mapcache_request_get_capab
       }
     }
 
-    if(tileset->wgs84bbox.minx != tileset->wgs84bbox.maxx) {
-      ezxml_t bbox = ezxml_add_child(layer,"ows:WGS84BoundingBox",0);
-      ezxml_set_txt(ezxml_add_child(bbox,"ows:LowerCorner",0),
-                    apr_psprintf(ctx->pool,"%f %f",tileset->wgs84bbox.minx, tileset->wgs84bbox.miny));
-      ezxml_set_txt(ezxml_add_child(bbox,"ows:UpperCorner",0),
-                    apr_psprintf(ctx->pool,"%f %f",tileset->wgs84bbox.maxx, tileset->wgs84bbox.maxy));
-    }
 
     for(i=0; i<tileset->grid_links->nelts; i++) {
       mapcache_grid_link *grid_link = APR_ARRAY_IDX(tileset->grid_links,i,mapcache_grid_link*);
@@ -413,16 +414,17 @@ void _create_capabilities_wmts(mapcache_context *ctx, mapcache_request_get_capab
         int j;
         for(j=0; j<grid_link->grid->nlevels; j++) {
           ezxml_t matrixlimits = ezxml_add_child(limits,"TileMatrixLimits",0);
+          int row;
           ezxml_set_txt(ezxml_add_child(matrixlimits,"TileMatrix",0),
                         apr_psprintf(ctx->pool,"%s:%d",grid_link->grid->name,j));
           ezxml_set_txt(ezxml_add_child(matrixlimits,"MinTileRow",0),
-                        apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].minx));
-          ezxml_set_txt(ezxml_add_child(matrixlimits,"MaxTileRow",0),
-                        apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].maxx-1));
-          ezxml_set_txt(ezxml_add_child(matrixlimits,"MinTileCol",0),
                         apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].miny));
-          ezxml_set_txt(ezxml_add_child(matrixlimits,"MaxTileCol",0),
+          ezxml_set_txt(ezxml_add_child(matrixlimits,"MaxTileRow",0),
                         apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].maxy-1));
+          ezxml_set_txt(ezxml_add_child(matrixlimits,"MinTileCol",0),
+                        apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].minx));
+          ezxml_set_txt(ezxml_add_child(matrixlimits,"MaxTileCol",0),
+                        apr_psprintf(ctx->pool,"%d",grid_link->grid_limits[j].maxx-1));
         }
       }
 
@@ -701,7 +703,7 @@ void _mapcache_service_wmts_parse_request(mapcache_context *ctx, mapcache_servic
           continue;
         }
       }
-      if(tileset->timedimension) {
+      if(!timedim && tileset->timedimension) {
         timedim = key;
         continue;
       }
@@ -779,19 +781,20 @@ void _mapcache_service_wmts_parse_request(mapcache_context *ctx, mapcache_servic
     for(i=0; i<tileset->dimensions->nelts; i++) {
       char *tmpval;
       int ok;
+      const char *value;
       mapcache_dimension *dimension = APR_ARRAY_IDX(tileset->dimensions,i,mapcache_dimension*);
-      const char *value = apr_table_get(dimtable,dimension->name);
+      if(dimension->skip_validation) continue;
+      value = apr_table_get(dimtable,dimension->name);
       if(value) {
         tmpval = apr_pstrdup(ctx->pool,value);
         ok = dimension->validate(ctx,dimension,&tmpval);
         GC_CHECK_ERROR(ctx);
         if(ok != MAPCACHE_SUCCESS) {
           ctx->set_error(ctx,404,"dimension \"%s\" value \"%s\" fails to validate",
-                         dimension->name, value);
+                  dimension->name, value);
           if(kvp) ctx->set_exception(ctx,"InvalidParameterValue","%s",dimension->name);
           return;
         }
-
         /* re-set the eventually modified value in the dimension table */
         apr_table_set(dimtable,dimension->name,tmpval);
       }
@@ -913,7 +916,7 @@ void _mapcache_service_wmts_parse_request(mapcache_context *ctx, mapcache_servic
     mapcache_request_get_tile *req = (mapcache_request_get_tile*)apr_pcalloc(
                                        ctx->pool,sizeof(mapcache_request_get_tile));
     
-    req->request.type = MAPCACHE_REQUEST_GET_TILE;
+    ((mapcache_request*)req)->type = MAPCACHE_REQUEST_GET_TILE;
     if(timedim) {
       timedim_selected = mapcache_timedimension_get_entries_for_value(ctx,
               tileset->timedimension, tileset, grid_link->grid, extent, timedim);
