@@ -37,6 +37,7 @@
 #include <time.h>
 #ifndef _WIN32
 #include <unistd.h>
+#include <sys/stat.h>
 #define USE_FORK
 #include <sys/time.h>
 #endif
@@ -53,6 +54,7 @@ int msqid;
 
 #include <apr_queue.h>
 apr_queue_t *work_queue;
+apr_queue_t *log_queue;
 
 #if defined(USE_OGR) && defined(USE_GEOS)
 #define USE_CLIPPERS
@@ -80,10 +82,13 @@ int verbose = 0;
 int force = 0;
 int sig_int_received = 0;
 int error_detected = 0;
+double percent_failed_allowed = 1.0;
+int n_metatiles_tot=0;
+FILE *failed_log = NULL, *retry_log = NULL;
+#define FAIL_BACKLOG_COUNT 1000
 
 apr_time_t age_limit = 0;
-int seededtilestot=0, seededtiles=0, queuedtilestot=0;
-struct mctimeval lastlogtime,starttime;
+struct mctimeval starttime;
 
 typedef enum {
   MAPCACHE_CMD_SEED,
@@ -96,7 +101,8 @@ typedef enum {
 typedef enum {
   MAPCACHE_ITERATION_UNSET,
   MAPCACHE_ITERATION_DEPTH_FIRST,
-  MAPCACHE_ITERATION_LEVEL_FIRST
+  MAPCACHE_ITERATION_LEVEL_FIRST,
+  MAPCACHE_ITERATION_LOG
 } mapcache_iteration_mode;
 
 mapcache_iteration_mode iteration_mode = MAPCACHE_ITERATION_UNSET;
@@ -108,6 +114,18 @@ struct seed_cmd {
   int z;
 };
 
+typedef enum {
+  MAPCACHE_STATUS_OK,
+  MAPCACHE_STATUS_FAIL,
+  MAPCACHE_STATUS_FINISHED
+} s_status;
+
+struct seed_status {
+  s_status status;
+  int x,y,z;
+  char *msg;
+};
+
 #ifdef USE_FORK
 struct msg_cmd {
   long mtype;
@@ -119,6 +137,7 @@ cmd mode = MAPCACHE_CMD_SEED; /* the mode the utility will be running in: either
 
 int push_queue(struct seed_cmd cmd)
 {
+  struct seed_cmd *pcmd;
 #ifdef USE_FORK
   if(nprocesses > 1) {
     struct msg_cmd mcmd;
@@ -131,7 +150,7 @@ int push_queue(struct seed_cmd cmd)
     return APR_SUCCESS;
   }
 #endif
-  struct seed_cmd *pcmd = calloc(1,sizeof(struct seed_cmd));
+  pcmd = calloc(1,sizeof(struct seed_cmd));
   *pcmd = cmd;
   return apr_queue_push(work_queue,pcmd);
 }
@@ -152,7 +171,7 @@ int pop_queue(struct seed_cmd *cmd)
     return APR_SUCCESS;
   }
 #endif
-  
+
   ret = apr_queue_pop(work_queue, (void**)&pcmd);
   if(ret == APR_SUCCESS) {
     *cmd = *pcmd;
@@ -191,6 +210,7 @@ int trypop_queue(struct seed_cmd *cmd)
 static const apr_getopt_option_t seed_options[] = {
   /* long-option, short-option, has-arg flag, description */
   { "config", 'c', TRUE, "configuration file (/path/to/mapcache.xml)"},
+  { "cache", 'C', TRUE, "override cache used by selected tileset (useful for selectively seeding fallback/multitier caches)"},
 #ifdef USE_CLIPPERS
   { "ogr-datasource", 'd', TRUE, "ogr datasource to get features from"},
 #endif
@@ -203,12 +223,15 @@ static const apr_getopt_option_t seed_options[] = {
 #ifdef USE_CLIPPERS
   { "ogr-layer", 'l', TRUE, "layer inside datasource"},
 #endif
+  { "log-failed", 'L', TRUE, "log failed tiles to [file]"},
   { "mode", 'm', TRUE, "mode: seed (default), delete or transfer" },
   { "metasize", 'M', TRUE, "override metatile size while seeding, eg 8,8" },
   { "nthreads", 'n', TRUE, "number of parallel threads to use (incompatible with -p/--nprocesses)" },
   { "older", 'o', TRUE, "reseed tiles older than supplied date (format: year/month/day hour:minute, eg: 2011/01/31 20:45" },
   { "nprocesses", 'p', TRUE, "number of parallel processes to use (incompatible with -n/--nthreads)" },
+  { "percent", 'P', TRUE, "percent of failed requests allowed from the last 1000 before we abort (default: 1(%), set to 0 to abort on first error)" },
   { "quiet", 'q', FALSE, "don't show progress info" },
+  { "retry-failed", 'R', TRUE, "retry failed requests logged to [file] by --log-failed" },
 #ifdef USE_CLIPPERS
   { "ogr-sql", 's', TRUE, "sql to filter inside layer"},
 #endif
@@ -259,6 +282,10 @@ int ogr_features_intersect_tile(mapcache_context *ctx, mapcache_tile *tile)
 {
   mapcache_metatile *mt = mapcache_tileset_metatile_get(ctx,tile);
   GEOSCoordSequence *mtbboxls = GEOSCoordSeq_create(5,2);
+  GEOSGeometry *mtbbox = GEOSGeom_createLinearRing(mtbboxls);
+  GEOSGeometry *mtbboxg = GEOSGeom_createPolygon(mtbbox,NULL,0);
+  int i;
+  int intersects = 0;
   GEOSCoordSeq_setX(mtbboxls,0,mt->map.extent.minx);
   GEOSCoordSeq_setY(mtbboxls,0,mt->map.extent.miny);
   GEOSCoordSeq_setX(mtbboxls,1,mt->map.extent.maxx);
@@ -269,10 +296,6 @@ int ogr_features_intersect_tile(mapcache_context *ctx, mapcache_tile *tile)
   GEOSCoordSeq_setY(mtbboxls,3,mt->map.extent.maxy);
   GEOSCoordSeq_setX(mtbboxls,4,mt->map.extent.minx);
   GEOSCoordSeq_setY(mtbboxls,4,mt->map.extent.miny);
-  GEOSGeometry *mtbbox = GEOSGeom_createLinearRing(mtbboxls);
-  GEOSGeometry *mtbboxg = GEOSGeom_createPolygon(mtbbox,NULL,0);
-  int i;
-  int intersects = 0;
   for(i=0; i<nClippers; i++) {
     const GEOSPreparedGeometry *clipper = clippers[i];
     if(GEOSPreparedIntersects(clipper,mtbboxg)) {
@@ -286,108 +309,39 @@ int ogr_features_intersect_tile(mapcache_context *ctx, mapcache_tile *tile)
 
 #endif
 
-int lastmsglen = 0;
-void progresslog(int x, int y, int z)
-{
-  char msg[1024];
-  int nworkers;
-  if(quiet) return;
-  nworkers = nthreads;
-  if(nprocesses >= 1) nworkers = nprocesses;
-
-  sprintf(msg,"seeding tile %d %d %d",x,y,z);
-  if(lastmsglen) {
-    char erasestring[1024];
-    int len = MAPCACHE_MIN(1023,lastmsglen);
-    memset(erasestring,' ',len);
-    erasestring[len+1]='\0';
-    sprintf(erasestring,"\r%%%ds\r",lastmsglen);
-    printf(erasestring," ");
-  }
-  lastmsglen = strlen(msg);
-  printf("%s",msg);
-  fflush(NULL);
-  return;
-
-  if(queuedtilestot>nworkers) {
-    struct mctimeval now_t;
-    float duration;
-    float totalduration;
-    seededtilestot = queuedtilestot - nworkers;
-
-    mapcache_gettimeofday(&now_t,NULL);
-    duration = ((now_t.tv_sec-lastlogtime.tv_sec)*1000000+(now_t.tv_usec-lastlogtime.tv_usec))/1000000.0;
-    totalduration = ((now_t.tv_sec-starttime.tv_sec)*1000000+(now_t.tv_usec-starttime.tv_usec))/1000000.0;
-    if(duration>=5) {
-      int Nx, Ny, Ntot, Ncur, ntilessincelast;
-      Nx = (grid_link->grid_limits[z].maxx-grid_link->grid_limits[z].minx)/tileset->metasize_x;
-      Ny = (grid_link->grid_limits[z].maxy-grid_link->grid_limits[z].miny)/tileset->metasize_y;
-      Ntot = Nx*Ny;
-      Ncur = ((y-grid_link->grid_limits[z].miny)/tileset->metasize_y ) * Nx +
-          (x-grid_link->grid_limits[z].minx+1)/tileset->metasize_x;
-      ntilessincelast = seededtilestot-seededtiles;
-      sprintf(msg,"seeding level %d [%d/%d]: %f metatiles/sec (avg since start: %f)",z,Ncur,Ntot,ntilessincelast/duration,
-              seededtilestot/totalduration);
-      lastlogtime=now_t;
-      seededtiles=seededtilestot;
-    } else {
-      return;
-    }
-  } else {
-    sprintf(msg,"seeding level %d",z);
-
-  }
-  if(lastmsglen) {
-    char erasestring[1024];
-    int len = MAPCACHE_MIN(1023,lastmsglen);
-    memset(erasestring,' ',len);
-    erasestring[len+1]='\0';
-    sprintf(erasestring,"\r%%%ds\r",lastmsglen);
-    printf(erasestring," ");
-  }
-  lastmsglen = strlen(msg);
-  printf("%s",msg);
-  fflush(NULL);
-}
-
 cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
 {
   int action = MAPCACHE_CMD_SKIP;
-  int intersects = -1;
-  int tile_exists = force?0:tileset->cache->tile_exists(ctx,tile);
+  int tile_exists;
+
+#ifdef USE_CLIPPERS
+  /* check we are in the requested features before checking the tile */
+  if(nClippers > 0 && ogr_features_intersect_tile(ctx,tile) == 0)
+    return MAPCACHE_CMD_SKIP;
+#endif
+
+  tile_exists = force?0:tileset->_cache->tile_exists(ctx,tileset->_cache,tile);
 
   /* if the tile exists and a time limit was specified, check the tile modification date */
   if(tile_exists) {
     if(age_limit) {
-      if(tileset->cache->tile_get(ctx,tile) == MAPCACHE_SUCCESS) {
+      if(tileset->_cache->tile_get(ctx,tileset->_cache, tile) == MAPCACHE_SUCCESS) {
         if(tile->mtime && tile->mtime<age_limit) {
           /* the tile modification time is older than the specified limit */
-#ifdef USE_CLIPPERS
-          /* check we are in the requested features before deleting the tile */
-          if(nClippers > 0) {
-            intersects = ogr_features_intersect_tile(ctx,tile);
-          }
-#endif
-          if(intersects != 0) {
-            /* the tile intersects the ogr features, or there was no clipping asked for: seed it */
-            if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
-              mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
-              /* if we are in mode transfer, delete it from the dst tileset */
-              if (mode == MAPCACHE_CMD_TRANSFER) {
-                tile->tileset = tileset_transfer;
-                if (tileset_transfer->cache->tile_exists(ctx,tile)) {
-                  mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
-                }
-                tile->tileset = tileset;
-              }
-              action = mode;
-            } else { //if(action == MAPCACHE_CMD_DELETE)
-              action = MAPCACHE_CMD_DELETE;
-            }
-          } else {
-            /* the tile does not intersect the ogr features, and already exists, do nothing */
-            action = MAPCACHE_CMD_SKIP;
+      if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
+        mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
+        /* if we are in mode transfer, delete it from the dst tileset */
+        if (mode == MAPCACHE_CMD_TRANSFER) {
+          tile->tileset = tileset_transfer;
+          if (tile->tileset->_cache->tile_exists(ctx,tile->tileset->_cache, tile)) {
+        mapcache_tileset_tile_delete(ctx,tile,MAPCACHE_TRUE);
           }
+          tile->tileset = tileset;
+        }
+        action = mode;
+      } else { //if(action == MAPCACHE_CMD_DELETE)
+        action = MAPCACHE_CMD_DELETE;
+      }
         }
       } else {
         //BUG: tile_exists returned true, but tile_get returned a failure. not sure what to do.
@@ -401,7 +355,7 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
         /* the tile exists in the source tileset,
            check if the tile exists in the destination cache */
         tile->tileset = tileset_transfer;
-        if (tileset_transfer->cache->tile_exists(ctx,tile)) {
+        if (tile->tileset->_cache->tile_exists(ctx,tile->tileset->_cache, tile)) {
           action = MAPCACHE_CMD_SKIP;
         } else {
           action = MAPCACHE_CMD_TRANSFER;
@@ -415,20 +369,7 @@ cmd examine_tile(mapcache_context *ctx, mapcache_tile *tile)
   } else {
     // the tile does not exist
     if(mode == MAPCACHE_CMD_SEED || mode == MAPCACHE_CMD_TRANSFER) {
-#ifdef USE_CLIPPERS
-      /* check we are in the requested features before deleting the tile */
-      if(nClippers > 0) {
-        if(ogr_features_intersect_tile(ctx,tile)) {
-          action = mode;
-        } else {
-          action = MAPCACHE_CMD_SKIP;
-        }
-      } else {
-        action = mode;
-      }
-#else
       action = mode;
-#endif
     } else {
       action = MAPCACHE_CMD_SKIP;
     }
@@ -450,9 +391,7 @@ void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
   if(sig_int_received || error_detected) { //stop if we were asked to stop by hitting ctrl-c
     //remove all items from the queue
     struct seed_cmd entry;
-    while (trypop_queue(&entry)!=APR_EAGAIN) {
-      queuedtilestot--;
-    }
+    while (trypop_queue(&entry)!=APR_EAGAIN) /*do nothing*/;
     return;
   }
 
@@ -466,8 +405,6 @@ void cmd_recurse(mapcache_context *cmd_ctx, mapcache_tile *tile)
     cmd.z = tile->z;
     cmd.command = action;
     push_queue(cmd);
-    queuedtilestot++;
-    progresslog(tile->x,tile->y,tile->z);
   }
 
   //recurse into our 4 child metatiles
@@ -559,11 +496,16 @@ void cmd_worker()
       if(sig_int_received || error_detected) { //stop if we were asked to stop by hitting ctrl-c
         //remove all items from the queue
         struct seed_cmd entry;
-        while (trypop_queue(&entry)!=APR_EAGAIN) {
-          queuedtilestot--;
-        }
+        while (trypop_queue(&entry)!=APR_EAGAIN) /* do nothing */;
         break;
       }
+      if(iteration_mode == MAPCACHE_ITERATION_LOG) {
+        if(3 != fscanf(retry_log,"%d,%d,%d\n",&x,&y,&z)) {
+          break;
+        } else {
+          printf("from log: %d %d %d\n",x,y,z);
+        }
+      }
       tile->x = x;
       tile->y = y;
       tile->z = z;
@@ -577,8 +519,6 @@ void cmd_worker()
         cmd.z = z;
         cmd.command = action;
         push_queue(cmd);
-        queuedtilestot++;
-        progresslog(x,y,z);
       }
 
       //compute next x,y,z
@@ -603,10 +543,6 @@ void cmd_worker()
     cmd.command = MAPCACHE_CMD_STOP;
     push_queue(cmd);
   }
-
-  if(error_detected && ctx.get_error_message(&ctx)) {
-    printf("%s\n",ctx.get_error_message(&ctx));
-  }
 }
 
 
@@ -633,27 +569,55 @@ void seed_worker()
     if(cmd.command == MAPCACHE_CMD_SEED) {
       /* aquire a lock on the metatile ?*/
       mapcache_metatile *mt = mapcache_tileset_metatile_get(&seed_ctx, tile);
-      int isLocked = mapcache_lock_or_wait_for_resource(&seed_ctx, mapcache_tileset_metatile_resource_key(&seed_ctx,mt));
+      void *lock;
+      int isLocked = mapcache_lock_or_wait_for_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), &lock);
       if(isLocked == MAPCACHE_TRUE) {
         /* this will query the source to create the tiles, and save them to the cache */
         mapcache_tileset_render_metatile(&seed_ctx, mt);
-        mapcache_unlock_resource(&seed_ctx, mapcache_tileset_metatile_resource_key(&seed_ctx,mt));
+        if(GC_HAS_ERROR(&seed_ctx)) {
+          /* temporarily clear error state so we don't mess up with error handling in the locker */
+          void *error;
+          seed_ctx.pop_errors(&seed_ctx,&error);
+          mapcache_unlock_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), lock);
+          seed_ctx.push_errors(&seed_ctx,error);
+        } else {
+          mapcache_unlock_resource(&seed_ctx, seed_ctx.config->locker, mapcache_tileset_metatile_resource_key(&seed_ctx,mt), lock);
+        }
       }
     } else if (cmd.command == MAPCACHE_CMD_TRANSFER) {
       int i;
       mapcache_metatile *mt = mapcache_tileset_metatile_get(&seed_ctx, tile);
       for (i = 0; i < mt->ntiles; i++) {
         mapcache_tile *subtile = &mt->tiles[i];
-        mapcache_tileset_tile_get(&seed_ctx, subtile);
-        subtile->tileset = tileset_transfer;
-        tileset_transfer->cache->tile_set(&seed_ctx, subtile);
+        int cache_ret;
+        cache_ret = tileset->_cache->tile_get(&seed_ctx, tileset->_cache, subtile);
+        if(cache_ret == MAPCACHE_SUCCESS && !GC_HAS_ERROR(&seed_ctx)) {
+          subtile->tileset = tileset_transfer;
+          subtile->tileset->_cache->tile_set(&seed_ctx, subtile->tileset->_cache, subtile);
+        }
       }
     } else { //CMD_DELETE
       mapcache_tileset_tile_delete(&seed_ctx,tile,MAPCACHE_TRUE);
     }
-    if(seed_ctx.get_error(&seed_ctx)) {
-      error_detected++;
-      ctx.log(&ctx,MAPCACHE_INFO,seed_ctx.get_error_message(&seed_ctx));
+
+    {
+      struct seed_status *st = calloc(1,sizeof(struct seed_status));
+      st->x=tile->x;
+      st->y=tile->y;
+      st->z=tile->z;
+      if(seed_ctx.get_error(&seed_ctx)) {
+        st->status = MAPCACHE_STATUS_FAIL;
+        st->msg = strdup(seed_ctx.get_error_message(&seed_ctx));
+        seed_ctx.clear_errors(&seed_ctx);
+      } else {
+        st->status = MAPCACHE_STATUS_OK;
+      }
+      ret = apr_queue_push(log_queue,(void*)st);
+      if(ret != APR_SUCCESS)
+      {
+        printf("FATAL ERROR: unable to log progress\n");
+        break;
+      }
     }
   }
 }
@@ -669,6 +633,66 @@ static void* APR_THREAD_FUNC seed_thread(apr_thread_t *thread, void *data) {
   return NULL;
 }
 
+static void* APR_THREAD_FUNC log_thread_fn(apr_thread_t *thread, void *data) {
+  size_t cur;
+  double last_time;
+  double now_time;
+  int i;
+  int nfailed;
+  int ntotal;
+  double pct;
+  char failed[FAIL_BACKLOG_COUNT];
+  memset(failed,-1,FAIL_BACKLOG_COUNT);
+  cur=0;
+  last_time=0;
+  while(1) {
+    struct seed_status *st;
+    apr_status_t ret = apr_queue_pop(log_queue, (void**)&st);
+    if(ret != APR_SUCCESS || !st) break;
+    if(st->status == MAPCACHE_STATUS_FINISHED)
+      return NULL;
+    if(st->status == MAPCACHE_STATUS_OK) {
+      failed[cur]=0;
+      n_metatiles_tot++;
+      if(!quiet) {
+        struct mctimeval now;
+        mapcache_gettimeofday(&now,NULL);
+        now_time = now.tv_sec + now.tv_usec / 1000000.0;
+        if((now_time - last_time) > 1.0) {
+          printf("                                                                                               \r");
+          printf("seeded %d tiles, now at z%d x%d y%d\r",n_metatiles_tot*tileset->metasize_x*tileset->metasize_y, st->z,st->x,st->y);
+          fflush(stdout);
+          last_time = now_time;
+        }
+      }
+    } else {
+      /* count how many errors and successes we have */
+      failed[cur]=1;
+      nfailed=0;
+      ntotal=0;
+      if(failed_log) {
+        fprintf(failed_log,"%d,%d,%d\n",st->x,st->y,st->z);
+      }
+      for(i=0; i<FAIL_BACKLOG_COUNT; i++) {
+        if(failed[i]>=0) ntotal++;
+        if(failed[i]==1) nfailed++;
+      }
+      ctx.log(&ctx, MAPCACHE_WARN, "failed to seed tile z%d,x%d,y%d:\n%s\n", st->z,st->x,st->y,st->msg);
+      pct = ((double)nfailed / (double)ntotal) * 100;
+      if(pct > percent_failed_allowed) {
+        ctx.log(&ctx, MAPCACHE_ERROR, "aborting seed as %.1f%% of the last %d requests failed\n", pct, FAIL_BACKLOG_COUNT);
+        error_detected = 1;
+        return NULL;
+      }
+    }
+    if(st->msg) free(st->msg);
+    free(st);
+    cur++;
+    cur %= FAIL_BACKLOG_COUNT;
+  }
+  return NULL;
+}
+
 void
 notice(const char *fmt, ...)
 {
@@ -734,10 +758,12 @@ int main(int argc, const char **argv)
   apr_getopt_t *opt;
   const char *configfile=NULL;
   apr_thread_t **threads;
+  apr_thread_t *log_thread;
   apr_threadattr_t *thread_attrs;
   const char *tileset_name=NULL;
   const char *tileset_transfer_name=NULL;
   const char *grid_name = NULL;
+  const char *cache_override = NULL;
   int *zooms = NULL;//[2];
   mapcache_extent *extent = NULL;//[4];
   int optch;
@@ -752,6 +778,8 @@ int main(int argc, const char **argv)
   double *extent_array = NULL;
 
 #ifdef USE_CLIPPERS
+  OGRFeatureH hFeature;
+  GEOSWKTReader *geoswktreader;
   const char *ogr_where = NULL;
   const char *ogr_layer = NULL;
   const char *ogr_sql = NULL;
@@ -762,15 +790,12 @@ int main(int argc, const char **argv)
   (void) signal(SIGINT,handle_sig_int);
   apr_pool_create(&ctx.pool,NULL);
   mapcache_context_init(&ctx);
-  ctx.process_pool = ctx.pool;
   cfg = mapcache_configuration_create(ctx.pool);
   ctx.config = cfg;
   ctx.log= mapcache_context_seeding_log;
   apr_getopt_init(&opt, ctx.pool, argc, argv);
 
-  seededtiles=seededtilestot=queuedtilestot=0;
   mapcache_gettimeofday(&starttime,NULL);
-  lastlogtime=starttime;
   argdimensions = apr_table_make(ctx.pool,3);
 
 
@@ -792,6 +817,9 @@ int main(int argc, const char **argv)
       case 'c':
         configfile = optarg;
         break;
+      case 'C':
+        cache_override = optarg;
+        break;
       case 'g':
         grid_name = optarg;
         break;
@@ -810,6 +838,18 @@ int main(int argc, const char **argv)
           return usage(argv[0],"invalid iteration mode, expecting \"drill-down\" or \"level-by-level\"");
         }
         break;
+      case 'L':
+        failed_log = fopen(optarg,"w");
+        if(!failed_log) {
+          return usage(argv[0],"failed to open -L|--log-failed file for writing");
+        }
+        break;
+      case 'R':
+        retry_log = fopen(optarg,"r");
+        if(!retry_log) {
+          return usage(argv[0],"failed to open -R|--retry_failed file for writing");
+        }
+        break;
       case 'm':
         if(!strcmp(optarg,"delete")) {
           mode = MAPCACHE_CMD_DELETE;
@@ -835,6 +875,12 @@ int main(int argc, const char **argv)
 #else
         return usage(argv[0], "multi process seeding not available on this platform");
 #endif
+      case 'P':
+        percent_failed_allowed = (double)strtol(optarg, NULL, 10);
+        if(percent_failed_allowed<0 || percent_failed_allowed>100 )
+          return usage(argv[0], "failed to parse percent, expecting number between 0 and 100");
+        break;
+
       case 'e':
         if ( MAPCACHE_SUCCESS != mapcache_util_extract_double_list(&ctx, (char*)optarg, ",", &extent_array, &n) ||
              n != 4 || extent_array[0] >= extent_array[2] || extent_array[1] >= extent_array[3] ) {
@@ -914,6 +960,7 @@ int main(int argc, const char **argv)
     mapcache_configuration_post_config(&ctx,cfg);
     if(ctx.get_error(&ctx))
       return usage(argv[0],ctx.get_error_message(&ctx));
+    mapcache_connection_pool_create(&ctx.connection_pool, ctx.pool);
   }
 
 #ifdef USE_CLIPPERS
@@ -926,6 +973,7 @@ int main(int argc, const char **argv)
   }
 
   if(ogr_datasource) {
+    int f=0;
     OGRDataSourceH hDS = NULL;
     OGRLayerH layer = NULL;
     OGRRegisterAll();
@@ -973,21 +1021,20 @@ int main(int argc, const char **argv)
     clippers = (const GEOSPreparedGeometry**)malloc(nClippers*sizeof(GEOSPreparedGeometry*));
 
 
-    OGRFeatureH hFeature;
-    GEOSWKTReader *geoswktreader = GEOSWKTReader_create();
+    geoswktreader = GEOSWKTReader_create();
     OGR_L_ResetReading(layer);
-    extent = apr_palloc(ctx.pool,4*sizeof(mapcache_extent));
-    int f=0;
+    extent = apr_palloc(ctx.pool,sizeof(mapcache_extent));
     while( (hFeature = OGR_L_GetNextFeature(layer)) != NULL ) {
+      char *wkt;
+      GEOSGeometry *geosgeom;
+      OGREnvelope ogr_extent;
       OGRGeometryH geom = OGR_F_GetGeometryRef(hFeature);
       if(!geom ||  !OGR_G_IsValid(geom)) continue;
-      char *wkt;
       OGR_G_ExportToWkt(geom,&wkt);
-      GEOSGeometry *geosgeom = GEOSWKTReader_read(geoswktreader,wkt);
+      geosgeom = GEOSWKTReader_read(geoswktreader,wkt);
       free(wkt);
       clippers[f] = GEOSPrepare(geosgeom);
       //GEOSGeom_destroy(geosgeom);
-      OGREnvelope ogr_extent;
       OGR_G_GetEnvelope  (geom, &ogr_extent);
       if(f == 0) {
         extent->minx = ogr_extent.MinX;
@@ -1036,6 +1083,38 @@ int main(int argc, const char **argv)
         return usage(argv[0],"grid not configured for tileset");
       }
     }
+#ifdef USE_CLIPPERS
+    if(ogr_datasource) {
+      /* check that the provided ogr features are compatible with the grid units */
+      if(grid_link->grid->unit == MAPCACHE_UNIT_DEGREES) {
+        if(extent->minx < -181.0 ||
+           extent->maxx > 181.0 ||
+           extent->miny < -91.0 ||
+           extent->maxy > 91.0) {
+          printf("\n********************************************************************************\n"
+                 "* WARNING!!!: you are seeding a grid in latlon degreees,\n"
+                 "* but your provided OGR intersection features span (%f,%f,%f,%f).\n"
+                 "* this seems like an error, you should be providing OGR features that\n"
+                 "* are in the same projection as the grid you want to seed\n"
+                 "********************************************************************************\n\n",
+                 extent->minx, extent->miny, extent->maxx, extent->maxy);
+        }
+      } else {
+        if(extent->minx > -181.0 &&
+           extent->maxx < 181.0 &&
+           extent->miny > -91.0 &&
+           extent->maxy < 91.0) {
+          printf("\n********************************************************************************\n"
+                 "* WARNING!!!: you are seeding a grid that is not in latlon degreees,\n"
+                 "* but your provided OGR intersection features span (%f,%f,%f,%f) which seem to be in degrees.\n"
+                 "* this seems like an error, you should be providing OGR features that\n"
+                 "* are in the same projection as the grid you want to seed\n"
+                 "********************************************************************************\n\n",
+                 extent->minx, extent->miny, extent->maxx, extent->maxy);
+        }
+      }
+    }
+#endif
     if(iteration_mode == MAPCACHE_ITERATION_UNSET) {
       if(!strcmp(grid_link->grid->name,"g") || !strcmp(grid_link->grid->name,"WGS84")
               || !strcmp(grid_link->grid->name,"GoogleMapsCompatible")) {
@@ -1044,6 +1123,10 @@ int main(int argc, const char **argv)
         iteration_mode = MAPCACHE_ITERATION_LEVEL_FIRST;
       }
     }
+    if(retry_log) {
+      iteration_mode = MAPCACHE_ITERATION_LOG;
+    }
+
     if(minzoom == -1 && maxzoom == -1) {
       minzoom = grid_link->minz;
       maxzoom = grid_link->maxz - 1;
@@ -1067,6 +1150,14 @@ int main(int argc, const char **argv)
       }
     }
 
+    if(cache_override) {
+      mapcache_cache *co = mapcache_configuration_get_cache(cfg, cache_override);
+      if(!co) {
+        return usage(argv[0], "overrided cache\"%s\" to not found in configuration", cache_override);
+      } else {
+        tileset->_cache = co;
+      }
+    }
 
   }
 
@@ -1160,11 +1251,22 @@ int main(int argc, const char **argv)
         apr_table_set(dimensions,tileset->timedimension->key,APR_ARRAY_IDX(timedim_selected,0,char*));
       } else {
         return usage(argv[0],"tileset references a TIME dimension, but none supplied on commandline. (hint: -D %s=<timestamp>",tileset->timedimension->key);
-        
+
       }
     }
   }
 
+  {
+  /* start the logging thread */
+    //create the queue where the seeding statuses will be put
+    apr_threadattr_t *log_thread_attrs;
+    apr_queue_create(&log_queue,2,ctx.pool);
+
+    //start the rendering threads.
+    apr_threadattr_create(&log_thread_attrs, ctx.pool);
+    apr_thread_create(&log_thread, log_thread_attrs, log_thread_fn, NULL, ctx.pool);
+  }
+
   if(nthreads == 0 && nprocesses == 0) {
     nthreads = 1;
   }
@@ -1238,23 +1340,32 @@ int main(int argc, const char **argv)
       apr_thread_join(&rv, threads[n]);
     }
   }
-  if(ctx.get_error(&ctx)) {
-    printf("%s",ctx.get_error_message(&ctx));
+  {
+    struct seed_status *st = calloc(1,sizeof(struct seed_status));
+    st->status = MAPCACHE_STATUS_FINISHED;
+    apr_queue_push(log_queue,(void*)st);
+    apr_thread_join(&rv, log_thread);
   }
 
-  if(seededtilestot>0) {
+  if(n_metatiles_tot>0) {
     struct mctimeval now_t;
     float duration;
+    int ntilestot = n_metatiles_tot*tileset->metasize_x*tileset->metasize_y;
     mapcache_gettimeofday(&now_t,NULL);
     duration = ((now_t.tv_sec-starttime.tv_sec)*1000000+(now_t.tv_usec-starttime.tv_usec))/1000000.0;
-    printf("\nseeded %d metatiles at %g tiles/sec\n",seededtilestot, seededtilestot/duration);
+
+    printf("\nseeded %d metatiles (%d tiles) in %.1f seconds at %.1f tiles/sec\n",n_metatiles_tot, ntilestot, duration, ntilestot/duration);
+  } else {
+    if(!error_detected) {
+      printf("0 tiles needed to be seeded, exiting\n");
+    }
   }
   apr_terminate();
 
   if (error_detected > 0) {
-    exit(1);  
+    exit(1);
   }
-  
+
   return 0;
 }
 /* vim: ts=2 sts=2 et sw=2
